/******************************************************************************
*                                                                             *
*  File: AutoConfigServer.cpp                                                 *
*                                                                             *
*  Description: The main() function for the Automatic Configuration Server.   *
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************
   Revision Control Information
   $Revision: 1.6 $
   $Log: AutoConfigServer.cpp,v $
   Revision 1.6  2000/10/09 20:26:22  f-kon
   New namespace for CORBA modules combined with Tomo's changes to support Windows.

   Revision 1.5  2000/09/30 18:23:25  yamane
   Patches for Windows

   Revision 1.3  2000/01/27 17:41:38  f-kon
   several small improvements

   Revision 1.2  2000/01/18 22:08:38  f-kon
   various optimizations and bug fixes

   Revision 1.1.1.2  1999/10/15 22:10:22  yamane
   Clean up for native exception support.

   Revision 1.1.1.1  1999/10/12 00:42:46  f-kon
   Version compatible with TAO 1.0

   Revision 1.2  1999/08/11 13:57:26  f-kon
   improved version with most bugs fixed

   Revision 1.1  1999/08/03 22:47:47  f-kon
   Initial version of the Automatic Configuration Systems. It is not yet fully debugged.

******************************************************************************/

#include "AutoConfigServer.h"
#include "ace/CORBA_macros.h"
#include "ace/Get_Opt.h"
//#include "2k/Utilities/KKUtil.h"
#include "orbsvcs/orbsvcs/CosNamingC.h"

int
main (int argc, char**argv)
{
  AutoConfigServer acServer;

  if (acServer.init (argc, argv) < 0)
    return -1;

  ACE_DECLARE_NEW_CORBA_ENV;
  ACE_TRY
    {
      acServer.run (ACE_TRY_ENV);
      ACE_TRY_CHECK;
    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION (ACE_ANY_EXCEPTION, "AutoConfigServer main function");
      return -1;
    }
  ACE_ENDTRY;
  ACE_CHECK_RETURN (1);

  return 0;

}// main ()


AutoConfigServer::AutoConfigServer ()
  : verbose_(0), ior_output_file_(0), time_(0)
{
}// constructor

int
AutoConfigServer::init (int argc, char**argv)
{
  CORBA::String_var str;
  CosNaming::Name query_name (4);
  
  ACE_DECLARE_NEW_CORBA_ENV;
  ACE_TRY
    {
      orbManager_.init_child_poa (argc,
                                  argv,
                                  "child_poa",
                                  ACE_TRY_ENV);
      ACE_TRY_CHECK;
      
      if (parse_args (argc, argv) < 0)
        return -1;
      
      str = orbManager_.activate_under_child_poa ("AutoConfigServer",
                                                  &autoConfigImpl_, 
                                                  ACE_TRY_ENV);
      ACE_TRY_CHECK;
 
      autoConfigImpl_.init(ACE_TRY_ENV);
      ACE_TRY_CHECK;

      if (verbose_)
        ACE_OS::printf("The IOR is: <%s>\n", str.in ());

      if (ior_output_file_)
        {
          ACE_OS::fprintf (ior_output_file_, "%s\n", str.in ());
          ACE_OS::fclose (ior_output_file_);
        }

    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION (ACE_ANY_EXCEPTION, "AutoConfigServer::init");
      return -1;
    }
  ACE_ENDTRY;
  
  // Now, register ourselves with the name server.
  CosNaming::NamingContext_ptr nameServiceRoot;
  ACE_TRY_EX (A)
    {
      CORBA::Object_var obj = orbManager_.orb()->resolve_initial_references ("NameService");
      
      if (CORBA::is_nil (obj.in ()))
        ACE_ERROR_RETURN ((LM_ERROR, 
                           "(%P|%t) Unable to resolve the Name Service.\n"), -1);
      
      nameServiceRoot = CosNaming::NamingContext::_narrow (obj.in (), ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (A); 
      
      if (CORBA::is_nil (nameServiceRoot))
        ACE_ERROR_RETURN ((LM_ERROR, 
                           "(%P|%t) Unable to resolve the Name Service.\n"), -1);
      
      query_name.length (1);
      query_name[0].id = CORBA::string_dup ("local"); // is this leaking?
      nameServiceRoot->bind_new_context (query_name, ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (A);
    }
  ACE_CATCH (CosNaming::NamingContext::AlreadyBound, varName)
    {
      // It was already bound. So, that's OK we don't need to do anything.
    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION 
        (ACE_ANY_EXCEPTION, 
         "ComponentRepository::init caught exception while contacting the Name Service");
      return -1;
    }
  ACE_ENDTRY;

  ACE_TRY_EX (B)
    {
       query_name.length (2);
       query_name[1].id = CORBA::string_dup ("machines"); // is this leaking? 
       nameServiceRoot->bind_new_context (query_name, ACE_TRY_ENV);
       ACE_TRY_CHECK_EX (B);
    }
  ACE_CATCH (CosNaming::NamingContext::AlreadyBound, varName)
    {
      // It was already bound. So, that's OK we don't need to do anything.
    }
  ACE_ENDTRY;
  
  ACE_TRY_EX (C)
    {
      char machine_name[MAXHOSTNAMELEN];
      ACE_OS::hostname (machine_name, MAXHOSTNAMELEN);
      query_name.length (3);
      query_name[2].id = CORBA::string_dup (machine_name); // Is this leaking?

      nameServiceRoot->bind_new_context (query_name, ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (C);
    }
  ACE_CATCH (CosNaming::NamingContext::AlreadyBound, varName)
    {
      // It was already bound. So, that's OK we don't need to do anything.
    }
  ACE_ENDTRY;

  ACE_TRY_EX (D)
    {
      char autoConfName[32];
      ACE_OS::sprintf (autoConfName, "AutoConfig:%d", ACE_OS::getpid());

      query_name.length (4);
      query_name[3].id = CORBA::string_dup (autoConfName); // Is this leaking?
      
      nameServiceRoot->rebind (query_name, orbManager_.orb()->string_to_object(str), ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (D);
    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION 
        (ACE_ANY_EXCEPTION, 
         "ComponentRepository::init caught exception while registering with the Name Service");
      return -1;
    }
  ACE_ENDTRY;

  return 0;
  
}// init ()

int
AutoConfigServer::run (CORBA_Environment& ACE_TRY_ENV)
{
  if (time_ == 0)
      return orbManager_.run (ACE_TRY_ENV);
  else
    {
      ACE_Time_Value *t = new ACE_Time_Value (time_);
      return orbManager_.run (ACE_TRY_ENV, t);
    }

}// run()


int
AutoConfigServer::parse_args (int argc, char *argv[])
{
  ACE_Get_Opt get_opts (argc, argv, "do:t:v");
  int c;
//  int size;
  int time;

  while ((c = get_opts ()) != -1)
    switch (c)
      {
      case 'd':  // debug flag.
        TAO_debug_level++;
        break;
      case 'v':  // debug flag.
        verbose_ = 1;
        break;
      case 'o': // outputs the AutoConfig service ior to a file.
        ior_output_file_ =
          ACE_OS::fopen (get_opts.optarg, "w");

        if (ior_output_file_ == 0)
          ACE_ERROR_RETURN ((LM_ERROR,
                             "Unable to open %s for writing: %p\n",
                             get_opts.optarg), -1);
        break;
      case 't':
        time = ACE_OS::atoi (get_opts.optarg);
        if (time >= 0)
          time_ = time;
        break;
      case '?':
      default:
        ACE_ERROR_RETURN ((LM_ERROR,
                           "usage:  %s "
                           "[-o <ior_output_file> ]"
                           "[-t <time_in_seconds> ]"
                           "[-v ]"
                           "\n",
                           argv [0]),
                          -1);
      }
  return 0;

}// parse_args ()
