/******************************************************************************
*                                                                             *
*  File: AutoConfig_i.cpp                                                     *
*                                                                             *
*  Description: Implementation of the Automatic Configuration Service.        *
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************
   Revision Control Information
   $Revision: 1.6 $
   $Log: AutoConfig_i.cpp,v $
   Revision 1.6  2000/10/09 20:26:20  f-kon
   New namespace for CORBA modules combined with Tomo's changes to support Windows.

   Revision 1.5  2000/09/30 18:24:00  yamane
   Patches for Windows

   Revision 1.4  2000/01/27 17:41:40  f-kon
   several small improvements

   Revision 1.3  2000/01/24 15:56:40  f-kon
   minor improvements

   Revision 1.2  2000/01/18 22:08:39  f-kon
   various optimizations and bug fixes

   Revision 1.1.1.3  1999/10/15 22:10:24  yamane
   Clean up for native exception support.

   Revision 1.1.1.2  1999/10/15 21:45:38  yamane
   Clean up for native exception support

   Revision 1.1.1.1  1999/10/12 00:42:46  f-kon
   Version compatible with TAO 1.0

   Revision 1.3  1999/08/11 13:57:27  f-kon
   improved version with most bugs fixed

   Revision 1.2  1999/08/10 15:49:05  f-kon
   little bug fixes

   Revision 1.1  1999/08/03 22:47:48  f-kon
   Initial version of the Automatic Configuration Systems. It is not yet fully debugged.

******************************************************************************/

#include "AutoConfig_i.h"
#include "ace/Dynamic_Service.h"
#include "AutoConfigExceptionsC.h"
#include "SPDFParser.h"
#include "SimpleResolver.h"
#include "dynamicTAO/src/Components/Simple_Component_Factory.h"

#define MEASURING_PERFORMANCE
#ifdef MEASURING_PERFORMANCE
#include "tao/Timeprobe.h"
#endif

AutoConfig_i::AutoConfig_i ()
  : parser_ (0), resolver_ (0)
{
}// constructor

////////////////////////////////////////////////////////////////////////////////
// Initializes the AutoConfig implementation by findig a reference to the local
// DynamicConfigurator and by creating a SPDFParser and a SimpleResolver.
////////////////////////////////////////////////////////////////////////////////
// IMPROVE: In the future, it should be extended to be able to generate
// different kinds of parsers and resolvers.
//          In the far future, we should have a list of parsers and use different ones
// according to the type of the prerequisite specification. 
////////////////////////////////////////////////////////////////////////////////
void
AutoConfig_i::init (CORBA_Environment& ACE_TRY_ENV)
{
  // First locate our local dynamicConfigurator and save it in the appropriate
  // member variable. We do that by calling resolve_initial_references on the ORB.
  //  CORBA::ORB_var orb = TAO_ORB_Core_instance ()->orb ();
  
  CORBA::Object_var dynconf_obj = 
    TAO_ORB_Core_instance ()->orb ()->resolve_initial_references ("DynamicConfigurator");
  
  if (CORBA::is_nil (dynconf_obj.in ()))
    {
      ACE_ERROR ((LM_ERROR, "(%P|%t) Unable to resolve the DynamicConfigurator Service.\n"));
      ACE_THROW (TwoK2::AutoConfig::LoadFailure ());
    }
  else 
    dynamicConfigurator_ = DynamicConfigurator::_narrow (dynconf_obj.in ());

  // we should put some default things to be dynamically loaded here.
  parser_ = new SPDFParser ();
  resolver_ = new SimpleResolver ();

}// init ()


////////////////////////////////////////////////////////////////////////////////
// Uses the PrerequisteParser and PrerequisiteResolver to (1) fetch a component
// from a given implementation repository (copying it to the local repository)
// (2) load the component into the runtime and (3) recursively load all the
// prerequisites of the requested component.
// This function is the one usually called by clients of the Automatic
// Configuration Service.
// Throws LoadFailure and NotFound (the latter not implememted yet) exceptions.
////////////////////////////////////////////////////////////////////////////////
TwoK::Configuration::ComponentConfigurator_ptr
AutoConfig_i::load_component (DynamicConfigurator_ptr implRep,
                              const TwoK2::CompSpec & compSpecification,
                              TwoK::Configuration::CCFactory_ptr factory,
                              CORBA::Environment &ACE_TRY_ENV)
    ACE_THROW_SPEC ((
      CORBA::SystemException,
      TwoK2::AutoConfig::NotFound,
      TwoK2::AutoConfig::LoadFailure
    ))
{
  TwoK::Configuration::ComponentConfigurator_ptr ptr;

#ifdef MEASURING_PERFORMANCE
  ACE_Time_Value time1, time2, time_difference;
  double mtime;
  time1 = ACE_OS::gettimeofday();
#endif

  ACE_TRY {
    parser_->init   (dynamicConfigurator_.in (), resolver_, ACE_TRY_ENV);
    ACE_TRY_CHECK;
  }
  ACE_CATCHANY
  {
    ACE_THROW_RETURN (TwoK2::AutoConfig::NotFound (), 
                      TwoK::Configuration::ComponentConfigurator::_nil ());
  }
  ACE_ENDTRY;

  ACE_TRY_EX(L2) {
    resolver_->init (_this (ACE_TRY_ENV), implRep, 
                     dynamicConfigurator_.in (), factory, ACE_TRY_ENV);
    ACE_TRY_CHECK_EX(L2);
  }
  ACE_CATCHANY
  {
    ACE_THROW_RETURN (TwoK2::AutoConfig::NotFound (), 
                      TwoK::Configuration::ComponentConfigurator::_nil ());
  }
  ACE_ENDTRY;

  ACE_TRY_EX(L3)
  {
    ptr = resolver_->resolvePrereq (TwoK::Configuration::ComponentConfigurator::_nil (), NULL, 
                                    compSpecification, ACE_TRY_ENV);
    ACE_TRY_CHECK_EX(L3);
  }
  ACE_CATCH (NotFound, ex) // Do not prefix with the AutoConfig:: here. It
                           // seems that we'll have to refactor this to get rid of
                           // AutoConfigExceptions.idl.
    {
      ACE_THROW_RETURN (TwoK2::AutoConfig::NotFound (), TwoK::Configuration::ComponentConfigurator::_nil ());
    }
  ACE_CATCHANY
  {
    ACE_THROW_RETURN (TwoK2::AutoConfig::LoadFailure (), 
                      TwoK::Configuration::ComponentConfigurator::_nil ());
  }
  ACE_ENDTRY;

#ifdef MEASURING_PERFORMANCE
      time2 = ACE_OS::gettimeofday();
      time_difference = time2 - time1;
      mtime = (double) time_difference.msec();
      ACE_OS::printf("total time as perceived by AutoConfig_i::load_component = %f ms\n", mtime);
#endif

  return(ptr);  
}// load_component ()


////////////////////////////////////////////////////////////////////////////////
// Loads a component in local address space. Creates a CORBAConfigurator for it
// and makes sure all its dependencies are resolved by calling the prerequiste
// parser.
// This function is usually used by the Automatic Configuration Service itself 
// and not by clients of this service.
// Throws LoadFailure and NotFound (the latter not implememted yet) exceptions.
////////////////////////////////////////////////////////////////////////////////
TwoK::Configuration::ComponentConfigurator_ptr
AutoConfig_i::load_component_aux (const TwoK2::Name & compSpecification,
                                  TwoK::Configuration::CCFactory_ptr factory,
                                  CORBA::Environment &ACE_TRY_ENV)
    ACE_THROW_SPEC ((
      CORBA::SystemException,
      TwoK2::AutoConfig::NotFound,
      TwoK2::AutoConfig::LoadFailure
    ))
{
  int version;

  // Initialize compConfig with nil in case we fail to create a non-nil one.
  TwoK::Configuration::ComponentConfigurator_var compConfig = 
    TwoK::Configuration::ComponentConfigurator::_nil ();
  
  // First figure out which component should be loaded. 
 
  // IMPORTANT NOTE: The following temporary solution assumes that the
  // CosNaming::Name has two components (categoryName and implementation
  // name). 
  // another IMPORTANT NOTE: Due to a bug in an early version of
  // the TAO_IDL compiler, we are simply using a string sequence here. It should be
  // fixed in the current version of the compiler.

  //  const char *categoryName = (const char *) compSpecification[0];
  const char *impName = (const char *) compSpecification[1];
  const char *params;

  // MOD
  // Modificado por Fabio e Jeferson 29/01/2002
  char localCatName[MAXNAMELEN];
  ACE_OS::strcpy (localCatName, "downloaded_");
  ACE_OS::strcat (localCatName, compSpecification[0]);
  const char *categoryName = (const char *) localCatName;
  // FIM-MODg


  if (compSpecification.length () == 3)
    params = (const char *) compSpecification[2];
  else
    params = "";
  
  ACE_TRY
    {
      // Dynamically loads the new implementation into our address space. This
      // will call init () on the component.
      version = dynamicConfigurator_->load_implementation(categoryName, impName, params, 
                                                          factory, compConfig, ACE_TRY_ENV);
      ACE_TRY_CHECK;
    }
  ACE_CATCHANY
    {
      ACE_ANY_EXCEPTION._tao_print_exception("Could not load_implementation in local address space.");
      ACE_THROW_RETURN (TwoK2::AutoConfig::LoadFailure (), compConfig._retn ());
    }
  ACE_ENDTRY;

  /////////////////////////////////////////////////////////////////////////////      
  // Now that the implementation is loaded, let's call the prerequisite parser
  // to take care of its prerequisites.
  /////////////////////////////////////////////////////////////////////////////      
  TwoK2::CompSpec compSpec;
  compSpec.comp_name (compSpecification);
  
  ACE_TRY_EX(LABEL2) {
    parser_->parse (compSpec, compConfig.in (), ACE_TRY_ENV);
    ACE_TRY_CHECK_EX(LABEL2);
  }
//  ACE_CATCH (BaseException , e)
//  ACE_CATCH (BaseException , e)
  ACE_CATCHANY
    {
      ACE_DEBUG ((LM_WARNING, "(%P|%t) Could not parse prerequisites for %s.\n", impName));

      // Unfortunately, I cannot really catch BaseExceptions here using the ACE macros unless
      // they are turned into a subclass of some ACE-defined exceptions.
      // Later this should be fixed. 

      // if the exception type is NotFound, this is not a problem, we may ignore it.
      // if the exception type is ParseError, we should throw the following exception
      //      ACE_THROW_RETURN (TwoK2::AutoConfig::LoadFailure (), compConfig._retn ());
    }
  ACE_ENDTRY;
  
  //////////////////////////////////////////////////////////////////////////////
  // Now that all the prerequisites were solved, call start () on the Component.
  //////////////////////////////////////////////////////////////////////////////

  char name[MAXNAMELEN];
  ACE_OS::sprintf(name, "%s:%d", categoryName, version);

  Simple_Strategy_Factory *ssf =
    ACE_Dynamic_Service<Simple_Strategy_Factory>::instance (name);

  if (ssf == 0)
    {
      ACE_DEBUG ((LM_ERROR, "(%P|%t) Could not locate loaded implementation %s.\n",
                 name));
      ACE_THROW_RETURN (TwoK2::AutoConfig::LoadFailure (), compConfig._retn ());
    }
  
  if (!ssf->typeHasCorbaConfigurator())
    {
      ACE_DEBUG ((LM_DEBUG, 
                  "(%P|%t) Warning: The type of %s does not have a CORBA Configurator.\n",
                 name));
      return compConfig._retn ();
    }

  ACE_DEBUG ((LM_DEBUG, "Type %s does have a CORBA Configurator.\n", name));

  // At this point it seems we can be sure it's a Simple_Component_Factory. 
  // Later, when available, we should use RTTI here.
  Simple_Component_Factory *scf = (Simple_Component_Factory *) ssf;
  
  if (scf == 0 || scf->start (compConfig) < 0)
    {
      ACE_ERROR ((LM_ERROR, "(%P|%t) start method on Component %s failed.\n", impName));
      ACE_THROW_RETURN (TwoK2::AutoConfig::LoadFailure (), compConfig._retn ());
    }

  // Ufs, we got here finally. Everything went well, let's return the component configurator.

  return compConfig._retn ();

}// load_component_aux ()


////////////////////////////////////////////////////////////////////////////////
// Loads the prerequisites related to a compSpecification. It assumes that the 
// prerequisite file is already available in the local machine and it does not
// load the component itself. Creates a CORBAConfigurator using the givem 
// factory and and makes sure all its dependencies are resolved by calling the 
// prerequiste parser.
// Throws LoadFailure and NotFound (the latter not implememted yet) exceptions.
////////////////////////////////////////////////////////////////////////////////
TwoK::Configuration::ComponentConfigurator_ptr
AutoConfig_i::load_prereq (const TwoK2::Name & prereqSpecification,
                           TwoK::Configuration::CCFactory_ptr factory,
                           CORBA::Environment &ACE_TRY_ENV)
  ACE_THROW_SPEC ((CORBA::SystemException,
                   TwoK2::AutoConfig::NotFound,
                   TwoK2::AutoConfig::LoadFailure
                   ))
{
//  int version;

  // Initialize compConfig with nil in case we fail to create one.
  TwoK::Configuration::ComponentConfigurator_var compConfig = 
    TwoK::Configuration::ComponentConfigurator::_nil ();

  // IMPORTANT NOTE: The following temporary solution assumes that the
  // CosNaming::Name has two components (categoryName and implementation name). 
  const char *impName = (const char *) prereqSpecification[1];
  
  // If we received a factory as a parameter, we use it to create a ComponentConfigurator.
  if (!CORBA::is_nil (factory))
    {
      compConfig = factory->newConfigurator (impName, CORBA::Object::_nil (), ACE_TRY_ENV);
      ACE_CHECK_RETURN (TwoK::Configuration::ComponentConfigurator::_nil ());
    }
  
  /////////////////////////////////////////////////////////////////////////////      
  // Call the prerequisite parser to take care of the prerequisites.
  /////////////////////////////////////////////////////////////////////////////      
  TwoK2::CompSpec compSpec;
  compSpec.comp_name (prereqSpecification);
  
  ACE_TRY_EX(LABEL2) {
    parser_->parse (compSpec, compConfig.in (), ACE_TRY_ENV);
    ACE_TRY_CHECK_EX(LABEL2);
  }
  ACE_CATCHANY
    {
      ACE_DEBUG ((LM_ERROR, "(%P|%t) Could not parse prerequisites for %s.\n", impName));
      ACE_THROW_RETURN (TwoK2::AutoConfig::LoadFailure (), compConfig._retn ());
    }
  ACE_ENDTRY;
  
  // Ufs, we got here finally. Everything went well, let's return the component configurator.

  return compConfig._retn ();

}// load_prereq ()


////////////////////////////////////////////////////////////////////////////////
// Loads the implementation of a new Prerequisite Parser and makes it the 
// default one.
////////////////////////////////////////////////////////////////////////////////
void 
AutoConfig_i::setParser (const char * impName,
                         const char * params,
                         CORBA::Environment &ACE_TRY_ENV)
    ACE_THROW_SPEC ((
      CORBA::SystemException,
      TwoK2::AutoConfig::NotFound,
      TwoK2::AutoConfig::LoadFailure
    ))
{
  // load new parser

  // delete old parser

  // assign new parser as default

}// setParser ()

  
////////////////////////////////////////////////////////////////////////////////
// Loads the implementation of a new Prerequisite Resolver and make it the 
// default one.
////////////////////////////////////////////////////////////////////////////////
void 
AutoConfig_i::setResolver (const char * impName,
                           const char * params,
                           CORBA::Environment &ACE_TRY_ENV)
    ACE_THROW_SPEC ((
      CORBA::SystemException,
      TwoK2::AutoConfig::NotFound,
      TwoK2::AutoConfig::LoadFailure
    ))
{
  // load new resolver

  // delete old resolver

  // assign new resolver as default

}// setResolver ()

  
DynamicConfigurator_ptr 
AutoConfig_i::getRepository (CORBA::Environment &ACE_TRY_ENV)
    ACE_THROW_SPEC ((
      CORBA::SystemException,
      TwoK2::AutoConfig::NotFound
    ))
{
  return DynamicConfigurator::_duplicate (dynamicConfigurator_);
}// getRepository ()
