/******************************************************************************
*                                                                             *
*  File: SPDFParser.cpp                                                       *
*                                                                             *
*  Description: A prerequisite parser for the Simple Prerequisite Description *
*               Format (SPDF).                                                *
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************
   Revision Control Information
   $Revision: 1.5 $
   $Log: SPDFParser.cpp,v $
   Revision 1.5  2000/10/09 20:26:15  f-kon
   New namespace for CORBA modules combined with Tomo's changes to support Windows.

   Revision 1.4  2000/09/30 18:25:41  yamane
   Patches for Windows

   Revision 1.3  2000/01/24 15:56:41  f-kon
   minor improvements

   Revision 1.2  2000/01/18 22:08:42  f-kon
   various optimizations and bug fixes

   Revision 1.1.1.4  1999/10/17 18:33:11  yamane
   Cleanup for native exception and linux support

   Revision 1.1.1.3  1999/10/15 22:10:34  yamane
   Clean up for native exception support.

   Revision 1.1.1.2  1999/10/15 21:58:27  yamane
   Clean up for native exception support

   Revision 1.1.1.1  1999/10/12 00:42:46  f-kon
   Version compatible with TAO 1.0

   Revision 1.3  1999/08/11 13:57:33  f-kon
   improved version with most bugs fixed

   Revision 1.2  1999/08/10 15:49:08  f-kon
   little bug fixes

   Revision 1.1  1999/08/03 22:47:49  f-kon
   Initial version of the Automatic Configuration Systems. It is not yet fully debugged.

******************************************************************************/

#include "SPDFParser.h"

const char *const SPDFParser::HARDWARE_REQUIREMENTS_HEADER = ":hardware requirements";
const char *const SPDFParser::SOFTWARE_REQUIREMENTS_HEADER = ":software requirements";
const char *const SPDFParser::OPTIONAL_REQUIREMENTS = "(optional)";


void SPDFParser::init (CORBA::Object_ptr repository, PrereqResolver *resolver,
                       CORBA::Environment& ACE_TRY_ENV)
  ACE_THROW_SPEC((NotFound, InitializationFailure))
{   
  repository_ = DynamicConfigurator::_narrow (repository);
  PrereqParser::init (repository_.in (), resolver, ACE_TRY_ENV);

  if (CORBA::is_nil (repository_.in()))
      {
        ACE_THROW(InitializationFailure("Unrecognized implementation repository"));
      }
}// init ()


////////////////////////////////////////////////////////////////////////////////
// Parses a SPDF file and issue calls to the resolver to resolve specified
// dependencies.
// Throws NotFound if the SPDF file is not found and ParseError if it cannot
// parse the file.
// The variable client referes to the client component whose
// prerequisites are being parsed.
////////////////////////////////////////////////////////////////////////////////
void SPDFParser::parse (PrereqSpec prereqs, TwoK::Configuration::ComponentConfigurator_ptr client,
                        CORBA::Environment& ACE_TRY_ENV) 
  ACE_THROW_SPEC((NotFound, ParseError))
{
  // First get the proper file name by adding the SUFFIX (.prq)
  const char *categoryName = (const char *) prereqs.comp_name()[0];
  char *SPDFFilename = CORBA::string_alloc (MAXNAMELEN);
  
  ACE_OS::sprintf(SPDFFilename, "%s%s", 
                  (const char *) prereqs.comp_name()[1], 
                  PrereqParser::PREREQ_SUFFIX);
  
  DynamicConfigurator::implCode_var prereq_spec;
  
  ACE_TRY
    {
      repository_->download_implementation(categoryName, SPDFFilename, 
                                           prereq_spec,  ACE_TRY_ENV);
      ACE_TRY_CHECK;
    }
  ACE_CATCHANY
    {
      // This is not necessarily an error, it may simply mean that this component doesn't
      // have any prerequisite. In this case, we're done.
      ACE_DEBUG ((LM_DEBUG, "(%P|%t) Warning: Component %s has no prerequisite specification\n",
                  (const char *) prereqs.comp_name()[1]));
      CORBA::string_free (SPDFFilename);
      ACE_THROW (NotFound ("Could not locate prerequisite specification"));
//      ACE_THROW (AutoConfig::NotFound ());
    }
  ACE_ENDTRY;
  
  CORBA::string_free (SPDFFilename);
  char *start = (char *) prereq_spec->get_buffer();
  long length = prereq_spec->length();      
  
  #define DEBUG
  #ifdef DEBUG
  if (length > 0) start[length-1] = '\0';
  #endif
  ACE_DEBUG ((LM_DEBUG, "length = %d:\n%s\n\n",length, start));
  
  this->parse (start, length, client, ACE_TRY_ENV);

}// parse (PrereqSpec)


////////////////////////////////////////////////////////////////////////////////
// Parses a buffer in SPDF format and issue calls to the resolver to resolve
// the specified dependencies.
// Throws ParseError if it cannot parse the file.
// The variable client refers to the client component whose
// prerequisites are being parsed.
////////////////////////////////////////////////////////////////////////////////
void SPDFParser::parse (char *start, long length,
                        TwoK::Configuration::ComponentConfigurator_ptr client,
                        CORBA::Environment& ACE_TRY_ENV) 
  ACE_THROW_SPEC((ParseError))
{
  char *current,   // current position in the buffer to parse. 
       *nextLine,  // start of the next line to parse.
//       *nextToken, // start of the next token to parse.
       *end;       // first character beyond the buffer to parse.

  current  = start;
  end      = start + length;
  nextLine = 0;
  
  // First locate the start of the hardware requirements section
  current = locateLine (current, end, HARDWARE_REQUIREMENTS_HEADER, &nextLine);

  if (current >= end)
    ACE_THROW(ParseError ("Did not locate hardware requirements section in prerequisite specification"));

  // Here we should parse the hardware requirements section.
  // By now, we don't do anything with it and go to the next section.

  // Now locate the start of the software requirements section
  current = locateLine (nextLine, end, SOFTWARE_REQUIREMENTS_HEADER, &nextLine);
  
  if (current >= end)
    ACE_THROW(ParseError ("Did not locate software requirements section in prerequisite specification"));

  // Now, scan each line at once until the end of the buffer
  for (current = ACE_OS::strtok_r(NULL, "\n\r", &nextLine);
       current < end && current != NULL;
       current = ACE_OS::strtok_r(NULL, "\n\r", &nextLine))
    {
      // If the line starts with "#", "%" or ";" it's a comments line.
      if (*current == '#' || *current == '%' || *current == ';')
        continue; // go to the next line
        
      // get hookName, spec, and optional
      char *hookName, *spec, *last = 0;
      int optional = 0;
      
      hookName = ACE_OS::strtok_r(current, " \t", &last);
      spec     = ACE_OS::strtok_r(NULL, " \t", &last);
      last     = ACE_OS::strtok_r(NULL, " \t", &last);
      if (last && !ACE_OS::strcasecmp(last, OPTIONAL_REQUIREMENTS))
        optional = 1;
      
      // now, divide spec in categoryName and impName
      char *categoryName, *impName;
      categoryName = ACE_OS::strtok_r(spec, " (", &last);
      for (impName = categoryName + ACE_OS::strlen (categoryName);
           impName > categoryName && *impName != '/' && *impName != '\\';
           impName--)
        ;
      if (*impName == '/' || *impName == '\\')
        {
          *impName = '\0';
          impName++;
        }

//       categoryName = ACE_OS::strtok_r(spec, "/\\", &last);
//       impName      = ACE_OS::strtok_r(NULL, " (", &last);
            
      TwoK2::CompSpec compSpec;
      TwoK2::Name specName;
      specName.length (2);
      specName[0] = CORBA::string_dup (categoryName);
      specName[1] = CORBA::string_dup (impName);
      compSpec.comp_name (specName);

      ACE_TRY
        {
          TwoK::Configuration::ComponentConfigurator_var ignore = 
            resolver_->resolvePrereq (client, hookName, compSpec, ACE_TRY_ENV);
          ACE_TRY_CHECK;
        }
      //      catch (BaseException e)
      ACE_CATCHANY
        {
          if (!optional)
            { 
              // It's a mandatory prerequisite and we failed to resolve it!
              char msg[256];
              // unfortunately, ACE_OS doesn't have snprintf so we have to use
              // this unsafe version.
              ACE_OS::sprintf(msg,
                              "Could not resolve mandatory prerequisite, hook = <%s>, "
                              "spec = <%s>",
                              hookName, spec);
              ACE_THROW(ParseError (msg));
            }
        }
      ACE_ENDTRY;
      
    }// for

}// parse (buf, length)


////////////////////////////////////////////////////////////////////////////////
// Returns a pointer to the beginning of the first line in the buffer that
// starts with the given prefix. 
// If no such a line is found, it returns a pointer to a location >= end.
// nextLine is an output parameter that holds a pointer to the beginning of the 
// line following the line with the prefix.
////////////////////////////////////////////////////////////////////////////////
char *
SPDFParser::locateLine (char *start, char *end, const char *prefix, char **nextLine)
{
  ACE_OS::strtok_r(start, "\n\r", nextLine);
  
  start += ACE_OS::strspn (start, "\n\r");

  while (start < end && start != NULL)
    {
      // Note that there's a more efficient way to do this. But we probably don't need it.
      if (!ACE_OS::strncasecmp(start, prefix, ACE_OS::strlen(prefix)))
        break;
      
      // Advance to the next line.
      start = ACE_OS::strtok_r(NULL, "\n\r", nextLine);
    }
  
  return start;
  
}// locateLine ()
