/******************************************************************************
*                                                                             *
*  File: SimpleResolver.cpp                                                   *
*                                                                             *
*  Description: A simple prerequisite resolver (with no caching).             *
*               Resolves prerequisites by loading components from a given     *
*               2K implementation repository accessible through a             *
*               DynamicConfigurator CORBA object.                             *
*               Explicit dependencies between the components are constructed. *
*                                                                             *
*  Note: The components are loaded in the address space of the                *
*        DynamicConfigurator CORBA object (which may be remote.               *
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************
   Revision Control Information
   $Revision: 1.6 $
   $Log: SimpleResolver.cpp,v $
   Revision 1.6  2000/10/09 20:26:12  f-kon
   New namespace for CORBA modules combined with Tomo's changes to support Windows.

   Revision 1.5  2000/09/30 18:26:55  yamane
   Patches for Windows

   Revision 1.4  2000/01/27 17:41:41  f-kon
   several small improvements

   Revision 1.3  2000/01/24 15:56:42  f-kon
   minor improvements

   Revision 1.2  2000/01/18 22:08:43  f-kon
   various optimizations and bug fixes

   Revision 1.1.1.3  1999/10/15 22:10:37  yamane
   Clean up for native exception support.

   Revision 1.1.1.2  1999/10/15 21:58:46  yamane
   Clean up for native exception support

   Revision 1.1.1.1  1999/10/12 00:42:46  f-kon
   Version compatible with TAO 1.0

   Revision 1.2  1999/08/11 13:57:35  f-kon
   improved version with most bugs fixed

   Revision 1.1  1999/08/03 22:47:49  f-kon
   Initial version of the Automatic Configuration Systems. It is not yet fully debugged.

******************************************************************************/

#include "SimpleResolver.h"

//#define ACE_NDEBUG 1 // comment this out if you want the debugging messages

#define MEASURING_PERFORMANCE
#ifdef MEASURING_PERFORMANCE
#include "tao/Timeprobe.h"
#endif

void 
SimpleResolver::init (TwoK2::AutoConfig_ptr autoConfig,
                      DynamicConfigurator_ptr componentRep, 
                      DynamicConfigurator_ptr localDynConf,
                      TwoK::Configuration::CCFactory_ptr configFactory,
                      CORBA::Environment& ACE_TRY_ENV)
  ACE_THROW_SPEC((InitializationFailure))
{
  autoConfig_    = autoConfig;
  remoteDynConf_ = DynamicConfigurator::_duplicate (componentRep);
  localDynConf_  = DynamicConfigurator::_duplicate (localDynConf);

  if (CORBA::is_nil(autoConfig_.in ()))
    {
      ACE_THROW( InitializationFailure("Bad Automatic Configurator"));
    }
  if (CORBA::is_nil(remoteDynConf_.in ()))
    {
      ACE_THROW(InitializationFailure("Bad implementation repository"));
    }
  if (CORBA::is_nil(localDynConf_.in ()))
    {
      ACE_THROW(InitializationFailure("Bad local Dynamic Configurator"));
    }

  if(!CORBA::is_nil(configFactory))
    configFactory_ = TwoK::Configuration::CCFactory::_duplicate (configFactory);
  else
    configFactory_ = TwoK::Configuration::CCFactory::_nil ();

  ACE_Log_Msg::disable_debug_messages ();

}// init ()

////////////////////////////////////////////////////////////////////////////////
// Some "syntatic sugar" before calling the next version of resolvePrereq.
////////////////////////////////////////////////////////////////////////////////
TwoK::Configuration::ComponentConfigurator_ptr
SimpleResolver::resolvePrereq (TwoK::Configuration::ComponentConfigurator_ptr client,
                               const char *hookName,
                               PrereqSpec &compSpec,
                               CORBA::Environment &ACE_TRY_ENV)
  ACE_THROW_SPEC((InitializationFailure, NotFound, Unsupported, LoadFailure))
{
  if (CORBA::is_nil (localDynConf_.in ()) || CORBA::is_nil (remoteDynConf_.in ()))
    {
      ACE_THROW_RETURN (InitializationFailure ("Local or remote Implementation Repository"
                                               "not initialized"), 
                        TwoK::Configuration::ComponentConfigurator::_nil ());
    }
  
  switch (compSpec._d ()) {
  case  TwoK2::CompName:
    {
      return resolvePrereq (client, hookName, compSpec.comp_name (), ACE_TRY_ENV);
    }
  default:
    ACE_THROW_RETURN (Unsupported ("SimpleResolver::resolvePrereq: "
                                   "Unsupported component spec type"), 
                      TwoK::Configuration::ComponentConfigurator::_nil ());
  }
  
}// resolvePrereq (PrereqSpec &)


////////////////////////////////////////////////////////////////////////////////
// Resolves a prerequite for a client component by fetching the implementation
// of a new component from the 2K implementation repository. It loads the new 
// component, attaches it to the given hook in the client component and 
// registers the client with the new component. 
// Throws a NotFound exception if it cannot locate or load the component.
////////////////////////////////////////////////////////////////////////////////
TwoK::Configuration::ComponentConfigurator_ptr
SimpleResolver::resolvePrereq(TwoK::Configuration::ComponentConfigurator_ptr client,
                              const char *hookName,
                              TwoK2::Name compSpec,
                              CORBA::Environment &ACE_TRY_ENV)
  ACE_THROW_SPEC((NotFound, LoadFailure))
{
#ifdef MEASURING_PERFORMANCE
  static ACE_Time_Value now, last_time = ACE_OS::gettimeofday(), time_difference;
  double mtime;
  //  last_time = ACE_OS::gettimeofday();
#endif

  if (compSpec.length () < 2)
    {
      ACE_THROW_RETURN(NotFound("Bad component implementation specification, "
                                "format: <categoryName, impName>"),
		TwoK::Configuration::ComponentConfigurator::_nil ());
    }
  
//  int version;
  int componentFound = 0;
  
  TwoK::Configuration::ComponentConfigurator_var newComponent;
  DynamicConfigurator::implCode_var binCode;
  char localCatName[MAXNAMELEN];
  char *prereqName = CORBA::string_alloc (MAXNAMELEN);

  const char *categoryName = (const char *) compSpec[0];
  char *impName = CORBA::string_dup (compSpec[1]);
  ACE_OS::sprintf (prereqName, "%s%s", impName, PrereqParser::PREREQ_SUFFIX);

  // We'll save the code in the local repository using the local dynamicConfigurator
  // but append a fixed prefix to the category name.
  ACE_OS::strcpy (localCatName, "downloaded_");
  ACE_OS::strcat (localCatName, categoryName);
  
  ACE_TRY
    {
#ifdef MEASURING_PERFORMANCE
      now = ACE_OS::gettimeofday();
      time_difference = now - last_time;
      mtime = (double) time_difference.msec();
      ACE_OS::printf("entering: unknown delay = %f ms\n", mtime);
      last_time = now;
#endif

      // Fetch the implementation code from the given (remote) repository
      // note that download_implementation will fix impName if necessary
      remoteDynConf_->download_implementation (categoryName, impName, binCode, ACE_TRY_ENV);

#ifdef MEASURING_PERFORMANCE
      now = ACE_OS::gettimeofday();
      time_difference = now - last_time;
      mtime = (double) time_difference.msec();
      ACE_OS::printf("time for fetching the code = %f ms\n", mtime);
      last_time = now;
#endif
      ACE_TRY_CHECK;


      // The DynamicConfigurator should create the category if it doesn't exist.
      localDynConf_->upload_implementation (localCatName, impName, binCode, ACE_TRY_ENV);
      ACE_TRY_CHECK;
#ifdef MEASURING_PERFORMANCE
      now = ACE_OS::gettimeofday();
      time_difference = now - last_time;
      mtime = (double) time_difference.msec();
      ACE_OS::printf("time for saving code to file = %f ms\n", mtime);
      last_time = now;
#endif

      componentFound = 1;

    }
  ACE_CATCHANY
    {
      // We couldn't load the component but we may be able to process its prerequisites.
      ACE_ERROR ((LM_ERROR, "(%P|%t) Warning: could not fetch component %s/%s.\n"
                  "(%P|%t) Will try to locate prerequisites anyway\n", categoryName, impName));
    }
  ACE_ENDTRY;
  
  ACE_TRY_EX (second_block)
    {
#ifdef MEASURING_PERFORMANCE
          now = ACE_OS::gettimeofday();
          time_difference = now - last_time;
          mtime = (double) time_difference.msec();
          ACE_OS::printf("unknwon delay = %f ms\n", mtime);
          last_time = now;
#endif
      // Now, do the same for the file with the Component Prerequisites.
      remoteDynConf_->download_implementation (categoryName, prereqName, binCode, ACE_TRY_ENV);

#ifdef MEASURING_PERFORMANCE
      now = ACE_OS::gettimeofday();
      time_difference = now - last_time;
      mtime = (double) time_difference.msec();
      ACE_OS::printf("time for fetching prerequisites = %f ms\n", mtime);
      last_time = now;
#endif

      ACE_TRY_CHECK_EX(second_block);
      
      localDynConf_->upload_implementation (localCatName, prereqName, binCode, ACE_TRY_ENV);
      ACE_TRY_CHECK_EX(second_block);

#ifdef MEASURING_PERFORMANCE
      now = ACE_OS::gettimeofday();
      time_difference = now - last_time;
      mtime = (double) time_difference.msec();
      ACE_OS::printf("time for saving prerequisites to file = %f ms\n", mtime);
      last_time = now;
#endif

    }
  ACE_CATCHANY
    {
      ACE_DEBUG ((LM_DEBUG, "(%P|%t) Warning: Prerequisite file %s:%s not found.\n", 
                  categoryName, prereqName));
      // If we didn't find either the component implementation or the
      // prerequisite file, there's definitely something wrong, let's throw an exception.
      if (!componentFound)
        {
          CORBA::string_free (prereqName);
          // ACE_ANY_EXCEPTION._tao_print_exception 
          //  ("Could not fetch both the component and the prerequisite specification, reason:");
          ACE_THROW_RETURN (NotFound("Component could not be properly loaded"),
                            TwoK::Configuration::ComponentConfigurator::_nil ());
        }
    }
  ACE_ENDTRY;

  ACE_TRY_EX (third_block)
    {
      CORBA::string_free (prereqName);
      
      if (componentFound)
        {
#ifdef MEASURING_PERFORMANCE
          now = ACE_OS::gettimeofday();
          time_difference = now - last_time;
          mtime = (double) time_difference.msec();
          ACE_OS::printf("unknwon delay = %f ms\n", mtime);
          last_time = now;
#endif
          // Since all the necessary files are in the local DynamicConfigurator, we are
          // ready to load the requested component.
          newComponent = autoConfig_->load_component_aux (compSpec, configFactory_, ACE_TRY_ENV);

#ifdef MEASURING_PERFORMANCE
          now = ACE_OS::gettimeofday();
          time_difference = now - last_time;
          mtime = (double) time_difference.msec();
          ACE_OS::printf("time for loading component = %f ms\n", mtime);
          last_time = now;
#endif

          ACE_TRY_CHECK_EX (third_block);

        }
      else
        {
          // This is a prerequisite file without a component associated to it, let's resolve the
          // prerequisites without trying to load the component
          newComponent = autoConfig_->load_prereq (compSpec, configFactory_, ACE_TRY_ENV);
          ACE_TRY_CHECK_EX (third_block);
#ifdef MEASURING_PERFORMANCE
      now = ACE_OS::gettimeofday();
      time_difference = now - last_time;
      mtime = (double) time_difference.msec();
      ACE_OS::printf("time for resolving prereqs = %f ms\n", mtime);
      last_time = now;
#endif
        }
      
      // If necessary, we establish the explicit dependencies between the
      // client and its new hooked component.
      if (!CORBA::is_nil (client) && !CORBA::is_nil(newComponent.in ()))
        {
          client->addHook (hookName, ACE_TRY_ENV); ACE_TRY_ENV.clear ();
          client->hook (hookName, newComponent, ACE_TRY_ENV);
          ACE_TRY_CHECK_EX (third_block);
          newComponent->registerClient (client, hookName, ACE_TRY_ENV);
          ACE_TRY_CHECK_EX (third_block);

#ifdef MEASURING_PERFORMANCE
          now = ACE_OS::gettimeofday();
          time_difference = now - last_time;
          mtime = (double) time_difference.msec();
          ACE_OS::printf("time for reifying dependencies = %f ms\n", mtime);
          last_time = now;
#endif
        }
    }
  ACE_CATCHANY
    {
      CORBA::string_free (prereqName);
      ACE_ANY_EXCEPTION._tao_print_exception ("Could not resolve prerequisite, reason:");
      ACE_THROW_RETURN (LoadFailure("Component could not be properly loaded"),
                        TwoK::Configuration::ComponentConfigurator::_nil ());
    }
  ACE_ENDTRY;

  return newComponent._retn ();
}// resolvePrereq (Name)
 
