/******************************************************************************
*                                                                             *
*  File: ComponentConfigurator_i.cpp                                          *
*                                                                             *
*  Description: CORBA implementation of the ComponentConfigurator             *
*                                                                             *
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************

   Revision Control Information

   $Revision: 1.5 $

   $Log: Configuration_i.cpp,v $
   Revision 1.5  2000/10/09 21:02:36  f-kon
   Changes to use new TwoK:: namespace for 2K services integrated with Tomo's
   changes for Windows compilation.

   Revision 1.4  2000/02/16 17:08:45  f-kon
   *** empty log message ***

   Revision 1.3  2000/02/02 21:04:03  f-kon
   minor updates

   Revision 1.2  1999/12/17 20:46:20  f-kon
   *** empty log message ***

   Revision 1.1.1.3  1999/10/17 18:33:44  yamane
   Cleanup for native exception and linux support

   Revision 1.1.1.2  1999/10/15 21:21:55  yamane
   Clean up for native exception support

   Revision 1.1.1.1  1999/10/12 00:42:46  f-kon
   Version compatible with TAO 1.0

   Revision 1.1.1.1  1999/04/01 18:16:36  f-kon
   This new 2K source tree contains the most up-to-date version of ACE and TAO
   and including the dynamicTAO code as well as some 2K tests and preliminary
   versions of some of its high-level systems.


   Revision 1.1  1999/02/15 18:03:17  f-kon
   *** empty log message ***

******************************************************************************/

#include "tao/corba.h"
#include "Configuration_i.h"

// Constructor for initially empty configurator.
ComponentConfigurator_i::ComponentConfigurator_i (PortableServer::POA_ptr poa)
{
    ComponentConfigurator_i (poa, 
			     CORBA::string_dup("<no name>"), 
                             CORBA::string_dup("<no info>"), 
                             CORBA_Object::_nil());
}


// Constructor.
// Note that I'm assuming that this constructor gets the ownership of the
// parameters it receives.
ComponentConfigurator_i::ComponentConfigurator_i (PortableServer::POA_ptr poa,
						  char *name, 
                                                  char *info,
                                                  CORBA::Object_ptr implementation)
    : poa_ (PortableServer::POA::_duplicate (poa)), 
      name_ (name), info_ (info), implementation_ (implementation)
{
  // Set the initial hookList_ maximum size to 10 and length to 0.
  TwoK::Configuration::DependencySpecification *buf = 
    TwoK::Configuration::DependencyList::allocbuf(10);
  hookList_.replace (10, 0, buf, 1);

  // The same thing with the clientList_ now.
  buf = TwoK::Configuration::DependencyList::allocbuf(10);
  clientList_.replace (10, 0, buf, 1);
  if (name_)
    ACE_DEBUG ((LM_DEBUG, "Constructed ComponentConfigurator <%s>\n", name_));
}// constructor


ComponentConfigurator_i::~ComponentConfigurator_i (void)
{
    // Note that for a real cleanup, the method destroyComponentConfigurator
    // must be called instead.
    //delete this->lock_;
}// destructor



void ComponentConfigurator_i::destroy (CORBA::Environment& ACE_TRY_ENV)
    ACE_THROW_SPEC ((CORBA::SystemException))
{
    ACE_DEBUG ((LM_DEBUG, "Destroying ComponentConfigurator <%s>\n", name_));

    // where is this defined?
    //    ACE_GUARD_THROW (ACE_Lock,
    //                     ace_mon,
    //                     *this->lock_,
    //                     CORBA::INTERNAL (CORBA::COMPLETED_NO));
    
    if(implementation_)
        {
            CORBA::release (implementation_);
            implementation_ = TwoK::Configuration::ComponentConfigurator::_nil ();
        }
    
    if (name_)
        CORBA::string_free (name_);   
    if (info_)
        CORBA::string_free (info_);   
    
    info_ = name_ = (char *)"<destroyed component>";

    // Do we need to cleanup clientList_ and hookList_ or it is done automatically??
    // I think this is done automatically, but I'm not sure. Fabio.

    // Now, tell the POA to deactivate this CORBA object.
    TAO_TRY
        {
	  PortableServer::ObjectId_var id =  poa_->servant_to_id (this, TAO_TRY_ENV);
	  TAO_CHECK_ENV;
	  
	  poa_->deactivate_object (id.in (), TAO_TRY_ENV);
	  TAO_CHECK_ENV;
	}
    TAO_CATCHANY
      {
	TAO_RETHROW;
      }
    TAO_ENDTRY;
    
    // We *must* be dynamically allocated for this delete to work right...
    delete this;

}// destroyComponentConfigurator()


void 
ComponentConfigurator_i::addHook (const char * hookName, CORBA::Environment& ACE_TRY_ENV)
    ACE_THROW_SPEC ((
        CORBA::SystemException,
        TwoK::Configuration::ComponentConfigurator::ElementExists
    ))
{
  if (hookName == 0 || ACE_OS::strlen (hookName) == 0) // must have a valid hookName
    {
      ACE_THROW(TwoK::Configuration::ComponentConfigurator::invalidArgument ());
    }

  // First check if the hook already exists
  if (findHook (hookName) != 0)
    {
      // The hook does exist, raise exception.
      ACE_THROW(TwoK::Configuration::ComponentConfigurator::ElementExists ());
    }
  
  // Then, check if the sequence is big enough. If it is not, make it bigger.

  CORBA::ULong size = hookList_.length ();
  if (size >= hookList_.maximum ())
    {
      // Allocate space for 10 more items. This will increase length_ and maximum_ by 10.
      hookList_.length (size + 10);
      // Note that we'll reset length to the proper value bellow. This is a
      // weird solution but it's not my fault, it seems it's the best we can
      // do with the way CORBA specifies the C++ sequence mapping. If we
      // have problems with that, what we should do is construct a new
      // buffer and replace for the existing one.
    }
  
  hookList_[size].name = CORBA::string_dup (hookName);
  hookList_[size].componentConfig = TwoK::Configuration::ComponentConfigurator::_nil ();
  hookList_.length (size + 1); // This sets length_ to size+1 and does not change maximum_.
  
}//addHook ()


void 
ComponentConfigurator_i::deleteHook (const char * hookName, CORBA::Environment& ACE_TRY_ENV) 
    ACE_THROW_SPEC ((
        CORBA::SystemException,
        TwoK::Configuration::ComponentConfigurator::NotFound
    ))
{
 if (hookName == 0)
      {
          // The hook does not exist, raise exception.
          ACE_THROW(TwoK::Configuration::ComponentConfigurator::NotFound ());
      }

 // Look for the hook in the hook list.
 CORBA::ULong pos, i;
 CORBA::ULong size = hookList_.length();
 
 for (pos = 0; pos < size; pos++)
     if (ACE_OS::strcmp(hookName, hookList_[pos].name ) == 0)
         break;
 
 if (pos < size)
   {
     
     for (i = pos; i < size-1; i++)
       hookList_[i] = hookList_[i+1];
     hookList_.length(size-1);
   }
 else      // The hook does not exist, raise exception.
   {
     ACE_THROW(TwoK::Configuration::ComponentConfigurator::NotFound ());
   } 
}// deleteHook ()


void 
ComponentConfigurator_i::hook (const char * hookName, TwoK::Configuration::ComponentConfigurator_ptr cc, 
                               CORBA::Environment& ACE_TRY_ENV)
    ACE_THROW_SPEC ((
        CORBA::SystemException,
        TwoK::Configuration::ComponentConfigurator::NotFound
    ))
{
  TwoK::Configuration::DependencySpecification *hs = findHook (hookName);
 
  if (hs != 0)
    {
      if (!CORBA::is_nil (hs->componentConfig.ptr ()))
        {
          // The hook is not vacant.  Shall we do something about the old
          // hooked component here?  We should try to send an event to the old
          // hooked component and ignore any exception.
        }
      
      hs->componentConfig = TwoK::Configuration::ComponentConfigurator::_duplicate (cc); 
      // I'm afraid that this may delete the old hooked
      // component even if it is hooked to other places and should not be
      // deleted. Test it. I may need to call hs->_retn first.
      // No, probably not since it does proper reference counting. 

      // But I think it may work badly because it seems that the factory is
      // not leaving the reference counting correctly. I added a duplicate in
      // the factory but it seems that it is still problematic. It's still
      // destroying the object even though the client has a reference to
      // it. try to have the client calling this object after it is deleted
      // here to see what happens.
    }
  else // hs == 0
    {
      // The hook does not exist, raise exception.
      ACE_THROW(TwoK::Configuration::ComponentConfigurator::NotFound ());
    }
}// hook ()


void 
ComponentConfigurator_i::unhook (const char * hookName, CORBA::Environment& ACE_TRY_ENV) 
    ACE_THROW_SPEC ((
        CORBA::SystemException,
        TwoK::Configuration::ComponentConfigurator::HookVacant,
        TwoK::Configuration::ComponentConfigurator::NotFound
    ))
{
  TwoK::Configuration::DependencySpecification *hs = findHook (hookName);
 
  if (hs != 0)
    {
      if (!CORBA::is_nil (hs->componentConfig.ptr ()))
        {
          hs->componentConfig = TwoK::Configuration::ComponentConfigurator::_nil ();
          // I'm afraid that this may delete the old hooked
          // component even if it is hooked to other places and should not be
          // deleted. Test it. I may need to call hs->_retn first.
        }
      else // The hook is already vacant.
        {
          ACE_THROW(TwoK::Configuration::ComponentConfigurator::HookVacant ());
        }
    }
  else // hs == 0
    {
      // The hook does not exist, raise exception.
        ACE_THROW(TwoK::Configuration::ComponentConfigurator::NotFound ());
    }
}// unhook ()


void 
ComponentConfigurator_i::registerClient (TwoK::Configuration::ComponentConfigurator_ptr client,
                                         const char * hookNameInClient, CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException,
        TwoK::Configuration::ComponentConfigurator::ElementExists
      ))
{
  if (!findClient (client, hookNameInClient))
    {
      // Add new DependencySpecification to the list of clients.
      // Then, check if the sequence is big enough. If it is not, make it bigger.
      
      CORBA::ULong size = clientList_.length ();
      if (size >= clientList_.maximum ())
        {
          // Allocate space for 10 more items. This will increase length_ and maximum_ by 10.
          clientList_.length (size + 10);
          // Note that we'll reset length to the proper value bellow. This is a
          // weird solution but it's not my fault, it seems it's the best we can
          // do with the way CORBA specifies the C++ sequence mapping. If we
          // have problems with that, what we should do is construct a new
          // buffer and replace for the existing one.
        }
      
      clientList_.length (size + 1); // This sets length_ to size+1 and does not change maximum_.
      clientList_[size].name = CORBA::string_dup (hookNameInClient);
      clientList_[size].componentConfig = TwoK::Configuration::ComponentConfigurator::_duplicate (client);
    }
  else
    {
      // The element is already in the list, raise exception.
      ACE_THROW(TwoK::Configuration::ComponentConfigurator::ElementExists ());
    }
}// registerClient


void 
ComponentConfigurator_i::unregisterClient ( TwoK::Configuration::ComponentConfigurator_ptr client,
                                           const char * hookNameInClient, CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException,
        TwoK::Configuration::ComponentConfigurator::NotFound
      ))
{
  // Look for an entry in the clientList_ matching the given arguments.
  CORBA::ULong pos, i;
  CORBA::ULong size = clientList_.length();
  
  for (pos = 0; pos < size; pos++)
    if (client->_is_equivalent (clientList_[pos].componentConfig.ptr (), ACE_TRY_ENV) &&
        (ACE_OS::strcmp(hookNameInClient, clientList_[pos].name ) == 0))
      break;
  
  if (pos < size) // entry found
    {
      for (i = pos; i < size-1; i++)
        clientList_[i] = clientList_[i+1]; 
      // should this previous line release one reference and add another? It's not doing it.
      clientList_.length(size-1);
    }
  else // The pair <client, hookNameInClient> does not exist, raise exception.
    {
      ACE_THROW(TwoK::Configuration::ComponentConfigurator::NotFound ());
    }
}// unregisterClient ()

////////////////////////////////////////////////////////////////////////////////
// ttl == 0 means always forward it. ttl == n, n>0, means decrement it and
// forward if it is still > 0 after decrementing it.
////////////////////////////////////////////////////////////////////////////////
void 
ComponentConfigurator_i::eventFromHookedComponent 
          (TwoK::Configuration::ComponentConfigurator_ptr hookedComponent,
           const TwoK::Configuration::Event & e, CORBA::UShort timeToLive, CORBA::Environment& ACE_TRY_ENV)
  ACE_THROW_SPEC ((
                   CORBA::SystemException
                   ))
{
  CORBA::ULong i, size;
  CORBA::UShort ttl;
  
  // This should be customized by applications and systems to carry out
  // reconfiguration based on events. This implementation just generates a
  // debugging message for each event and forward it to clients and hooked
  // components.

  ACE_DEBUG ((LM_INFO, 
              "Component <%s> received event '%s' from hooked component <%s>:\n\t%s\n", 
              this->name (ACE_TRY_ENV), eventToString (e.type), 
              hookedComponent->name (ACE_TRY_ENV), (const char *)e.description));

  if (timeToLive != 0)
      {
          ttl = timeToLive - 1;
          if (ttl <= 0)
              return; // do not forward it.
      }
  else
      ttl = 0;
  
  // Forward the event to the clients. Do not forward it to the
  // component that sent the event originally.
  size = clientList_.length();
  for (i = 0; i < size; i++)
      {
          TwoK::Configuration::ComponentConfigurator_var c = 
            TwoK::Configuration::ComponentConfigurator::_duplicate 
             (clientList_[i].componentConfig.in());
          if (c.ptr () != TwoK::Configuration::ComponentConfigurator::_nil () &&
              !c->_is_equivalent (hookedComponent, ACE_TRY_ENV))
              c->eventFromHookedComponent (_this (ACE_TRY_ENV), e, ttl, 
                                         ACE_TRY_ENV);
      }

  // does the use of _this generate memory leaks??? do I need to delete it???

  // Now, forward it to the hooked components. Do not forward it to the
  // component that sent the event originally.
  size = hookList_.length();
  for (i = 0; i < size; i++)
      {
        TwoK::Configuration::ComponentConfigurator_var c = 
          TwoK::Configuration::ComponentConfigurator::_duplicate
            (hookList_[i].componentConfig.in());
        if (c.ptr () != TwoK::Configuration::ComponentConfigurator::_nil () &&
            !c->_is_equivalent (hookedComponent, ACE_TRY_ENV))
          c->eventFromClient (_this (ACE_TRY_ENV), e, ttl, ACE_TRY_ENV);
      }
} // eventFromHookedComponent()

////////////////////////////////////////////////////////////////////////////////
// ttl == 0 means always forward it. ttl == n, n>0, means decrement it and
// forward if it is still > 0 after decrementing it.
////////////////////////////////////////////////////////////////////////////////
void 
ComponentConfigurator_i::eventFromClient (TwoK::Configuration::ComponentConfigurator_ptr client,
                                        const TwoK::Configuration::Event & e,
                                        CORBA::UShort timeToLive, CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
  CORBA::ULong i, size;
  CORBA::UShort ttl;

  // This should be customized by applications and systems to carry out
  // reconfiguration based on events. This implementation just generates a
  // debugging message for each event and forward it to clients and hooked
  // components.
    ACE_DEBUG ((LM_INFO, 
                "Component <%s> received event '%s' from client component <%s>:\n\t%s\n", 
                this->name (ACE_TRY_ENV), eventToString (e.type), 
                client->name (ACE_TRY_ENV), (const char *)e.description));

    if (timeToLive != 0)
        {
            ttl = timeToLive - 1;
            if (ttl <= 0)
                return; // do not forward it.
        }
    else
        ttl = 0;

    // Forward the event to the clients except the one sent the event originally.
    size = clientList_.length();
    for (i = 0; i < size; i++)
        {
          TwoK::Configuration::ComponentConfigurator_var c = 
            TwoK::Configuration::ComponentConfigurator::_duplicate
              (clientList_[i].componentConfig.in());
          if (c.ptr () != TwoK::Configuration::ComponentConfigurator::_nil () &&
              !c->_is_equivalent (client, ACE_TRY_ENV))
            c->eventFromHookedComponent (_this (ACE_TRY_ENV), e, ttl, 
                                       ACE_TRY_ENV);
        }
    // Now, forward it to the hooked components except the one which sent the event originally.
    size = hookList_.length();
    for (i = 0; i < size; i++)
      {
        TwoK::Configuration::ComponentConfigurator_var c = 
          TwoK::Configuration::ComponentConfigurator::_duplicate
            (hookList_[i].componentConfig.in());
        if (c.ptr () != TwoK::Configuration::ComponentConfigurator::_nil () &&
            !c->_is_equivalent (client, ACE_TRY_ENV))
          c->eventFromClient (_this (ACE_TRY_ENV), e, ttl, ACE_TRY_ENV);
      }
} // eventFromClient ()




////////////////////////////////////////////////////////////////////////////////
// ttl == 0 means always forward it. ttl == n, n>0, means decrement it and
// forward if it is still > 0 after decrementing it.
// (one way versio)
////////////////////////////////////////////////////////////////////////////////
void 
ComponentConfigurator_i::eventFromHookedComponent_ow 
          (TwoK::Configuration::ComponentConfigurator_ptr hookedComponent,
           const TwoK::Configuration::Event & e, CORBA::UShort timeToLive, CORBA::Environment& ACE_TRY_ENV)
  ACE_THROW_SPEC ((
                   CORBA::SystemException
                   ))
{
  CORBA::ULong i, size;
  CORBA::UShort ttl;
  
  // This should be customized by applications and systems to carry out
  // reconfiguration based on events. This implementation just generates a
  // debugging message for each event and forward it to clients and hooked
  // components.

  ACE_DEBUG ((LM_INFO, 
              "Component <%s> received event '%s' from hooked component <%s>:\n\t%s\n", 
              this->name (ACE_TRY_ENV), eventToString (e.type), 
              hookedComponent->name (ACE_TRY_ENV), (const char *)e.description));

  if (timeToLive != 0)
      {
          ttl = timeToLive - 1;
          if (ttl <= 0)
              return; // do not forward it.
      }
  else
      ttl = 0;
  
  // Forward the event to the clients. Do not forward it to the
  // component that sent the event originally.
  size = clientList_.length();
  for (i = 0; i < size; i++)
      {
          TwoK::Configuration::ComponentConfigurator_var c = 
            TwoK::Configuration::ComponentConfigurator::_duplicate 
             (clientList_[i].componentConfig.in());
          if (c.ptr () != TwoK::Configuration::ComponentConfigurator::_nil () &&
              !c->_is_equivalent (hookedComponent, ACE_TRY_ENV))
              c->eventFromHookedComponent_ow (_this (ACE_TRY_ENV), e, ttl, 
                                            ACE_TRY_ENV);
      }

  // does the use of _this generate memory leaks??? do I need to delete it???

  // Now, forward it to the hooked components. Do not forward it to the
  // component that sent the event originally.
  size = hookList_.length();
  for (i = 0; i < size; i++)
      {
        TwoK::Configuration::ComponentConfigurator_var c = 
          TwoK::Configuration::ComponentConfigurator::_duplicate
            (hookList_[i].componentConfig.in());
        if (c.ptr () != TwoK::Configuration::ComponentConfigurator::_nil () &&
            !c->_is_equivalent (hookedComponent, ACE_TRY_ENV))
          c->eventFromClient_ow (_this (ACE_TRY_ENV), e, ttl, ACE_TRY_ENV);
      }
} // eventFromHookedComponent_ow()



////////////////////////////////////////////////////////////////////////////////
// ttl == 0 means always forward it. ttl == n, n>0, means decrement it and
// forward if it is still > 0 after decrementing it.
// (one way version)
////////////////////////////////////////////////////////////////////////////////
void 
ComponentConfigurator_i::eventFromClient_ow (TwoK::Configuration::ComponentConfigurator_ptr client,
                                           const TwoK::Configuration::Event & e,
                                           CORBA::UShort timeToLive, CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
  CORBA::ULong i, size;
  CORBA::UShort ttl;

  // This should be customized by applications and systems to carry out
  // reconfiguration based on events. This implementation just generates a
  // debugging message for each event and forward it to clients and hooked
  // components.
    ACE_DEBUG ((LM_INFO, 
                "Component <%s> received event '%s' from client component <%s>:\n\t%s\n", 
                this->name (ACE_TRY_ENV), eventToString (e.type), 
                client->name (ACE_TRY_ENV), (const char *)e.description));

    if (timeToLive != 0)
        {
            ttl = timeToLive - 1;
            if (ttl <= 0)
                return; // do not forward it.
        }
    else
        ttl = 0;

    // Forward the event to the clients except the one sent the event originally.
    size = clientList_.length();
    for (i = 0; i < size; i++)
        {
          TwoK::Configuration::ComponentConfigurator_var c = 
            TwoK::Configuration::ComponentConfigurator::_duplicate
              (clientList_[i].componentConfig.in());
          if (c.ptr () != TwoK::Configuration::ComponentConfigurator::_nil () &&
              !c->_is_equivalent (client, ACE_TRY_ENV))
            c->eventFromHookedComponent_ow (_this (ACE_TRY_ENV), e, ttl, 
                                          ACE_TRY_ENV);
        }
    // Now, forward it to the hooked components except the one which sent the event originally.
    size = hookList_.length();
    for (i = 0; i < size; i++)
      {
        TwoK::Configuration::ComponentConfigurator_var c = 
          TwoK::Configuration::ComponentConfigurator::_duplicate
            (hookList_[i].componentConfig.in());
        if (c.ptr () != TwoK::Configuration::ComponentConfigurator::_nil () &&
            !c->_is_equivalent (client, ACE_TRY_ENV))
          c->eventFromClient_ow (_this (ACE_TRY_ENV), e, ttl, ACE_TRY_ENV);
      }
} // eventFromClient_ow ()




char * 
ComponentConfigurator_i::name (CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
    ACE_UNUSED_ARG(ACE_TRY_ENV);
    return CORBA::string_dup(name_);
}


char * 
ComponentConfigurator_i::info (CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
    ACE_UNUSED_ARG(ACE_TRY_ENV);
    return CORBA::string_dup(info_);
}

TwoK::Configuration::DependencyList * 
ComponentConfigurator_i::listHooks (CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
  ACE_UNUSED_ARG(ACE_TRY_ENV);
  return new TwoK::Configuration::DependencyList (hookList_); 
}//listHooks ()

    
TwoK::Configuration::ComponentConfigurator_ptr 
ComponentConfigurator_i::getHookedComponent (const char * hookName, CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
  TwoK::Configuration::DependencySpecification *hs = findHook (hookName);
    
  if (hs != 0)
    return TwoK::Configuration::ComponentConfigurator::_duplicate (hs->componentConfig);
    //  return hs->componentConfig; 
  else
    {
      // The hook does not exist, raise exception.
      ACE_THROW_RETURN(TwoK::Configuration::ComponentConfigurator::NotFound (),
		 TwoK::Configuration::ComponentConfigurator::_nil ());
    }
}// getHookedComponent ()


CORBA::UShort 
ComponentConfigurator_i::numberOfClients (CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
    ACE_UNUSED_ARG(ACE_TRY_ENV);
    return clientList_.length ();
}


TwoK::Configuration::DependencyList * 
ComponentConfigurator_i::listClients (CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
  ACE_UNUSED_ARG(ACE_TRY_ENV);

    // can I write the following????
    //  return &clientList_;
    
  return new TwoK::Configuration::DependencyList (clientList_); 

} // listClients ()



CORBA::Object_ptr 
ComponentConfigurator_i::implementation (CORBA::Environment& ACE_TRY_ENV)
      ACE_THROW_SPEC ((
        CORBA::SystemException
      ))
{
    ACE_UNUSED_ARG(ACE_TRY_ENV);
    CORBA_Object::_duplicate (implementation_);
    return implementation_;
}

// Didn't work from one request to the next. Just worked immediately after the
// stub was created.  void ComponentConfigurator_i::self
// (TwoK::Configuration::ComponentConfigurator_ptr POAstub) { self_ = POAstub;
// CORBA::Environment e;
//    
//     ACE_OS::printf("Testing self, my name is %s\n", self_->name(e));
//}


////////////////////////////////////////////////////////////////////////////////
// Private auxiliary function. Given a hook name, returns a pointer to the
// dependency specification associated with that name. Returns 0 if the element
// is not found.
////////////////////////////////////////////////////////////////////////////////
TwoK::Configuration::DependencySpecification *
ComponentConfigurator_i::findHook (const char *hookName)
{
  if (hookName == 0)
      return 0;
  
  for (CORBA::ULong i = 0; i < hookList_.length(); i++)
      {
          if (ACE_OS::strcmp(hookName, hookList_[i].name ) == 0)
              {
                  return &hookList_[i]; // I can use that because this is a private function.
              }
      }
  
  return 0; // element not found.
  
}// findHook ()


////////////////////////////////////////////////////////////////////////////////
// Protected auxiliary function that can be used by our subclasses. 
// Given a pointer to a ComponentConfigurator, returns a pointer to the
// first dependency specification found in the hookList_ that contains that
// ComponentConfirator. Returns 0 if the element is not found.
////////////////////////////////////////////////////////////////////////////////
TwoK::Configuration::DependencySpecification *
ComponentConfigurator_i::findHookedComponent (TwoK::Configuration::ComponentConfigurator_ptr cc)
{
  if (CORBA::is_nil (cc))
    return 0;

  CORBA::Environment env;

  for (CORBA::ULong i = 0; i < hookList_.length(); i++)
      {
        if (cc->_is_equivalent (hookList_[i].componentConfig.ptr (), env))
          {
            return &hookList_[i]; // I can use that because this is not a CORBA function.
          }
      }
  
  return 0; // element not found.
  
}// findHook ()

////////////////////////////////////////////////////////////////////////////////
// Private auxiliary function. Given a ComponentConfigurator_ptr and a hook
// name, returns a pointer to the dependency specification matching that pair.
// Returns 0 if the element is not found.
////////////////////////////////////////////////////////////////////////////////
TwoK::Configuration::DependencySpecification *
ComponentConfigurator_i::findClient (TwoK::Configuration::ComponentConfigurator_ptr client, 
                                     const char *name)
{
  CORBA::Environment env;
  
  CORBA::ULong size = clientList_.length();
  
  for (CORBA::ULong i = 0; i < size; i++)
   if (client->_is_equivalent (clientList_[i].componentConfig.ptr (), env) &&
        (ACE_OS::strcmp(name, clientList_[i].name ) == 0))
      {
        return &clientList_[i]; // I can use that because this is a private function.
      }

  return 0; // element not found.
}// findHook ()

const char *
ComponentConfigurator_i::eventToString (TwoK::Configuration::EventType type)
{
  switch (type)
    {
    case TwoK::Configuration::STARTED:
      return "STARTED";
    case TwoK::Configuration::FINISHED:
      return "FINISHED";
    case TwoK::Configuration::SHUTDOWN:
      return "SHUTDOWN";
    case TwoK::Configuration::RECONFIGURED:
      return "RECONFIGURED";
    case TwoK::Configuration::REPLACED:
      return "REPLACED";
    case TwoK::Configuration::MIGRATED:
      return "MIGRATED";
    case TwoK::Configuration::DELETED:
      return "DELETED";
    case TwoK::Configuration::FAILED:
      return "FAILED";
    case TwoK::Configuration::APP_SPECIFIC:
      return "APP_SPECIFIC";

    default:
      return "UNKNOWN";
    }
}// eventToString ()
