/******************************************************************************
*                                                                             *
*  File: ComponentRepository.cpp                                              *
*                                                                             *
*  Description: A Repository of Components built on top                       *
*               of the DynamicConfigurator.                                   *
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************
   Revision Control Information
   $Revision: 1.6 $
   $Log: ComponentRepository.cpp,v $
   Revision 1.6  2000/09/30 18:31:26  yamane
   Patches for Windows

   Revision 1.3  2000/02/20 08:11:37  yamane
   Supports multiple platforms (OS name/Processor name)

   Revision 1.2  2000/01/27 17:43:23  f-kon
   added registration to the NameService

   Revision 1.1  2000/01/24 15:59:47  f-kon
   first working version

******************************************************************************/

#include "ComponentRepository.h"
#include "ace/CORBA_macros.h"
#include "ace/Get_Opt.h"

int
main (int argc, char**argv)
{
  ComponentRepository CR;

  if (CR.init (argc, argv) < 0)
    return -1;

  ACE_DECLARE_NEW_CORBA_ENV;
  ACE_TRY
    {
      CR.run (ACE_TRY_ENV);
      ACE_TRY_CHECK;
    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION (ACE_ANY_EXCEPTION, 
                           "ComponentRepository main function failed, reason:");
      return -1;
    }
  ACE_ENDTRY;
  ACE_CHECK_RETURN (-1);

  return 0;

}// main ()


ComponentRepository::ComponentRepository ()
  : verbose_(0), ior_output_file_(0), time_(0)
{
}// constructor

int
ComponentRepository::init (int argc, char**argv)
{
  CORBA::ORB_var orb_aux;
  CosNaming::Name query_name (3);

  ACE_DECLARE_NEW_CORBA_ENV;
  ACE_TRY_EX (A)
    {
      if (orbManager_.init (argc,argv, ACE_TRY_ENV) == -1)
        {
          ACE_ERROR_RETURN((LM_ERROR,
                            "%p\n","ComponentRepository::init: orbManager_.init failed"),
                           -1);
        }
      ACE_TRY_CHECK_EX (A);
      orb_aux = orbManager_.orb();

      // Locate our local DynamicConfigurator
      CORBA::Object_var obj = 
        orb_aux->resolve_initial_references ("DynamicConfigurator");
      
      if (CORBA::is_nil (obj.in ()))
        ACE_ERROR_RETURN ((LM_ERROR,
                           "ComponentRepository::init: "
                           "Unable to resolve the DynConf Service.\n"),-1);
      
      localDynConf_ = DynamicConfigurator::_narrow (obj.in (), ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (A);

      // Now, locate the Naming Service
      obj = orb_aux->resolve_initial_references ("NameService");
      
      if (CORBA::is_nil (obj.in ()))
        ACE_ERROR_RETURN ((LM_ERROR, 
                           "(%P|%t) Unable to resolve the Name Service.\n"), -1);
      
      nameServiceRoot_ = CosNaming::NamingContext::_narrow (obj.in (), ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (A); 

      if (CORBA::is_nil (nameServiceRoot_))
        ACE_ERROR_RETURN ((LM_ERROR, 
                           "(%P|%t) Unable to resolve the Name Service.\n"), -1);
      
      // Now, bind our local DynamicConfigurator to the Name Service
      // as the 2K ComponentRepository.

      query_name.length (1);
      query_name[0].id = CORBA::string_dup ("services"); // is this leaking?
      nameServiceRoot_->bind_new_context (query_name, ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (A);
    }
  ACE_CATCH (CosNaming::NamingContext::AlreadyBound, varName)
    {
      // It was already bound. So, that's OK we don't need to do anything.
    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION (ACE_ANY_EXCEPTION, "ComponentRepository::init A");
      return -1;
    }
  ACE_ENDTRY;

  ACE_TRY_EX (B)
    {
      // Now, bind our local DynamicConfigurator to the Name Service
      // as the 2K ComponentRepository.

      query_name.length (2);
      query_name[0].id = CORBA::string_dup ("services"); // is this leaking?
      query_name[1].id = CORBA::string_dup ("ComponentRepository"); // is this leaking?
      nameServiceRoot_->bind_new_context (query_name, ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (B);
    }
  ACE_CATCH (CosNaming::NamingContext::AlreadyBound, varName)
    {
      // It was already bound. So, that's OK we don't need to do anything.
    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION (ACE_ANY_EXCEPTION, "ComponentRepository::init B");
      return -1;
    }
  ACE_ENDTRY;

  if (parse_args (argc, argv) < 0)
    return -1;

  ACE_TRY_EX (C)
    {
      // At this point, we know the context exists, now let's try to bind to it.
      query_name.length (3);
      char arch[256];

      ACE_OS::sprintf(arch, "%s-%s", 
			ACE_2k_os::os_name(), ACE_2k_os::processor_name());

      query_name[2].id = CORBA::string_dup (arch); // is this leaking?

      // Possible Improvement: print a warning if overwriting a previous binding.
      nameServiceRoot_->rebind (query_name, localDynConf_, ACE_TRY_ENV);
      ACE_TRY_CHECK_EX (C); 

      if (verbose_)
        ACE_OS::printf("ComponentRepository IOR : <%s>\n", 
                       orb_aux->object_to_string (localDynConf_));

      if (ior_output_file_)
        {
          ACE_OS::fprintf (ior_output_file_, "%s\n", orb_aux->object_to_string (localDynConf_));
          ACE_OS::fclose (ior_output_file_);
        }

    }
  ACE_CATCHANY
    {
      ACE_PRINT_EXCEPTION (ACE_ANY_EXCEPTION, "ComponentRepository::init C");
      return -1;
    }
  ACE_ENDTRY;

  return 0;
}// init ()



int
ComponentRepository::run (CORBA_Environment& ACE_TRY_ENV)
{
  if (time_ == 0)
      return orbManager_.run (ACE_TRY_ENV);
  else
    {
      ACE_Time_Value *t = new ACE_Time_Value (time_);
      return orbManager_.run (ACE_TRY_ENV, t);
    }

}// run()




int
ComponentRepository::parse_args (int argc, char *argv[])
{
  ACE_Get_Opt get_opts (argc, argv, "do:t:v");
  int c;
  int time;

  while ((c = get_opts ()) != -1)
    switch (c)
      {
      case 'd':  // debug flag.
        TAO_debug_level++;
        break;
      case 'v':  // debug flag.
        verbose_ = 1;
        break;
      case 'o': // outputs the AutoConfig service ior to a file.
        ior_output_file_ =
          ACE_OS::fopen (get_opts.optarg, "w");

        if (ior_output_file_ == 0)
          ACE_ERROR_RETURN ((LM_ERROR,
                             "Unable to open %s for writing: %p\n",
                             get_opts.optarg), -1);
        break;
      case 't':
        time = ACE_OS::atoi (get_opts.optarg);
        if (time >= 0)
          time_ = time;
        break;
      case '?':
      default:
        ACE_ERROR_RETURN ((LM_ERROR,
                           "usage:  %s "
                           "[-o <ior_output_file> ]"
                           "[-t <lifetime_in_seconds>]"
                           "[-v ]"
                           "\n",
                           argv [0]),
                          -1);
      }
  return 0;

}// parse_args ()
