/* 
 * Released Date: March 7 99
 * This is the Dynamic Soft CPU Scheduler II release by the 
 *   Monet Group, 
 *   http://cairo.cs.uiuc.edu/
 *   Department of Computer Science, 
 *   University of Illinois at Urbana Champaign.  
 * 
 * This release has been tested under SUN OS 5.6.
 *
 * Contacting Authors:
 *  Hao-hua Chu (h-chu3@cs.uiuc.edu)
 *  Klara Nahrstedt (klara@cs.uiuc.edu)
 * 
 */

#include <stdio.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>
#include <stdlib.h>
#include "CpuApi.H"
#include "Timeval.H"


CpuApi::CpuApi()
{
  _pid = getpid();
  _wrShmArea = NULL;
  _msgQueue = new MsgQueue(0, _pid);
  _msg.mtype = 1;
  _msg.callerPid = _pid;
  _msg.cmd = CMD_INIT;

  // contact the server
  _msgQueue->send(&_msg);
  _ssSamplingInterval = 0;

  if (!_msgQueue->recv(&_msg, _pid, BLOCKING))
    printf("msgQueue did not receive\n");
  
  // create and attach the readable part of the shared memory with the server.
  _shm = new ShmCli();

  if (!_msg.res || 
      !(_p = _shm->attach(_pid, _msg.idx))) {
    _shm->detach();
    exit(1);
  }

  // create the writable part of the shared memory.
  _wrShmArea = _shm->getWrShmArea();
  _cpsBuf = NULL;
  _ssBuf = NULL;
  _psFname = NULL;
};


CpuApi::~CpuApi()
{
  if (_cpsBuf) {
    if (_psFname)
      _cpsBuf->writeFile(_psFname);
    delete _cpsBuf;
  }
  if (_ssBuf)
    delete _ssBuf;
  if (_msgQueue)
    destroy();
  _wrShmArea = NULL;
};


int CpuApi::destroy(int pid)
{
  _msg.mtype = 1;
  if (pid == 0) {
    _msg.cmd = CMD_DESTROY;
    _msgQueue->send(&_msg);
    delete _msgQueue;
    _msgQueue = NULL;
    _shm->detach();
    return(1);
  }
  else {
    _msg.cmd = CMD_IDESTROY;
    _msg.intArg[0] = pid;
    _msgQueue->send(&_msg);
    return(1);
  }
};


int CpuApi::reserve(CpuReserve* cr, int pid) 
{
  _msg.mtype = 1;
  if (pid == 0)
    _msg.cmd = CMD_RESERVE;
  else {
    _msg.cmd = CMD_IRESERVE;
    _msg.intArg[0] = pid;
  }
  _msg.cr = *cr;
  _msgQueue->send(&_msg);
  if (!_msgQueue->recv(&_msg, _pid, BLOCKING))
    printf("msgQueue did not receive\n");
  return(_msg.res);
};


int CpuApi::modify(CpuReserve* cr, int pid) 
{
  _msg.mtype = 1;
  if (pid == 0)
    _msg.cmd = CMD_MODIFY;
  else {
    _msg.cmd = CMD_IMODIFY;
    _msg.intArg[0] = pid;
  }
  _msg.cr = *cr;
  _msgQueue->send(&_msg);
  if (!_msgQueue->recv(&_msg, _pid, BLOCKING))
    printf("msgQueue did not receive\n");
  return(_msg.res);
};


int CpuApi::setAdaptStrategy(AdaptStrategy* as, int pid)
{
  _msg.mtype = 1;
  if (pid == 0) 
    _msg.cmd = CMD_SET_ADAPT_STRATEGY;
  else {
    _msg.cmd = CMD_ISET_ADAPT_STRATEGY;
    _msg.intArg[0] = pid;
  }
  _msg.as = *as;
  _msgQueue->send(&_msg);
  if (!_msgQueue->recv(&_msg, _pid, BLOCKING))
    printf("msgQueue did not receive\n");
  return(_msg.res);  
};


int CpuApi::free(int pid) 
{
  _msg.mtype = 1;
  if (pid == 0)
    _msg.cmd = CMD_FREE;
  else {
    _msg.cmd = CMD_IFREE;
    _msg.intArg[0] = pid;
  }
  _msgQueue->send(&_msg);
  if (!_msgQueue->recv(&_msg, _pid, BLOCKING))
    printf("msgQueue did not receive\n");
  return(_msg.res);
};


int CpuApi::start(timeval begin, int pid)
{
  _msg.mtype = 1;
  if (pid == 0)
  	_msg.cmd = CMD_START;
	else
		_msg.cmd = CMD_ISTART;

  _msg.intArg[0] = pid;
  _msg.intArg[1] = begin.tv_sec;
  _msg.intArg[2] = begin.tv_usec;

  _msgQueue->send(&_msg);
  if (!_msgQueue->recv(&_msg, _pid, BLOCKING))
    printf("msgQueue did not receive\n");
  // if (_msg.res)
  //  sema_wait(&(_wrShmArea->contSem));  
  return(_msg.res);
};


int CpuApi::stop(int pid) 
{
  _msg.mtype = 1;
  if (pid == 0)
    _msg.cmd = CMD_STOP;
  else
    _msg.cmd = CMD_ISTOP;
  _msg.intArg[0] = pid;
  _msgQueue->send(&_msg);
  if (!_msgQueue->recv(&_msg, _pid, BLOCKING))
    printf("msgQueue did not receive\n");
  return(_msg.res);
};


void CpuApi::svcDestroy()
{
  _msg.mtype = 1;
  _msg.cmd = CMD_SDESTROY;
  _msgQueue->send(&_msg);
};


void CpuApi::yield() 
{
  // _p->nowaitFlag = nowait;
  sem_post(&(_wrShmArea->yieldSem));
  sem_wait(&(_wrShmArea->contSem));

  if (_cpsBuf)
    _cpsBuf->insert(getProcStats());
  if (_ssBuf) {
    SchedStats* ss = getSchedStats();
    timeval t = Timeval::subr2(&(ss->clock), &(_ssBuf->getLast()->clock));
    if (Timeval::compare(&t, _ssSamplingInterval) >= 0)
      _ssBuf->update(ss);
  }
};



int CpuApi::setRecordProcStats(char* fname)
{
  _psFname = (char*)malloc(50*sizeof(char));
  if (fname)
    sprintf(_psFname, "%s-ps.dat", _psFname);
  else
    sprintf(_psFname, "%d-ps.dat", _pid);
  if (!_cpsBuf)
    _cpsBuf = new CliProcStatsBuf();
  return(1);
};


int CpuApi::setRecordSchedStats(char* fname, int samplingInterval)
{
  if (_ssBuf)
    return(1);
  char name[200];
  if (fname)
    sprintf(name, "%s-ss.dat", fname);
  else
    sprintf(name, "%d-ss.dat", _pid);
  _ssBuf = new SchedStatsBuf(BATCH_SIZE, name);
  _ssSamplingInterval = samplingInterval;
  return(1);
};


int CpuApi::setApDeadline(int deadline)
{
  if (deadline > TIMER_RESOLUTION) {
    _wrShmArea->apDeadline = deadline;
    return(1);
  }
  else
    return(0);
};


int CpuApi::probe(int period)
{
  if (period > 0)
    _probeCptFlag = 1;
  else
    _probeCptFlag = 0;

  if (_cpsBuf) {
    if (_psFname)
      _cpsBuf->writeFile(_psFname);
    delete _cpsBuf;
  }
  _cpsBuf = new CliProcStatsBuf();

  _msg.mtype = 1;
  _msg.cmd = CMD_PROBE;
  _msg.intArg[0] = period;
  _msgQueue->send(&_msg);
  if (!_msgQueue->recv(&_msg, _pid, BLOCKING)) {
    printf("msgQueue did not receive\n");
    return(0);
  }
  return(_msg.res);
};


int CpuApi::probeEnd()
{
  _msg.mtype = 1;
  _msg.cmd = CMD_PROBE_END;
  _msgQueue->send(&_msg);
  if (!_msgQueue->recv(&_msg, _pid, BLOCKING)) {
    printf("msgQueue did not receive\n");
    return(0);
  }
  return(_msg.res);
};


int CpuApi::probeMatch(CpuReserve* cr)
{
  if (!_cpsBuf)
    return(0);

  _cpsBuf->print();
  if (_probeCptFlag) {
    int avgUsage = (int)(_cpsBuf->findAvgUsage() * 1.1);
    int peakUsage = _cpsBuf->findPeakUsage();
    int bt = _cpsBuf->findBurstTolerance(avgUsage, SSBTR);
    if (bt == 0) {
      cr->serviceClass = CPU_RT_PCPT;
      cr->peakProcessingTime = avgUsage;
      cr->period = _cpsBuf->findPeriod();
    }
    else {
      cr->serviceClass = CPU_RT_PVPT;
      cr->peakProcessingTime = max(peakUsage, avgUsage);
      cr->sustainableProcessingTime = avgUsage;
      cr->burstTolerance = bt;
      cr->period = _cpsBuf->findPeriod();
    }
  } 
  else {
    float avgUtil = _cpsBuf->findAvgUtil() * 1.1;
    cr->serviceClass = CPU_RT_ACPU;
    cr->peakProcessingUtil = avgUtil;
  }
  return(1);
}


int CpuApi::profileSave(int argc, char* argv)
{
  return(0);
};


int CpuApi::profileGet(int argc, char* argv, float* util)
{
  return(0);
};


void CpuApi::printReserveUsage()
{
  printf("[-reserve CPU_RT_PCPT  period<ms> peakProcessingTime<ms>\n");
  printf("[         CPU_RT_PCCPT period<ms> peakProcessingTime<ms>\n");
  printf("          CPU_RT_PVPT  period<ms> peakProcessingTime<ms>\n");
  printf("                       sustainableProcessingTime<ms> burstTolerance<ms>\n");
  printf("          CPU_RT_ACPU  peakProcessingUtil]\n");
  printf("          CPU_RT_Event peakProcessingTime<ms>]\n");
}


void CpuApi::printAdaptUsage()
{
  printf("[-adapt ADAPT_EXPO windowSize alpha\n");
  printf("        ADAPT_EXPO_UP windowSize alpha\n");
  printf("        ADAPT_EXPO_RANGE windowSize alpha max min\n");
  printf("        ADAPT_STAT windowSize nonconformFreq]\n");
}


// returns the number of parsed arguments
int CpuApi::parseReserveArg(CpuReserve* cr, char* argv[], int mark) 
{
  if (strcmp(argv[mark], "CPU_RT_PCPT") == 0) {
    cr->serviceClass = CPU_RT_PCPT;
    cr->period = atoi(argv[mark+1]) * 1000; 
    cr->peakProcessingTime = atoi(argv[mark+2]) * 1000; 
    return(3);
  }
  else if (strcmp(argv[mark], "CPU_RT_PVPT") == 0) {
    cr->serviceClass = CPU_RT_PVPT;
    cr->period = atoi(argv[mark+1]) * 1000; 
    cr->peakProcessingTime = atoi(argv[mark+2]) * 1000; 
    cr->sustainableProcessingTime = atoi(argv[mark+3]) * 1000; 
    cr->burstTolerance = atoi(argv[mark+4]) * 1000; 
    return(5);
  }
  else if (strcmp(argv[mark], "CPU_RT_PCCPT") == 0) {
    cr->serviceClass = CPU_RT_PCCPT;
    cr->period = atoi(argv[mark+1]) * 1000; 
    cr->peakProcessingTime = atoi(argv[mark+2]) * 1000; 
    return(3);
  }
  else if (strcmp(argv[mark], "CPU_RT_ACPU") == 0) {
    cr->serviceClass = CPU_RT_ACPU;
    cr->peakProcessingUtil = atof(argv[mark+1]); 
    return(2);
  }
  else if (strcmp(argv[mark], "CPU_RT_EVENT") == 0) {
    cr->serviceClass = CPU_RT_EVENT;
    cr->peakProcessingTime = atoi(argv[mark+1]);
    return(2);
  }
  else
    return(0);
}


int CpuApi::parseAdaptArg(AdaptStrategy* as, char* argv[], int mark)
{
  if (strcmp(argv[mark], "ADAPT_EXPO") == 0) {
    as->strategy = ADAPT_EXPO;
    as->windowSize = atoi(argv[mark+1]);
    as->alpha = atof(argv[mark+2]);
    return(3);
  }
  else if (strcmp(argv[mark], "ADAPT_EXPO_UP") == 0) {
    as->strategy = ADAPT_EXPO_UP;
    as->windowSize = atoi(argv[mark+1]);
    as->alpha = atof(argv[mark+2]);
    return(3);
  }
  else if (strcmp(argv[mark], "ADAPT_EXPO_RANGE") == 0) {
    as->strategy = ADAPT_EXPO_RANGE;
    as->windowSize = atoi(argv[mark+1]);
    as->alpha = atof(argv[mark+2]);
    as->max = atof(argv[mark+3]);
    as->min = atof(argv[mark+4]);
    if (as->max > 1) {
      as->max *= 1000;
      as->min *= 1000;
    }
    return(5);
  }
  else if (strcmp(argv[mark], "ADAPT_STAT") == 0) {
    as->strategy = ADAPT_STAT;
    as->windowSize = atoi(argv[mark+1]);
    as->nonconformFreq = atof(argv[mark+2]);
    return(3);
  }
  else if (strcmp(argv[mark], "ADAPT_NONE") == 0) {
    as->strategy = ADAPT_NONE;
    return(1);
  }
  else
    return(0);
};



ProcStats* CpuApi::getProcStats(int pid) {
    Process* p = getProcess(pid);
    if (p == NULL)
	return(NULL);
    else
	return(&(p->ps));
}


Process* CpuApi::getProcess(int pid) { 
    if (pid == 0)
	return(_p);
    else	
	return(_shm->getProcessWithPid(pid));
}


int CpuApi::isPermNonconform(int pid) { 
    ProcStats* ps = getProcStats(pid);
    if (!ps)
	return(ps->permNonconformFlag);
    else
	return(0);
};


int CpuApi::getPidTab(int* pidTab, int len) {
    int n = (len > WINDOW_SIZE) ? WINDOW_SIZE : len;
    ShmArea* area = getShmArea();
    for (int i=0; i<n; i++) {
	if (area->pArrayFlag[i])
	    pidTab[i] = area->pArray[i].pid;
	else
	    pidTab[i] = 0;
    }

// Tomonori Yamane
    return (n);
}
