/* 
 * Released Date: March 7 99
 * This is the Dynamic Soft CPU Scheduler II release by the 
 *   Monet Group, 
 *   http://cairo.cs.uiuc.edu/
 *   Department of Computer Science, 
 *   University of Illinois at Urbana Champaign.  
 * 
 * This release has been tested under SUN OS 5.6.
 *
 * Contacting Authors:
 *  Hao-hua Chu (h-chu3@cs.uiuc.edu)
 *  Klara Nahrstedt (klara@cs.uiuc.edu)
 */

#include <string.h>
#include "CpuServiceClass.H"
#include "ProcStatsBuf.H"
#include "Timeval.H"


CpuPcptClass cpuPcptClass;
CpuPcptClass cpuPccptClass;
CpuPvptClass cpuPvptClass;
CpuAcptClass cpuAcptClass;


int CpuServiceClass::isValid(CpuReserve* cr) 
{
  switch (cr->serviceClass)
    {
    case CPU_RT_PCPT:
      if ((cr->period <= TIMER_RESOLUTION) ||
	  (cr->peakProcessingTime <= 0) ||
	  (cr->peakProcessingTime > cr->period))
	return(0);
      break;
    case CPU_RT_PCCPT:
      if ((cr->period <= TIMER_RESOLUTION) ||
	  (cr->peakProcessingTime <= 0) ||
	  (cr->peakProcessingTime > cr->period))
	return(0);
      break;
    case CPU_RT_PVPT:
      if ((cr->period <= TIMER_RESOLUTION) ||
	  (cr->peakProcessingTime <= 0) ||
	  (cr->sustainableProcessingTime <= 0) || 
	  (cr->burstTolerance < 0) ||
	  (cr->sustainableProcessingTime > cr->period) ||
	  (cr->sustainableProcessingTime >= cr->peakProcessingTime))
	return(0);
      break;
    case CPU_RT_ACPU:
      if (cr->peakProcessingUtil <= 0)
	return(0);
    case CPU_RT_EVENT:
    default:
      break;
    }
  return(1);
};



void CpuServiceClass::print(CpuReserve* cr)
{
  switch (cr->serviceClass) {
  case CPU_RT_PCPT:
    printf("(RT_PCPT: period=%d, PPT=%d)",
	   cr->period, cr->peakProcessingTime);
    break;
  case CPU_RT_PCCPT:
    printf("(RT_PCCPT: period=%d, PPT=%d)",
	   cr->period, cr->peakProcessingTime);
    break;
  case CPU_RT_PVPT:
    printf("(RT_PVPT: period=%d, PPT=%d, SPT=%d, BT=%d)",
	   cr->period, cr->peakProcessingTime,
	   cr->sustainableProcessingTime, cr->burstTolerance);
    break;
  case CPU_RT_ACPU:
    printf("(RT_ACPU: PPU=%f)",
	   cr->peakProcessingUtil);
    break;
  case CPU_RT_EVENT:
    printf("(RT_EVENT: PPT=%d)",
	   cr->peakProcessingTime);
    break;
  }
};



void CpuServiceClass::print(AdaptStrategy* as) 
{
  switch(as->strategy) {
  case ADAPT_NONE:
    printf("(ADAPT_NONE: ");
    break;
  case ADAPT_EXPO:
    printf("(ADAPT_EXPO: alpha=%f, ",
	   as->alpha);
    break;
  case ADAPT_EXPO_UP:
    printf("(ADAPT_EXPO_UP: alpha=%f, \n",
	   as->alpha);
    break;
  case ADAPT_EXPO_RANGE:
    printf("(ADAPT_EXPO_RANGE: alpha=%f, max=%f, min=%f, ",
	   as->alpha, as->max, as->min);
    break;
  case ADAPT_STAT:
    printf("(ADAPT_STAT: nonconformFreq=%f, ",
	   as->nonconformFreq);
    break;
  }
  printf("windowSize=%d)", as->windowSize);
};



float CpuServiceClass::getUtil(CpuReserve* cr)
{
  switch (cr->serviceClass)
    {
    case CPU_RT_PCPT:
    case CPU_RT_PCCPT:
      return((float)cr->peakProcessingTime / (float)cr->period);
    case CPU_RT_PVPT:
      return((float)cr->sustainableProcessingTime / (float)cr->period);
    case CPU_RT_ACPU:
      return(cr->peakProcessingUtil);
    default:
      printf("CpuServiceClass::getUtil unknown class\n");
      return(0.0);
    }
};


void CpuServiceClass::newReservation(Process* p)
{
  if (p->probeFlag) {
    if (p->cpuReserve.period != 0)
      p->cpuServiceClass = &cpuPcptClass;
    else if (p->cpuReserve.period != 0)
      p->cpuServiceClass = &cpuPccptClass;
    else
      p->cpuServiceClass = &cpuAcptClass;
    return;
  }

  switch (p->cpuReserve.serviceClass)
    {
    case CPU_RT_PCPT:
      p->cpuServiceClass = &cpuPcptClass;
      p->guaranteed = p->cpuReserve.peakProcessingTime;
      break;
    case CPU_RT_PCCPT:
      p->cpuServiceClass = &cpuPccptClass;
      p->guaranteed = p->cpuReserve.peakProcessingTime;
      break;
    case CPU_RT_PVPT:
      p->cpuServiceClass = &cpuPvptClass;
      p->guaranteed = p->cpuReserve.sustainableProcessingTime;
      break;
    case CPU_RT_ACPU:
      p->cpuServiceClass = &cpuAcptClass;
      p->guaranteed = (int)((float)(p->cpuReserve.period) * 
			    p->cpuReserve.peakProcessingUtil);
      break;
    }
  p->cpuServiceClass->leakingBucketInit(p, &(p->cpuReserve));
};


int CpuServiceClass::resetDeadline(Process* p, timeval* clock)
{
  int n = Timeval::incrTillGreater(clock, &(p->deadline), 
				   p->cpuReserve.period);
  if (p->state == PSTATE_OVERRUN)
    p->overrunItr += n;
  return(n);
};


void CpuPcptClass::adaptExpo(CpuReserve* cr, CpuReserve* newCr, 
			     AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  memcpy(newCr, cr, sizeof(CpuReserve));
  int start = psBuf->getIdx() - as->windowSize;
  int end = psBuf->getIdx() - 1;
  newCr->peakProcessingTime = 
    (int)(cr->peakProcessingTime * (1.0 - as->alpha) +
	  psBuf->findAvgUsage(start, end) * as->alpha);
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
};


void CpuPccptClass::adaptExpo(CpuReserve* cr, CpuReserve* newCr, 
			     AdaptStrategy* as, ProcStatsBuf* psBuf)
{
// TODO: need to look into this
/*
  memcpy(newCr, cr, sizeof(CpuReserve));
  int start = psBuf->getIdx() - as->windowSize;
  int end = psBuf->getIdx() - 1;
  newCr->peakProcessingTime = 
    (int)(cr->peakProcessingTime * (1.0 - as->alpha) +
	  psBuf->findAvgUsage(start, end) * as->alpha);
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
	  */
};

void CpuPvptClass::adaptExpo(CpuReserve* cr, CpuReserve* newCr, 
			     AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  memcpy(newCr, cr, sizeof(CpuReserve));
  int start = psBuf->getIdx() - as->windowSize;
  int end = psBuf->getIdx() - 1;
  newCr->sustainableProcessingTime = 
    (int)(cr->sustainableProcessingTime * (1.0 - as->alpha) +
	  psBuf->findAvgUsage(start, end) * as->alpha);
  if (newCr->sustainableProcessingTime > newCr->period)
      newCr->sustainableProcessingTime = newCr->period;
};


void CpuAcptClass::adaptExpo(CpuReserve* cr, CpuReserve* newCr, 
			     AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  memcpy(newCr, cr, sizeof(CpuReserve));
  int start = psBuf->getIdx() - as->windowSize;
  int end = psBuf->getIdx() - 1;
  newCr->peakProcessingUtil = 
    cr->peakProcessingUtil * (1.0 - as->alpha) +
    psBuf->findAvgUtil(start, end) * as->alpha;
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
};


void CpuPcptClass::adaptExpoUp(CpuReserve* cr, CpuReserve* newCr, 
			       AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  adaptExpo(cr, newCr, as, psBuf);
  newCr->peakProcessingTime = max(newCr->peakProcessingTime, 
				  cr->peakProcessingTime);
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
};

void CpuPccptClass::adaptExpoUp(CpuReserve* cr, CpuReserve* newCr, 
			       AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  // TODO: look into this
  /*
  adaptExpo(cr, newCr, as, psBuf);
  newCr->peakProcessingTime = max(newCr->peakProcessingTime, 
				  cr->peakProcessingTime);
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
	  */
};

void CpuPvptClass::adaptExpoUp(CpuReserve* cr, CpuReserve* newCr, 
			       AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  adaptExpo(cr, newCr, as, psBuf);
  newCr->sustainableProcessingTime = max(newCr->sustainableProcessingTime, 
					 cr->sustainableProcessingTime);
  if (newCr->sustainableProcessingTime > newCr->period)
      newCr->sustainableProcessingTime = newCr->period;
};


void CpuAcptClass::adaptExpoUp(CpuReserve* cr, CpuReserve* newCr, 
			       AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  adaptExpo(cr, newCr, as, psBuf);
  newCr->peakProcessingUtil = max(newCr->peakProcessingUtil,
				  cr->peakProcessingUtil);
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
};


void CpuPcptClass::adaptExpoRange(CpuReserve* cr, CpuReserve* newCr, 
				  AdaptStrategy* as, ProcStatsBuf* psBuf)
{ 
  adaptExpo(cr, newCr, as, psBuf);
  newCr->peakProcessingTime =  
    max(min((int)as->max, newCr->peakProcessingTime), (int)as->min);
};

void CpuPccptClass::adaptExpoRange(CpuReserve* cr, CpuReserve* newCr, 
				  AdaptStrategy* as, ProcStatsBuf* psBuf)
{ 
// TODO: look into this
/*
  adaptExpo(cr, newCr, as, psBuf);
  newCr->peakProcessingTime =  
    max(min((int)as->max, newCr->peakProcessingTime), (int)as->min);
	*/
};

void CpuPvptClass::adaptExpoRange(CpuReserve* cr, CpuReserve* newCr, 
				  AdaptStrategy* as, ProcStatsBuf* psBuf)
{ 
  adaptExpo(cr, newCr, as, psBuf);
  newCr->sustainableProcessingTime =  
    max(min((int)as->max, newCr->sustainableProcessingTime), (int)as->min);
};


void CpuAcptClass::adaptExpoRange(CpuReserve* cr, CpuReserve* newCr, 
				  AdaptStrategy* as, ProcStatsBuf* psBuf)
{ 
  adaptExpo(cr, newCr, as, psBuf);
  newCr->peakProcessingUtil =  
    max(min(as->max, newCr->peakProcessingUtil), as->min);
};


void CpuPcptClass::adaptStat(CpuReserve* cr, CpuReserve* newCr, 
			     AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  memcpy(newCr, cr, sizeof(CpuReserve));
  int start = psBuf->getIdx() - as->windowSize;
  int end = psBuf->getIdx() - 1;
  newCr->peakProcessingTime = psBuf->findUsageAtFreq(start, end, as->nonconformFreq);
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
};

void CpuPccptClass::adaptStat(CpuReserve* cr, CpuReserve* newCr, 
			     AdaptStrategy* as, ProcStatsBuf* psBuf)
{
// TODO: look into this
/*
  memcpy(newCr, cr, sizeof(CpuReserve));
  int start = psBuf->getIdx() - as->windowSize;
  int end = psBuf->getIdx() - 1;
  newCr->peakProcessingTime = psBuf->findUsageAtFreq(start, end, as->nonconformFreq);
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
	  */
};

void CpuPvptClass::adaptStat(CpuReserve* cr, CpuReserve* newCr, 
			     AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  memcpy(newCr, cr, sizeof(CpuReserve));
  int start = psBuf->getIdx() - as->windowSize;
  int end = psBuf->getIdx() - 1;
  newCr->sustainableProcessingTime = psBuf->findUsageAtFreq(start, end, as->nonconformFreq);
  if (newCr->sustainableProcessingTime > newCr->period)
      newCr->sustainableProcessingTime = newCr->period;
};


void CpuAcptClass::adaptStat(CpuReserve* cr, CpuReserve* newCr, 
			     AdaptStrategy* as, ProcStatsBuf* psBuf)
{
  memcpy(newCr, cr, sizeof(CpuReserve));
  int start = psBuf->getIdx() - as->windowSize;
  int end = psBuf->getIdx() - 1;
  newCr->peakProcessingUtil = psBuf->findUtilAtFreq(start, end, as->nonconformFreq);
  if (newCr->peakProcessingTime > newCr->period)
      newCr->peakProcessingTime = newCr->period;
};


void CpuPcptClass::leakingBucketInit(Process* p, CpuReserve* cr)
{
  p->leakingBucket->init(cr->peakProcessingTime, 
			 cr->peakProcessingTime * (1+SSBTR));
};

void CpuPccptClass::leakingBucketInit(Process* p, CpuReserve* cr)
{
  // TODO: make sure this is correct
  p->leakingBucket->init(cr->peakProcessingTime, 
			 cr->peakProcessingTime * (1+SSBTR));
};


void CpuPvptClass::leakingBucketInit(Process* p, CpuReserve* cr)
{
  p->leakingBucket->init(cr->sustainableProcessingTime, 
			 cr->sustainableProcessingTime * (1+SSBTR) + cr->burstTolerance);
  p->leakingBucket2->init(cr->peakProcessingTime, 
			  cr->peakProcessingTime * (1+SSBTR));
};

void CpuAcptClass::leakingBucketInit(Process* p, CpuReserve* cr)
{
  p->leakingBucket->init(cr->peakProcessingUtil * (float)cr->period, 
			 cr->peakProcessingUtil * (float)cr->period * (1+SSBTR));
};


void CpuPcptClass::leakingBucketAdjust(Process* p, CpuReserve* cr)
{
  p->leakingBucket->setLeak(cr->peakProcessingTime);
  p->leakingBucket->setDepth(cr->peakProcessingTime * (1+SSBTR));
};

void CpuPccptClass::leakingBucketAdjust(Process* p, CpuReserve* cr)
{
  //TODO: make sure this is correct
  p->leakingBucket->setLeak(cr->peakProcessingTime);
  p->leakingBucket->setDepth(cr->peakProcessingTime * (1+SSBTR));
};


void CpuPvptClass::leakingBucketAdjust(Process* p, CpuReserve* cr)
{
  p->leakingBucket->setLeak(cr->sustainableProcessingTime);
  p->leakingBucket->setDepth(cr->sustainableProcessingTime * (1+SSBTR) + cr->burstTolerance);
  p->leakingBucket2->setLeak(cr->peakProcessingTime);
  p->leakingBucket2->setDepth(cr->peakProcessingTime * (1+SSBTR));
};


void CpuAcptClass::leakingBucketAdjust(Process* p, CpuReserve* cr)
{
  p->leakingBucket->setLeak(cr->peakProcessingUtil * (float)cr->period);
  p->leakingBucket->setDepth(cr->peakProcessingUtil * (float)cr->period * (1+SSBTR));
};


int CpuPvptClass::leakingBucketFill(Process* p, int usage)
{
  p->leakingBucket->fill(usage);
  p->leakingBucket2->fill(usage);
  return(p->leakingBucket->isNonconform() ||
	 p->leakingBucket2->isNonconform());
};


int CpuServiceClass::leakingBucketFlush(Process* p, int n)
{ 
  for (int i=0; i<n; i++) {
    if (p->leakingBucket->isEmpty())
      break;
    p->leakingBucket->flush(); 
  }
  return(p->leakingBucket->isNonconform());
};


int CpuPvptClass::leakingBucketFlush(Process* p, int n)
{
  for (int i=0; i<n; i++) {
    if (p->leakingBucket->isEmpty() &&
	p->leakingBucket2->isEmpty())
      break;
    p->leakingBucket->flush();
    p->leakingBucket2->flush();
  }
  return(p->leakingBucket->isNonconform() ||
	 p->leakingBucket2->isNonconform());
};

