#ifndef DEF_H
#define DEF_H

#include <sys/time.h>
#include <semaphore.h>
#include "CpuReserve.H"
#include "LeakingBucket.H"

#define WINDOW_SIZE 20            // default value for detecting perm overrun
#define NONCONFORM_TOLERATE_FREQ 0.3  // default value for detecting perm overrun 
#define SSBTR 0.1 // System Specific Burst Tolerance Rate

#define CMD_INIT 0
#define CMD_DESTROY 1
#define CMD_RESERVE 2
#define CMD_MODIFY 3
#define CMD_FREE 4
#define CMD_START 5
#define CMD_STOP 6
#define CMD_SDESTROY 7
#define CMD_SET_ADAPT_STRATEGY 8
#define CMD_SYNC 9
#define CMD_PROBE 10
#define CMD_PROBE_END 11

#define CMD_IRESERVE 12
#define CMD_IMODIFY 13
#define CMD_IFREE 14
#define CMD_IDESTROY 15
#define CMD_ISET_ADAPT_STRATEGY 16
#define CMD_ISTOP 17
#define CMD_ISTART 18

#define BLOCKING 1
#define NONBLOCKING 0

#define AS_MUCH_FLAG 1

inline char* cmdToString(int cmd)
{
  switch (cmd) {
  case CMD_INIT:
    return("INIT");
  case CMD_DESTROY:
    return("DESTROY");
  case CMD_RESERVE:
    return("RESERVE");
  case CMD_MODIFY:
    return("MODIFY");
  case CMD_FREE:
    return("FREE");
  case CMD_START:
    return("START");
  case CMD_STOP:
    return("STOP");
  case CMD_ISTOP:
    return("ISTOP");
  case CMD_SDESTROY:
    return("SDESTROY");
  case CMD_SET_ADAPT_STRATEGY:
    return("SET_ADAPT_STRATEGY");
  case CMD_SYNC:
    return("SYNC");
  case CMD_PROBE:
    return("PROBE");
  case CMD_PROBE_END:
    return("PROBE_END");
  case CMD_IRESERVE:
    return("IRESERVE");
  case CMD_IMODIFY:
    return("IMODIFY");
  case CMD_IFREE:
    return("CMD_IFREE");
  case CMD_IDESTROY:
    return("CMD_IDESTROY");
  case CMD_ISET_ADAPT_STRATEGY:
    return("CMD_ISET_ADAPT_STRATEGY");
  default:
    return("UNKNOWN");
  }
}

#define MSG_QUEUE_KEY 9999

#define MAX_PROCESS 20
#define TIMER_RESOLUTION 10000  // in us

/*
  PSTATE_INIT: without any reservation.  init, init-reserve-start-stop-free 
               init-probe-start-stop-probeEnd.  
  PSTATE_RESERVE: with reservation but not running.  init-reserve, init-reserve-start-stop
  PSTATE_PROBE_INIT: with probing but not running.  init-probe, init-probe-start-stop
  PSTATE_RUN: running, but not finished its iteration.  init-reserve-start
  PSTATE_WAIT: running, but finished its iteration.  init-reserve-start
  PSTATE_OVERRUN: overrunning.  init-reserve-start
  PSTATE_SCHED: being dispatched.  init-reserve-start
  PSTATE_PROBE: probing.  init-probe-start
*/
#define PSTATE_INIT 0
#define PSTATE_RESERVE 1 
#define PSTATE_PROBE_INIT 2 
#define PSTATE_RUN 3     
#define PSTATE_WAIT 4    
#define PSTATE_OVERRUN 5 
#define PSTATE_SCHED 6   
#define PSTATE_PROBE 7   


inline char* pstateToString(int state) {
  switch (state) 
    {
    case PSTATE_INIT:
      return("PSTATE_INIT");
    case PSTATE_RESERVE:
      return("PSTATE_RESERVE");
    case PSTATE_PROBE_INIT:
      return("PSTATE_PROBE_INIT");
    case PSTATE_RUN:
      return("PSTATE_RUN");
    case PSTATE_WAIT:
      return("PSTATE_WAIT");
    case PSTATE_OVERRUN:
      return("PSTATE_OVERRUN");
    case PSTATE_SCHED:
      return("PSTATE_SCHED");
    case PSTATE_PROBE:
      return("PSTATE_PROBE");
    default:
      return("PSTATE_ERROR");
    }
}

typedef struct Msg {
  long mtype;
  int callerPid;
  int cmd;
  int idx;
  int res;
  union 
  {
    CpuReserve cr;
    AdaptStrategy as;
  };
  int intArg[2];
} Msg;


/*  
 * RESERVE/MODIFY/IRESERVE/IMODIFY
 *  pid = int intArg[0]
 * ADAPT/IADAPT:  
 *  pid = intArg[0]
 * SYNC:
 *  syncId = intArg[0]
 *  np = intArg[1]
 * IDESTROY/IFREE:
 *  pid = intArg[0]
 * START:
 *  start = intArg[0/1];
 */


const timeval TIMEVAL_INIT = {0, 0};


typedef struct ProcStats {
  timeval clock;
  int usage;           // its usage in us.
  int nonconformFlag;  // conformance 0/1.
  int overrunFlag; 
  CpuReserve cpuReserve;
  int apDeadline;
  int permNonconformFlag; // 0/1
  int overrunItr;      // its number of iterations that it may have overrun.

  timeval accUsage;
  int accNumNonconform;
  int accNumOverrun;
  int accNumPermNonconform;
  int accNumDeadlineMiss;
  int accItr;
} ProcStats;


typedef struct SchedStats {
  timeval clock;

  int numProcessor;
  float availPermUtil; 
  float overrunRatio;
  float permRatio;
  float tsRatio;

  timeval permUsage;
  timeval burstUsage;
  timeval nonconformUsage;
  timeval probeUsage;
  timeval tsUsage;
} SchedStats;


class ProcStatsBuf;
class PQueue;
class CpuServiceClass;


typedef struct WrShmArea {
  sem_t yieldSem;
  sem_t contSem;
  int apDeadline;     // For aperiodc class: the length of the next period.
} WrShmArea;


typedef struct Process {
  int idx;            // the idx in the pArray in the ShmArea.
  int pid; 
  int euid;
  int egid;
  int state;          // in any one of the PSTATE_XXX.
  PQueue* pq;           // holds the pointer to the PQueue.
  int cpu;
    
  CpuReserve cpuReserve;
  CpuServiceClass* cpuServiceClass;
  AdaptStrategy adaptStrategy;  // struct supporting adaptation
    
  timeval begin;      // begin time
  int guaranteed;     // the amount of guaranteed processing time.
  LeakingBucket* leakingBucket; // for conformance test.
  LeakingBucket* leakingBucket2; // for conformance test.
  int probeFlag;      // if it is in the probing phase.
  // int syncId;         // synchronization id.
    
  int usage;          // usage time(us) in the current iteration.
  int usageOverrunItr;  // usage time(us) accumulated from the previous overrunItr.
  int overrunFlag;    // if it is in an overrun.
  int nonconformFlag; // if it is conforming.
  int permNonconformFlag; 
  int overrunItr;     // the number of iterations it has overrun.
  timeval deadline;   // the expected deadline.

  timeval accUsage;
  int accNumNonconform;
  int accNumOverrun;
  int accNumPermNonconform;
  int accNumDeadlineMiss;
  int accItr;

  int yieldFlag;      // it has yielded.
  // int nowaitFlag;     // its yield should skip to next iteration.
  int resetContSemFlag;

  ProcStats ps;       // stats information about the last iteration.
  ProcStatsBuf* psBuf; // history for detecting permanent overrun and adaptation

  int procFd;
  int wrShmAreaFd;
  WrShmArea* wrShmArea;
  Process* next;      // used in PQueue.
} Process;




void msgPrint(Msg* msg);
void processPrint(Process* p);


#endif DEF_H
