/* 
 * Released Date: March 7 99
 * This is the Dynamic Soft CPU Scheduler II release by the 
 *   Monet Group, 
 *   http://cairo.cs.uiuc.edu/
 *   Department of Computer Science, 
 *   University of Illinois at Urbana Champaign.  
 * 
 * This release has been tested under SUN OS 5.6.
 *
 * Contacting Authors:
 *  Hao-hua Chu (h-chu3@cs.uiuc.edu)
 *  Klara Nahrstedt (klara@cs.uiuc.edu)
 */

#include <stdlib.h>
#include "ProcStatsBuf.H"

#define BLK_SIZE 10

ProcStatsBuf::ProcStatsBuf(int blkSize)
{
  _blkSize = blkSize;
  _blk = (ProcStats*)malloc(blkSize * sizeof(ProcStats));
  _idx = 0;
}


ProcStatsBuf::~ProcStatsBuf()
{
  _idx = 0;
  if (_blk)
    free(_blk);
  _blk = NULL;
}


void ProcStatsBuf::setBlkSize(int blkSize)
{
  if (_blkSize != blkSize) {
    _blkSize = blkSize;
    free(_blk);
    _blk = (ProcStats*)malloc(blkSize * sizeof(ProcStats));
  }
}


void ProcStatsBuf::insert(ProcStats* ps)
{
  _blk[_idx%_blkSize] = *ps;
  _idx++;
}


int ProcStatsBuf::findAvgUsage(int start, int end)
{
  int sum = 0;
  for (int i=start; i<=end; i++)
    sum += _blk[i%_blkSize].usage;
  return(sum/(end-start+1));
}


float ProcStatsBuf::findAvgUtil(int start, int end)
{
  float sum = 0;
  for (int i=start; i<=end; i++)
    sum += (float)(_blk[i%_blkSize].usage) / (float)(_blk[i%_blkSize].cpuReserve.period);
  return(sum/(end-start+1));
}


int ProcStatsBuf::findUsageAtFreq(int start, int end, float nonconformFreq) 
{
  int len = end - start + 1;
  int* utilArray = (int*)malloc(len*sizeof(int));
  // printf("usage: ");
  for (int i=start, j=0; i<=end, j<len; i++, j++) {
    // printf("%d ", _blk[i].usage);
    utilArray[j] = _blk[i%_blkSize].usage;
  }
  // printf("\n");
  qsort((char*)utilArray, len, sizeof(int), intcompare);
  int res = utilArray[(int)((len - 1)*(1.0 - nonconformFreq))];
  free(utilArray);
  return(res);
}


float ProcStatsBuf::findUtilAtFreq(int start, int end, float nonconformFreq) 
{
  int len = end - start + 1;
  float* utilArray = (float*)malloc(len*sizeof(float));
  // printf("usage: ");
  for (int i=start, j=0; i<=end, j<len; i++, j++) {
    // printf("%d ", _blk[i].usage);
    utilArray[j] = (float)(_blk[i%_blkSize].usage)/(float)(_blk[i%_blkSize].cpuReserve.period);
  }
  // printf("\n");
  qsort((char*)utilArray, len, sizeof(float), floatcompare);
  float res = utilArray[(int)((len - 1)*(1.0 - nonconformFreq))];
  free(utilArray);
  return(res);
}


float ProcStatsBuf::findNonconformFreq(int start, int end)
{
  int nonconform = 0;
  for (int i=start; i<=end; i++)
    nonconform += _blk[i%_blkSize].nonconformFlag;
  return(nonconform/(end-start+1));
}


CliProcStatsBuf::CliProcStatsBuf() : ProcStatsBuf(BLK_SIZE)
{
  _head = (ProcStatsLink*)malloc(sizeof(ProcStatsLink));
  _head->blk = _blk;
  _head->next = NULL;
  _tail = _head;
}


CliProcStatsBuf::~CliProcStatsBuf()
{
  while (_head != NULL) {
    ProcStatsLink* tmp = _head;
    free(tmp->blk);
    free(tmp);
    _head = _head->next;
  }
}


void CliProcStatsBuf::insert(ProcStats* ps)
{
  ProcStatsBuf::insert(ps);
  if (_idx % _blkSize == 0) {
    _blk = (ProcStats*)malloc(_blkSize * sizeof(ProcStats));
    ProcStatsLink* psl = (ProcStatsLink*)malloc(sizeof(ProcStatsLink));
    psl->blk = _blk;
    psl->next = NULL;
    _tail->next = psl;
    _tail = psl;
  }
}


void CliProcStatsBuf::writeFile(char* fname)
{
  FILE* fd;
  timeval begin = _head->blk[0].clock;
  timeval accumUsage = TIMEVAL_INIT;

  fd = fopen(fname, "w");
  fprintf(fd, "#%11s %10s %10s %5s %10s %5s %5s %10s %6s %6s %6s %6s\n",
	  "clk", "period", 
	  "guaran(us)", "resv",
	  "usage(us)",  "usage", "resv*", "accum", 
	  "ovFlg", "ncFlg", "pncFlg", "ovIt");
  
  ProcStatsLink* psl = _head;
  ProcStats* blk = psl->blk;
  for (int i=0; i<_idx; i++) {
    if (i != 0 && i%_blkSize == 0) {
      psl = psl->next;
      blk = psl->blk;
    }
    int j = i%_blkSize;
    int guaranteed;    
    switch(blk[j].cpuReserve.serviceClass) 
      {
      case CPU_RT_PCPT:
      case CPU_RT_PCCPT:
	guaranteed = blk[j].cpuReserve.peakProcessingTime;
	break;
      case CPU_RT_PVPT:
	guaranteed = blk[j].cpuReserve.sustainableProcessingTime;
	break;
      case CPU_RT_ACPU:
	guaranteed = (int)(blk[j].cpuReserve.peakProcessingUtil * (float)blk[j].cpuReserve.period);
	break;
      }
    int t = Timeval::subr(&(blk[j].clock), &begin);
    Timeval::incr(&accumUsage, blk[j].usage);
    timeval prevClock;
    fprintf(fd, "%12d %10d %10d %5.3f %10d %5.3f %5.3f %3d:%6d %6d %6d %6d %6d\n",
	    t, blk[j].cpuReserve.period, 
	    guaranteed, (float)guaranteed / (float)(blk[j].cpuReserve.period),
	    blk[j].usage, 
	    (i == 0) ?  0 : (float)blk[j].usage/Timeval::subr(&(blk[j].clock), &prevClock),
	    (float)blk[j].usage / (float)blk[j].cpuReserve.period,
	    (accumUsage.tv_sec)%1000, accumUsage.tv_usec,
	    blk[j].overrunFlag, blk[j].nonconformFlag,
	    blk[j].permNonconformFlag, blk[j].overrunItr);
    prevClock = blk[j].clock;
  }
  fclose(fd);
}


int CliProcStatsBuf::findPeriod()
{
  return(_head->blk[0].cpuReserve.period);
}


int CliProcStatsBuf::findAvgUsage()
{
  int sum = 0;
  ProcStatsLink* psl = _head;
  ProcStats* blk = psl->blk;
  for (int i=0; i<_idx; i++) {
    int j = i%_blkSize;
    if (i != 0 && j == 0) {
      psl = psl->next;
      blk = psl->blk;
    }
    sum += psl->blk[i].usage;
  }
  return(sum/_idx);
}


void CliProcStatsBuf::print()
{
  printf("%10s %10s\n", "clk", "usage");
  ProcStatsLink* psl = _head;
  ProcStats* blk = psl->blk;
  for (int i=0; i<_idx; i++) {
    int j = i%_blkSize;
    if (i != 0 && j == 0) {
      psl = psl->next;
      blk = psl->blk;
    }
    printf("%3d:%6d %10d\n", 
	   psl->blk[j].clock.tv_sec%100, 
	   psl->blk[j].clock.tv_usec, 
	   psl->blk[j].usage);
  }
}


float CliProcStatsBuf::findAvgUtil()
{
  float sum = 0;
  ProcStatsLink* psl = _head;
  ProcStats* blk = psl->blk;
  for (int i=0; i<_idx; i++) {
    int j = i%_blkSize;
    if (i != 0 && j == 0) {
      psl = psl->next;
      blk = psl->blk;
    }
    sum += (float)psl->blk[j].usage / (float)(psl->blk[j].cpuReserve.period);
  }
  return(sum/_idx);
}


int CliProcStatsBuf::findPeakUsage()
{
  int peakUsage = 0;
  ProcStatsLink* psl = _head;
  ProcStats* blk = psl->blk;
  for (int i=0; i<_idx; i++) {
    int j = i%_blkSize;
    if (i != 0 && j == 0) {
      psl = psl->next;
      blk = psl->blk;
    }
    if (peakUsage < psl->blk[j].usage)
      peakUsage = psl->blk[j].usage;
  }
  return(peakUsage);
}


int CliProcStatsBuf::findBurstTolerance(int sustainableProcessingTime, float maxNonconformFreq)
{
  int bt = 0;
  int* btArray = (int*)malloc(_idx * sizeof(int));
  printf("SPT = %d\n", sustainableProcessingTime);
  
  ProcStatsLink* psl = _head;
  ProcStats* blk = psl->blk;
  for (int i=0; i<_idx; i++) {
    int j = i%_blkSize;
    if (i != 0 && j == 0) {
      psl = psl->next;
      blk = psl->blk;
    }
    printf("%d ", psl->blk[j].usage);
    bt = max(0, bt + psl->blk[j].usage - sustainableProcessingTime);
    btArray[i] = bt;
  }
  printf("\n");
  qsort((char*)btArray, _idx, sizeof(int), intcompare);
  printf("nth = %d\n", (int)((_idx - 1)*(1.0 - maxNonconformFreq)));
  int res = btArray[(int)((_idx - 1)*(1.0 - maxNonconformFreq))];
  free(btArray);
  return(res);
}
