/******************************************************************************
*                                                                             *
*  File: NSBroker.cpp                                                         *
*                                                                             *
*  Description: Facilitates access to the name server.                        *
*                                                                             *
*  Project: 2K                                                                *
*  Author: Christopher Hess (ckhess@cs.uiuc.edu)                              *
*                                                                             *
*******************************************************************************
   Revision Control Information
   $Revision: 1.5 $
   $Log: NSBroker.cpp,v $
   Revision 1.5  2000/10/09 21:02:42  f-kon
   Changes to use new TwoK:: namespace for 2K services integrated with Tomo's
   changes for Windows compilation.

   Revision 1.4  2000/02/28 22:15:51  f-kon
   *** empty log message ***

   Revision 1.3  2000/02/16 18:06:32  ckhess
   Return _ptr instead of _var on resolve.

******************************************************************************/

#include "NSBroker.h"

/*
 * Constructor. Initialize member variable to 0.
 */
NSBroker::NSBroker ()
 : _rootContext (0)
{}

/*
 * Initialize the broker by getting a handle to the name server.
 */
int
NSBroker::init(CORBA::ORB_var orb)
{
  ACE_TRY_NEW_ENV {        
    CORBA::Object_var obj = orb->resolve_initial_references("NameService");
    if (CORBA::is_nil(obj.in()))
      ACE_ERROR_RETURN ((LM_ERROR, 
                         "(%P|%t) Unable to resolve the Name Service.\n"), -1);   
    _rootContext = CosNaming::NamingContext::_narrow(obj.in(), ACE_TRY_ENV);
    ACE_TRY_CHECK; 
  } ACE_CATCHANY {
    ACE_ERROR ((LM_ERROR, "error while resolving initial reference to name server\n"));
    return -1;
  }
  ACE_ENDTRY;
  return 0;
}

/*
 * Based on the given compoundName, creates a chain of contexts in the Name
 * Service. If some of the contexts already exists, it continues without any
 * warning. If successful, returns 0, otherwise, returs -1.
 */
int 
NSBroker::createChainOfContexts(CosNaming::Name &name)
{
  CORBA::ULong length = name.length();
  CosNaming::Name query_name(length);
  
  for (CORBA::ULong i = 0; i < length; i++) {
    query_name.length(i+1);
    query_name[i].id = name[i].id; // do I need to call CORBA::string_dup ?
    
    ACE_TRY_NEW_ENV {
      _rootContext->bind_new_context(query_name, ACE_TRY_ENV);
    } ACE_CATCH (CosNaming::NamingContext::AlreadyBound, varName) {
      // It was already bound. So, that's OK we don't need to do anything.
    } ACE_CATCHANY {
      ACE_ERROR ((LM_ERROR, "error while creating context <%s>\n", 
                  (const char *)query_name[i].id));
      return -1;
    }
    ACE_ENDTRY;
  }
  return 0;
}

/*
 * Bind a name to the root context.  If it already exists, we don't fail.
 * Returns 0 on success, -1 on failure.
 */
int
NSBroker::bind(CosNaming::Name &name, CORBA::Object_ptr obj)
{
  CORBA::Environment ACE_TRY_ENV;
  ACE_TRY_NEW_ENV {
    _rootContext->rebind(name, obj, ACE_TRY_ENV);
  } ACE_CATCHANY {
    ACE_ERROR ((LM_ERROR, "error while binding name <%s>\n", 
                (const char *)name[name.length()-1].id));
    return -1;
  }
  ACE_ENDTRY;
  return 0;
}

/*
 * Unbind a name or context.
 * Returns 0 on success, -1 on failure.
 */
int
NSBroker::unbind(CosNaming::Name &name)
{
  CORBA::Environment ACE_TRY_ENV;
  ACE_TRY_NEW_ENV {
    _rootContext->unbind(name, ACE_TRY_ENV);
  } ACE_CATCHANY {
    ACE_ERROR ((LM_ERROR, "error while unbinding name <%s>\n", 
                (const char *)name[name.length()-1].id));
    return -1;
  }
  ACE_ENDTRY;
  return 0;
}

/*
 * Resolve a name from the root context.  Returns the resolved object on
 * success or a null object on failure.
 */
CORBA::Object_ptr
NSBroker::resolve(CosNaming::Name &name)
{
  CORBA::Environment ACE_TRY_ENV;
  CORBA::Object_ptr object = NULL;
  ACE_TRY_NEW_ENV {
    object = _rootContext->resolve(name, ACE_TRY_ENV);
  } ACE_CATCHANY {
    ACE_ERROR ((LM_ERROR, "error while resolving name <%s>\n", 
                (const char *)name[name.length()-1].id));
  }
  ACE_ENDTRY;
  return object;
}

/*
 * Returns a pointer to the Name Service root. The init () function must have been
 * called before.
 */
CosNaming::NamingContext_ptr 
NSBroker::getRootContext ()
{
  return _rootContext;
}
