/******************************************************************************
*                                                                             *
*  File: ComponentConfigurator.h                                              *
*                                                                             *
*  Description: An abstract class for representing the dependencies between   *
*               components.                                                   *
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************

   Revision Control Information

   $Revision: 1.1.1.2 $

   $Log: ComponentConfigurator.h,v $
   Revision 1.1.1.2  2000/09/30 15:49:31  yamane
   Patches for Windows

   Revision 1.1.1.1  1999/10/12 00:43:29  f-kon
   Version compatible with TAO 1.0

   Revision 1.1.1.1  1999/04/01 18:16:36  f-kon
   This new 2K source tree contains the most up-to-date version of ACE and TAO
   and including the dynamicTAO code as well as some 2K tests and preliminary
   versions of some of its high-level systems.


   Revision 1.4  1999/03/03 16:11:52  f-kon
   Implemented DCP operation configure_implementation <loaded_implementation> <configuration_string>

   Revision 1.3  1999/02/22 22:20:39  f-kon
   Introduced NiceConfigurator and LockingConfigurator and performed several
   fixes on dynamic reconfiguration.

   Revision 1.2  1999/01/26 18:22:12  mroman
   DomainConfigurator running on Windows NT

   Revision 1.1  1999/01/07 22:43:17  f-kon
   Added TAOConfigurator and new DCP commands

******************************************************************************/

#if !defined (COMPONENT_CONFIGURATOR_H)
#define COMPONENT_CONFIGURATOR_H

#include "ComponentEvent.h"
#include "ace/Synch.h"

class ACE_Service_Object;
typedef ACE_Service_Object Object; 

class SIMPLECONF_2K_Export ComponentConfigurator {
  public:
    inline virtual ~ComponentConfigurator () { };

    /* Deletes the references to all internal data structures. */
    virtual void destroyComponentConfigurator () = 0;
        
    /* Adds a hook to the configurator.
     * @param   hookName  the name of the hook to be added.
     */
    virtual int addHook (const char *hookName) = 0;
    
    /* Deletes a hook from the configurator.
     * @param   hookName  the name of the hook to be deleted.
     */
    virtual int deleteHook (const char *hookName) = 0;
        
    /* Attaches a ComponentConfigurator to the given hook.
     * @param  hookName  the name of the hook.
     * @param  cc        the ComponentConfigurator to be attached to the hook.
     */
    virtual int hook (const char *hookName, ComponentConfigurator *cc) = 0;

    /* Detaches a ComponentConfigurator from the given hook.
     * @param hookName  the name of the hook.
     * @param cc        the ComponentConfigurator to be detached.
     */
    virtual int unhook (const char *hookName) = 0;
    
    /* Adds a new component to the list of clients.
     * @param  client  the ComponentConfigurator representing the new client.
     * @param  hookNameInClient the name of the hook to which 
     *         this component is attached.
     */
    virtual int registerClient (ComponentConfigurator *client, 
                                const char *hookNameInClient) = 0;
    
    /* Removes a pair (client, hook) from the list of clients.
     * @param  client  the ComponentConfigurator representing the new client.
     * @param  hookNameInClient the name of the hook to which 
     *         this component is attached.
     * @param  wasRemoved is an output parameter that specifies if this component
     *         was removed as a result from unregistering the client. This
     *         may happen if garbage collection is being carried out.  
     */
    virtual int unregisterClient (ComponentConfigurator *client, 
                                  const char *hookNameInClient, 
                                  int *wasRemoved = 0) = 0;
    
    /* This implementation simply prints the event to the standard output.
     * @param  hookedComponent the ComponentConfigurator that generated the event.
     * @param  e               the event that was generated. 
     */
    virtual void eventOnHookedComponent (const ComponentConfigurator *hookedComponent,
                                         const ComponentEvent *e) = 0;
    
    /* This implementation simply prints the event to the standard output.
     * @param  client  the ComponentConfigurator that generated the event.
     * @param  e       the event that was generated. 
     */
    virtual void eventOnClient (const ComponentConfigurator *client,
                                const ComponentEvent *e) = 0;
    
    /* Accessor for the component name.
     * @return a string containing the component name. 
     */
    virtual const char *name () const = 0;
    
    /* Accessor for the component information string.
     * @return a string containing component information.
     */
    virtual const char *info () const = 0;
    
    /* Provides a list of the hooks in this component.
     * @returns a list of DependencySpecifications.
     */
    virtual /*MANUELconst*/ DependencyList *listHooks () = 0; 
    
    /* Provides a list of the clients of this component.
     * @returns a vector of DependencySpecifications.
     */
    virtual ComponentConfigurator *getHookedComponent(const char *hookName) = 0;

    /* Provides the number of clients of this component. 
     * May be used for garbage collection.
     * @returns the size of the list of clients.
     */

    virtual int numberOfClients () = 0;

    /* Provides a list of the clients of this component.
     * @returns a vector of HookSpecifications.
     */
    virtual const DependencyList *listClients () = 0;
    
    /* Sets the name of this component.
     */
    virtual void name (const char *s) = 0;
    
    /* Sets the information string for this component.
     */
    virtual void info (const char *s) = 0;
    
    /* Accessor for getting the component implementation.
     * @return the object implementation.
     */
    virtual Object *implementation () = 0;

    /* returns a pointer to our lock, if we have one, or 0 o/w */
    virtual inline ACE_RW_Mutex *lock () { return 0; }

    /* Specifies that the configurator can be removed when the number of clients reaches 0. */
    virtual void toBeRemoved () = 0;
};

#endif
