/******************************************************************************
*                                                                             *
*  File: LockingConfigurator.h                                                *
*                                                                             *
*  Description: 
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************
   Revision Control Information
   $Revision: 1.1.1.2 $
   $Log: LockingConfigurator.h,v $
   Revision 1.1.1.2  2000/09/30 15:58:02  yamane
   Patches for Windows

   Revision 1.1.1.1  1999/10/12 00:43:29  f-kon
   Version compatible with TAO 1.0

   Revision 1.1.1.1  1999/04/01 18:16:36  f-kon
   This new 2K source tree contains the most up-to-date version of ACE and TAO
   and including the dynamicTAO code as well as some 2K tests and preliminary
   versions of some of its high-level systems.


   Revision 1.2  1999/03/03 16:11:54  f-kon
   Implemented DCP operation configure_implementation <loaded_implementation> <configuration_string>

   Revision 1.1  1999/02/22 22:22:43  f-kon
   Introducing NiceConfigurator and LockingConfigurator.

******************************************************************************/

#if !defined (LOCKING_CONFIGURATOR_H)
#define       LOCKING_CONFIGURATOR_H

#include "NiceConfigurator.h"
#include "ace/Synch.h"

//
// class SIMPLECONF_2K_Export LockingConfigurator : public NiceConfigurator {
//
// We don't have any non-inline function in this class.
// Therefore we don't need SIMPLECONF_2K_Export declaration here.
// - Tomonori Yamane
//
class LockingConfigurator : public NiceConfigurator {
private:
  ACE_RW_Mutex lock_;               // controls the access to the dependencies
  ACE_RW_Mutex implementationLock_; // controls the access to the component implementation
  
public:
  /* Creates a new configurator for a given object.
   *
   * @param     name             the component name.
   * @param     implementation   the object implementation.
   */
  inline LockingConfigurator (const char *name, Object *implementation)
    : NiceConfigurator (name, implementation)
    {
    }
  
  // returns a pointer to our locks.
  inline ACE_RW_Mutex *lock () { return &lock_; }
  inline ACE_RW_Mutex *implementationLock () { return &implementationLock_; }
  
  /* Deletes the references to all internal data structures. */
  inline void destroyComponentConfigurator ()
    {
      lock_.acquire_write();
      NiceConfigurator::destroyComponentConfigurator();
      lock_.release ();
    }
  
  /* Adds a hook to the configurator.
   * @param   hookName  the name of the hook to be added.
   */
  inline int addHook (const char *hookName)
    {
      lock_.acquire_write();
      int ret = NiceConfigurator::addHook(hookName);
      lock_.release ();
      return ret;
    }   

  /* Deletes a hook from the configurator.
   * @param   hookName  the name of the hook to be deleted.
   */
  inline int deleteHook (const char *hookName)
    {
      lock_.acquire_write();
      int ret = NiceConfigurator::deleteHook(hookName);
      lock_.release ();
      return ret;
    }   

  /* Attaches a ComponentConfigurator to the given hook.
   * @param  hookName  the name of the hook.
   * @param  cc        the ComponentConfigurator to be attached to the hook.
   */
  int hook (const char *hookName, ComponentConfigurator *cc)
    {
      //ACE_DEBUG ((LM_DEBUG, "hook: acquiring lock\n"));
      lock_.acquire_write();
      //ACE_DEBUG ((LM_DEBUG, "hook: lock acquired\n"));
      int ret = NiceConfigurator::hook(hookName, cc);
      //ACE_DEBUG ((LM_DEBUG, "hook: releasing lock\n"));
      lock_.release ();
      //ACE_DEBUG ((LM_DEBUG, "hook: lock released\n"));
      return ret;
    }  


  /* Detaches a ComponentConfigurator from the given hook.
   * @param hookName  the name of the hook.
   * @param cc        the ComponentConfigurator to be detached.
   */
  int unhook (const char *hookName)
    {
      //ACE_DEBUG ((LM_DEBUG, "unhook: acquiring lock\n"));
      lock_.acquire_write();
      //ACE_DEBUG ((LM_DEBUG, "unhook: lock acquired\n"));
      int ret = NiceConfigurator::unhook(hookName);
      //ACE_DEBUG ((LM_DEBUG, "unhook: releasing lock\n"));
      lock_.release ();
      //ACE_DEBUG ((LM_DEBUG, "unhook: lock released\n"));
      return ret;
    }

  /*
   * Adds a new component to the list of clients.
   * @param  client  the ComponentConfigurator representing the new client.
   * @param  hookNameInClient the name of the hook to which
   *          this component is attached.
   */
  int registerClient (ComponentConfigurator *client, 
                             const char *hookNameInClient)
    {
      //ACE_DEBUG ((LM_DEBUG, "registeringClient: acquiring lock\n"));
      lock_.acquire_write();
      //ACE_DEBUG ((LM_DEBUG, "registeringClient: lock acquired\n"));
      int ret = NiceConfigurator::registerClient(client, hookNameInClient);
      //ACE_DEBUG ((LM_DEBUG, "registeringClient: releasing lock\n"));
      lock_.release ();
      //ACE_DEBUG ((LM_DEBUG, "registeringClient: lock released\n"));
      return ret;
    }
    
  /* Removes a pair (client, hook) from the list of clients.
   * @param  client  the ComponentConfigurator representing the new client.
   * @param  hookNameInClient the name of the hook to which 
   *         this component is attached.
   */
  int unregisterClient (ComponentConfigurator *client, 
                               const char *hookNameInClient, 
                               int *wasRemoved = 0)
    {
      //ACE_DEBUG ((LM_DEBUG, "unregisteringClient: acquiring lock\n"));
      lock_.acquire_write();
      //ACE_DEBUG ((LM_DEBUG, "unregisteringClient: lock acquired\n"));
      int ret = NiceConfigurator::unregisterClient(client, hookNameInClient, wasRemoved);
      //ACE_DEBUG ((LM_DEBUG, "unregisteringClient: releasing lock\n"));
      lock_.release ();
      //ACE_DEBUG ((LM_DEBUG, "unregisteringClient: lock released\n"));
      return ret;
    }
   
  /* This implementation simply prints the event to the standard output.
   * @param  hookedComponent the ComponentConfigurator that generated the event.
   * @param  e               the event that was generated. 
   */
  inline void eventOnHookedComponent (const ComponentConfigurator *hookedComponent,
                                      const ComponentEvent *e)
    {
      // Applications could override this to treat different kinds of events
      // as write events or read events or events that do not require locking.
      lock_.acquire_write();
      NiceConfigurator::eventOnHookedComponent (hookedComponent, e);
      lock_.release ();
    }
    
  /* This implementation simply prints the event to the standard output.
   * @param  client  the ComponentConfigurator that generated the event.
   * @param  e       the event that was generated. 
   */
  inline void eventOnClient (const ComponentConfigurator *client, 
                             const ComponentEvent *e)
    {
      // Applications could override this to treat different kinds of events
      // as write events or read events or events that do not require locking.
      lock_.acquire_read ();
      NiceConfigurator::eventOnClient (client, e);
      lock_.release ();
    }

  /* Accessor for the component name.
   * @return a string containing the component name. 
   */
  //const char *name () const;
  
  /* Accessor for the component information string.
   * @return a string containing component information.
   */
  //const char *info () const;
  
  /*
   * Provides a list of the hooks in this component.
   * @returns a list of DependencySpecifications.
   */
  inline DependencyList *listHooks ()
    {
      lock_.acquire_read();
      DependencyList *ret = NiceConfigurator::listHooks();
      lock_.release ();
      return ret;
    }

  /* Provides a list of the clients of this component.
   * @returns a list of DependencySpecifications.
   */
  inline ComponentConfigurator *getHookedComponent (const char *hookName)
    {
      lock_.acquire_read ();
      ComponentConfigurator *ret = NiceConfigurator::getHookedComponent (hookName);
      lock_.release ();
      return ret;
    }

  /* Provides a list of the clients of this component.
   * @returns a list of DependencySpecifications.
   */
  inline const DependencyList *listClients ()
    {
      lock_.acquire_read ();
      const DependencyList *ret = NiceConfigurator::listClients ();
      lock_.release ();
      return ret;
    }
  
  /* Provides the number of clients of this component. 
   * May be used for garbage collection.
   * @returns the size of the list of clients.
   */
  inline int numberOfClients ()
    {
      lock_.acquire_read ();
      int ret = NiceConfigurator::numberOfClients ();
      lock_.release ();
      return ret;
    }

  /* Accessor for getting the component implementation.
   * @return the object implementation.
   */
  inline Object *implementation ()
    {
      lock_.acquire_read ();
      Object *ret = NiceConfigurator::implementation ();
      lock_.release ();
      return ret;
    }

//   /* Accessor for getting the component implementation and locking it to make
//    * sure it won't change until the releaseLockedImplementation () method is
//    * called.  
//    * @return the object implementation.  
//    */
//   inline Object *lockedImplementation ()
//     {
//       // Is it possible to add a timeout in the validity of the lock???
//       implementationLock_.acquire_read ();
//       Object *ret = NiceConfigurator::implementation ();
//       implementationLock_.release ();
//       return ret;
//     }

//   /* Releases the lock in the component implementation 
//    */
//   inline void releaseLockedImplementation ()
//     {
//       implementationLock_.release ();
//     }
  
  /* Accessor for setting the component implementation. 
   */
  void implementation (Object *implementation)
    {
      lock_.acquire_write ();
      NiceConfigurator::implementation (implementation);
      lock_.release ();
    }
};

#endif
