/******************************************************************************
*                                                                             *
*  File: SimpleConfigurator.h                                                 *
*                                                                             *
*  Description: A concrete class that reifies dependencies between components.*
*                                                                             *
*  Project: 2K - Automatic Configuration                                      *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************

   Revision Control Information

   $Revision: 1.1.1.2 $

   $Log: SimpleConfigurator.h,v $
   Revision 1.1.1.2  2000/09/30 16:02:14  yamane
   Patches for Windows

   Revision 1.1.1.1  1999/10/12 00:43:29  f-kon
   Version compatible with TAO 1.0

   Revision 1.1.1.1  1999/04/01 18:16:36  f-kon
   This new 2K source tree contains the most up-to-date version of ACE and TAO
   and including the dynamicTAO code as well as some 2K tests and preliminary
   versions of some of its high-level systems.


   Revision 1.5  1999/03/03 16:11:59  f-kon
   Implemented DCP operation configure_implementation <loaded_implementation> <configuration_string>

   Revision 1.4  1999/02/22 22:20:43  f-kon
   Introduced NiceConfigurator and LockingConfigurator and performed several
   fixes on dynamic reconfiguration.

   Revision 1.3  1999/02/19 20:03:24  f-kon
   *** empty log message ***

   Revision 1.2  1999/01/26 18:22:18  mroman
   DomainConfigurator running on Windows NT

   Revision 1.1  1999/01/07 22:43:18  f-kon
   Added TAOConfigurator and new DCP commands

******************************************************************************/

#if !defined (SIMPLE_CONFIGURATOR_H)
#define       SIMPLE_CONFIGURATOR_H

#include "ace/OS.h"
#include "ace/Service_Object.h"

#include "DependencySpecification.h"

/////////////////////////////////

#include "ComponentConfigurator.h"
#include "SimpleConfigurator_export.h"

class SIMPLECONF_2K_Export SimpleConfigurator : public ComponentConfigurator {
  private:
    /* The implementating this component. */
    Object *implementation_;

    /* A list containing the clients for this component.  */
    DependencyList clientList_;

    /* The name of this component. */
    char *name_;

    /* Information string about this component. */
    char *info_;

    /* Flag specifying if this configurator should be removed when the number of
       clients reach 0.
    */
    int toBeRemoved_;

  protected:
    /* A list containing the hooks for this component. */
    DependencyList hookList_;

  public:
    /* Creates a new configurator for a given object.
     *
     * @param     name             the component name.
     * @param     implementation   the object implementation.
     */
    SimpleConfigurator (const char *name, Object *implementation);

    virtual ~SimpleConfigurator ();
    
    /* Deletes the references to all internal data structures. */
    virtual void destroyComponentConfigurator ();

    /* Adds a hook to the configurator.
     * @param   hookName  the name of the hook to be added.
     */
    virtual int addHook (const char *hookName);
        
    /* Deletes a hook from the configurator.
     * @param   hookName  the name of the hook to be deleted.
     */
    virtual int deleteHook (const char *hookName);
        
    /* Attaches a ComponentConfigurator to the given hook.
     * @param  hookName  the name of the hook.
     * @param  cc        the ComponentConfigurator to be attached to the hook.
     */
    virtual int hook (const char *hookName, ComponentConfigurator *cc);

    /* Detaches a ComponentConfigurator from the given hook.
     * @param hookName  the name of the hook.
     * @param cc        the ComponentConfigurator to be detached.
     */
    virtual int unhook (const char *hookName);
    
    /*
     * Adds a new component to the list of clients.
     * @param  client  the ComponentConfigurator representing the new client.
     * @param  hookNameInClient the name of the hook to which
     *          this component is attached.
     */
    virtual int registerClient (ComponentConfigurator *client, 
                                const char *hookNameInClient);
    
    /* Removes a pair (client, hook) from the list of clients.
     * @param  client  the ComponentConfigurator representing the new client.
     * @param  hookNameInClient the name of the hook to which 
     *         this component is attached.
     */
    virtual int unregisterClient (ComponentConfigurator *client, 
                                  const char *hookNameInClient,
                                  int *wasRemoved = 0);
    
    /* This implementation simply prints the event to the standard output.
     * @param  hookedComponent the ComponentConfigurator that generated the event.
     * @param  e               the event that was generated. 
     */
    virtual void eventOnHookedComponent (const ComponentConfigurator *hookedComponent,
                                         const ComponentEvent *e);
    
    /* This implementation simply prints the event to the standard output.
     * @param  client  the ComponentConfigurator that generated the event.
     * @param  e       the event that was generated. 
     */
    virtual void eventOnClient (const ComponentConfigurator *client, 
                        const ComponentEvent *e);
    
    /* Accessor for the component name.
     * @return a string containing the component name. 
     */
    virtual const char *name () const;
    
    /* Accessor for the component information string.
     * @return a string containing component information.
     */
    virtual const char *info () const;
    
    /*
     * Provides a list of the hooks in this component.
     * @returns a list of DependencySpecifications.
     */
    virtual DependencyList *listHooks ();

    /* Provides a list of the clients of this component.
     * @returns a list of DependencySpecifications.
     */
    virtual ComponentConfigurator *getHookedComponent(const char *hookName);

    /* Provides a list of the clients of this component.
     * @returns a list of DependencySpecifications.
     */
    virtual const DependencyList *listClients ();
    
    /* Provides the number of clients of this component. 
     * May be used for garbage collection.
     * @returns the size of the list of clients.
     */
    virtual int numberOfClients ();

    /* Sets the name of this component. The string is copied*/
    virtual void name (const char *s);
    
    /* Sets the information string for this component. The string is copied*/
    virtual void info (const char *s);

    /* Accessor for getting the component implementation.
     * @return the object implementation.
     */
    virtual Object *implementation ();

    /* Accessor for setting the component implementation. */
    virtual void implementation (Object *implementation);

    /* Specifies that the configurator can be removed when the number of clients reaches 0.*/
    virtual void toBeRemoved ();

  protected:
    //////////////////////////////////////////////////////////////////////
    // Returns the index of the first hook with the name provided.
    // Returns -1 if such element is not found.
    //////////////////////////////////////////////////////////////////////
    DependencySpecification *findHook (const char *hookName);

    ////////////////////////////////////////////////////////////////////// 
    // Returns a reference to the first client hook with the
    // ComponentConfigurator and name provided.  
    // Returns 0 if such element is not found.
    //////////////////////////////////////////////////////////////////////
    DependencySpecification *findClientHook (const ComponentConfigurator *client, 
                                       const char *hookName);
};

#endif
