/******************************************************************************
*                                                                             *
*  File: Simple_Component_Factory.cpp                                         *
*                                                                             *
*  Description: A wrapper for 2K components that can be dynamically loaded.   *
*               This components are implemented as DLLs and they contain two  *
*               ComponentConfigurators (a local one that represents the       *
*               dependencies inside the process and a CORBA-based distributed *
*               one.                                                          *
*                                                                             *
*  Note: This class does not initialize the CORBA ComponentConfigurator, if   *
*        you need this initialization, use Component_Factory instead.         *
*                                                                             *
*  Project: 2K - Automatic Configuration.                                     *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************
   Revision Control Information
   $Revision: 1.2 $
   $Log: Simple_Component_Factory.cpp,v $
   Revision 1.2  2000/10/09 21:03:08  f-kon
   Changes to use new TwoK:: namespace for 2K services integrated with Tomo's
   changes for Windows compilation.

   Revision 1.1.1.2  1999/10/15 22:14:16  yamane
   Clean up for native exception support

   Revision 1.1.1.1  1999/10/12 00:43:30  f-kon
   Version compatible with TAO 1.0

******************************************************************************/

#include "Simple_Component_Factory.h"

Simple_Component_Factory::Simple_Component_Factory ()
  //  : thisTypeHasCorbaConfigurator_ (1) WHY THIS DOESN'T COMPILE???
{
  thisTypeHasCorbaConfigurator_ = 1;
}

int
Simple_Component_Factory::init (int argc, char *argv[])
{
  ACE_TRACE ("Simple_Component_Factory::init");
  
  localConfigurator_ = new SimpleConfigurator (0, this);
  
  // This class does not initialize the CORBA ComponentConfigurator, if
  // you need this initialization, use Component_Factory instead. 
  corbaConfigurator_ = TwoK::Configuration::ComponentConfigurator::_nil ();

  return 0;
}// init


int
Simple_Component_Factory::fini (void)
{
  ACE_TRACE ("Simple_Component_Factory::fini");
  // Calls the destructor on its local configurator. The configurator then may send
  // events to its clients.
  if (localConfigurator_ != 0)
    {
      delete localConfigurator_;
      localConfigurator_ = 0;
    }

  if (!CORBA::is_nil (corbaConfigurator_.in ()))
    {
      ACE_TRY_NEW_ENV
        {
          corbaConfigurator_->destroy(ACE_TRY_ENV);
          ACE_TRY_CHECK;
        }
      ACE_CATCHANY
        {
          ACE_ANY_EXCEPTION._tao_print_exception ("Could not destroy configurator");
          return -1;
        }
      ACE_ENDTRY;
    }
  
  return 0;
}// fini


////////////////////////////////////////////////////////////////////////////////
// It is called after the prerequisites of the component are resolved.
// At this time, its dependencies, if any, were attached to its CORBA
// ComponentConfigurator hooks.
////////////////////////////////////////////////////////////////////////////////
int
Simple_Component_Factory::start (const TwoK::Configuration::ComponentConfigurator_var &compConfig)
{
  ACE_TRACE ("Simple_Component_Factory::start");

#define DEBUG
#ifdef DEBUG
  if (!CORBA::is_nil (corbaConfigurator ().in()) && 
      !compConfig->_is_equivalent (corbaConfigurator ()))
        {
          ACE_DEBUG ((LM_ERROR, "(%P|%t) Internal error in Simple_Component_Factory::start.\n"));
          return -1;
        }
#endif

  // set the given ComponentConfigurator as our corbaConfigurator. 
  if (!CORBA::is_nil(compConfig.in ()))
    corbaConfigurator (compConfig);

  // Subclasses may now look at the hooked components on this configurator to
  // perform some initializations and start its job.

  return 0;
}// start ()



void *Simple_Component_Factory::component_implementation (void)
{
  return 0;
}

void *
Simple_Component_Factory::strategy (void)
{
  return component_implementation (); 
  // Note that component_implementation is virtual and can be customized by subclasses.
}


// localConfigurator_ accessor methods
ComponentConfigurator *
Simple_Component_Factory::localConfigurator (void) 
{
  return localConfigurator_;
}

void 
Simple_Component_Factory::localConfigurator (ComponentConfigurator *c) 
{
  localConfigurator_ = c;
}

// corbaConfigurator_ accessor methods
TwoK::Configuration::ComponentConfigurator_var
Simple_Component_Factory::corbaConfigurator (void) 
{
  return TwoK::Configuration::ComponentConfigurator::_duplicate (corbaConfigurator_);
}

void 
Simple_Component_Factory::corbaConfigurator (const TwoK::Configuration::ComponentConfigurator_var &c) 
{
  corbaConfigurator_ = c; 
}


