////////////////////////////////////////////////////////
//Auhor: Manuel Roman. Based on the ACE_Thread_Strategy/
////////////////////////////////////////////////////////
//Date: 9-29-1998                                      /
////////////////////////////////////////////////////////

// = DESCRIPTION
//		Creates a pool of threads when the open method is called. The number of threads is specified by the user.
//		a water-mark will be used, in such a way that when the number of threads in the pool goes beyond the mark,
//		new threads will be created.
//		Threads, when spawned, will execute a dummy function. This function waits on a semaphore, and once it starts
//		executing it gets the pointer to the actual SVC_HANDLER and executes its svc method.




////////////////////////////////////////////////////////////////////////////////////////////////////////
///////////////////////// ACE THREAD POOL STRATEGY /////////////////////////////////////////////////////
////////////////////////////////////////////////////////////////////////////////////////////////////////
#if !defined(_THREAD__POOL_H_)
#define _THREAD__POOL_H_

#include "ace/Strategies.h"
#include "svchandler_list.h"

#define MAX_HIGH_WATER_MARK	-1

template <class SVC_HANDLER>
class ACE_Thread_Pool_Strategy : public ACE_Concurrency_Strategy<SVC_HANDLER>
{
  // = TITLE
  //     Defines the interface for specifying a concurrency strategy
  //     for a <SVC_HANDLER> based on a thread pool.
  //
  // = DESCRIPTION
  //     This class provides a strategy that manages the creation of
  //     threads to handle requests from clients concurrently.  It
  //     uses a thread pool, so there is no overhead for creating a new thread
  //	 each time a connection from the client is established

public:
  // = Intialization and termination methods.
  ACE_Thread_Pool_Strategy (int flags = 0);
  // "Do-nothing constructor"

  ACE_Thread_Pool_Strategy (ACE_Thread_Manager *tm,
		       long thr_flags,
		       size_t n_threads = 1,
			   long low_water_mark = 0,
			   long high_water_mark = MAX_HIGH_WATER_MARK,
               int flags = 0);
  // Initialize the strategy.

  virtual int open (ACE_Thread_Manager *tm,
		    long thr_flags,
		    size_t n_threads = 1,
			long low_water_mark = 0,
			long high_water_mark = MAX_HIGH_WATER_MARK,
            int flags = 0);
  // Initialize the strategy.

  virtual ~ACE_Thread_Pool_Strategy (void);

  // = Factory method.
  virtual int activate_svc_handler (SVC_HANDLER *svc_handler,
				    void *arg = 0);
  // Activate the <svc_handler> with an appropriate concurrency
  // strategy.  This method activates the SVC_HANDLER by first calling
  // its open() method, assigning the pointer of the SVC_HANDLER to
  // the Current_SVC_HANDLER attribute and then calling signal on the shared
  // semaphore.

  void dump (void) const;
  // Dump the state of an object.

  static SvcHandler_List svc_handler_list_;
  //static CLinkedList svc_handler_list_;

  static ACE_Double_Linked_List_Iterator<SvcHandler_Record> *thp_li_; 
  //Iterator for the list.

  static ACE_Semaphore *sem_;
  //Semaphore used to synchronized the pool of threads.

  //Used to protect the variable that keeps track of the number of threads available.
  static ACE_Mutex mutex_;
  
  static void entry_point();

  static int available_threads_;

  //Number of incoming requests waiting to be managed.
  static int waiting_requests_;

  static long low_water_mark_;
  //When the number of threads reaches this value, new threads must be spawned

  static long high_water_mark_;
  //Fixes the maximum amount of threads in the pool. If needed, threads will be deleted.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.

protected:
  typedef ACE_Concurrency_Strategy<SVC_HANDLER> inherited;

  ACE_Thread_Manager *thr_mgr_;
  // Thread manager for this class (must be provided).

  long thr_flags_;
  // Flags to pass into the SVC_HANDLER::activate() method.

  size_t n_threads_;
  // Number of threads to spawn.
};


//All threads must be blocked until an incoming request arrives.
template<class SVC_HANDLER> ACE_Semaphore *
ACE_Thread_Pool_Strategy<SVC_HANDLER>::sem_=0;


template<class SVC_HANDLER> ACE_Mutex
ACE_Thread_Pool_Strategy<SVC_HANDLER>::mutex_;


template<class SVC_HANDLER> int
ACE_Thread_Pool_Strategy<SVC_HANDLER>::available_threads_=0;

template<class SVC_HANDLER> int
ACE_Thread_Pool_Strategy<SVC_HANDLER>::waiting_requests_=0;


template<class SVC_HANDLER> SvcHandler_List
ACE_Thread_Pool_Strategy<SVC_HANDLER>::svc_handler_list_;


template<class SVC_HANDLER> ACE_Double_Linked_List_Iterator<SvcHandler_Record> *
ACE_Thread_Pool_Strategy<SVC_HANDLER>::thp_li_=0;

template<class SVC_HANDLER> long
ACE_Thread_Pool_Strategy<SVC_HANDLER>::low_water_mark_=0;


template<class SVC_HANDLER> long
ACE_Thread_Pool_Strategy<SVC_HANDLER>::high_water_mark_=MAX_HIGH_WATER_MARK;


ACE_ALLOC_HOOK_DEFINE(ACE_Thread_Pool_Strategy)

template<class SVC_HANDLER> void 
ACE_Thread_Pool_Strategy<SVC_HANDLER>::entry_point()
{
	SVC_HANDLER *temp_svc_hdlr=0;
	SvcHandler_Record *tempNode;

	//ATTENTION: WE SHOULD HAVE REGISTERED THE THREAD WITH AN EXITING FUNCTION. LOOK AT TASK.CPP
	//NO. I just call the exit() method from the thread, but before that I call close on the handler
	//served by the thread, in order to close network connections used bu the SVC_HANDLER.

	while(1)
	{
		//Barrier semaphore
		ACE_Thread_Pool_Strategy::sem_->acquire();
		
		//The semaphore can be released due to two possibilities:
		//1. There is a svc_handler that must be executed
		//2. Thre high_water_mark has been changed, so some threads must be eliminated.
		//
		//In the last case, the semaphore's release will be called as many times as threads need to be deleted.
		//It will be known that we are in the last case because the list of service handlers will be empty.
		
		///////////////////////////////////////////////////////////////////////
		ACE_Thread_Pool_Strategy::mutex_.acquire();
		
		if(ACE_Thread_Pool_Strategy::svc_handler_list_.is_empty())
		{
			//High_Water_Mark has been changed.
			//Thread exiting
			ACE_Thread_Pool_Strategy::mutex_.release();	

			//After servicing the handler, let's call the close method on the ACE_SVC_HANDLER in order to unregister
			//it from the reactor.
			if(temp_svc_hdlr!=0)
				temp_svc_hdlr->close();

			ACE_Thread_Manager::instance()->exit(0);
		}
		else
		{
			ACE_Thread_Pool_Strategy::thp_li_->first();
			tempNode=/*(SvcHandlerNode*)*/ACE_Thread_Pool_Strategy::thp_li_->next();
			temp_svc_hdlr=tempNode->get_sh();
			ACE_Thread_Pool_Strategy::svc_handler_list_.delete_head();
			ACE_Thread_Pool_Strategy::mutex_.release();
		}
		///////////////////////////////////////////////////////////////////////
		
		temp_svc_hdlr->svc();

		//Check high water mark
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
		ACE_Thread_Pool_Strategy::mutex_.acquire();

		if( ACE_Thread_Pool_Strategy<SVC_HANDLER>::available_threads_ < ACE_Thread_Pool_Strategy<SVC_HANDLER>::high_water_mark_)
		{
			//If there are pending requests, don't increment waiting_requests and release the semaphore so this thread will 
			//serve the incoming request.
			if(ACE_Thread_Pool_Strategy::waiting_requests_==0)
			{
				ACE_Thread_Pool_Strategy<SVC_HANDLER>::available_threads_++;
			}
			else
			{
				ACE_Thread_Pool_Strategy::waiting_requests_--;
				//ACE_Thread_Pool_Strategy::sem_->release();
			}
		}
		else
		{
			//The thread must we killed
			ACE_Thread_Pool_Strategy::mutex_.release();	
		
			//After servicing the handler, let's call the close method on the ACE_SVC_HANDLER in order to unregister
			//it from the reactor.
			temp_svc_hdlr->close();
			ACE_Thread_Manager::instance()->exit(0);
		}
		ACE_Thread_Pool_Strategy::mutex_.release();
		/////////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	}
}


template <class SVC_HANDLER> void
ACE_Thread_Pool_Strategy<SVC_HANDLER>::dump (void) const
{
  ACE_TRACE ("ACE_Thread_Pool_Strategy<SVC_HANDLER>::dump");
}

template <class SVC_HANDLER> int
ACE_Thread_Pool_Strategy<SVC_HANDLER>::open (ACE_Thread_Manager *thr_mgr,
					     long thr_flags,
					     size_t n_threads,
					     long low_water_mark,
					     long high_water_mark,
					     int flags)
{
  ACE_TRACE ("ACE_Thread_Pool_Strategy<SVC_HANDLER>::open");
  this->thr_mgr_ = thr_mgr;
  this->n_threads_ = n_threads;
  this->thr_flags_ = thr_flags;
  this->flags_ = flags;
  ACE_Thread_Pool_Strategy::low_water_mark_  = low_water_mark;
  ACE_Thread_Pool_Strategy::high_water_mark_ = high_water_mark;
  ACE_Thread_Pool_Strategy::thp_li_ = 
    new ACE_Double_Linked_List_Iterator<SvcHandler_Record>
          (ACE_Thread_Pool_Strategy::svc_handler_list_);

  // Must have a thread manager!
  if (this->thr_mgr_ == 0)
    ACE_ERROR_RETURN ((LM_ERROR,
                       "error: must have a non-NULL thread manager\n"), -1);
  else
  {
	//Don't need to protect with mutex, since threads have not been created yet!!
	ACE_Thread_Pool_Strategy::available_threads_=n_threads;
	ACE_Thread_Pool_Strategy::waiting_requests_=0;
	  
	//Create the pool of threads.
	this->thr_mgr_->spawn_n (n_threads,ACE_THR_FUNC (&ACE_Thread_Pool_Strategy::entry_point));
  }
    return 0;
}


template <class SVC_HANDLER>
ACE_Thread_Pool_Strategy<SVC_HANDLER>::ACE_Thread_Pool_Strategy (ACE_Thread_Manager *thr_mgr,
                                                       long thr_flags,
                                                       size_t n_threads,
													   long low_water_mark,
													   long high_water_mark,
                                                       int flags)
{
  ACE_TRACE ("ACE_Thread_Strategy<SVC_HANDLER>::ACE_Thread_Strategy");

  
  ACE_Thread_Pool_Strategy::sem_=new ACE_Semaphore(0);	

  if (this->open (thr_mgr, thr_flags, n_threads, low_water_mark, high_water_mark, flags) == -1)
    ACE_ERROR ((LM_ERROR, "%p\n",
                "ACE_Thread_Strategy<SVC_HANDLER>::ACE_Thread_Strategy"));
}


template <class SVC_HANDLER> 
ACE_Thread_Pool_Strategy<SVC_HANDLER>::ACE_Thread_Pool_Strategy (int flags)
  : ACE_Concurrency_Strategy<SVC_HANDLER> (flags),
    thr_mgr_ (0),
    thr_flags_ (0),
    n_threads_ (1)
{
  ACE_TRACE ("ACE_Thread_Pool_Strategy<SVC_HANDLER>::ACE_Thread_Pool_Strategy");

  ACE_Thread_Pool_Strategy::sem_=new ACE_Semaphore(0);	

  ACE_Thread_Pool_Strategy::low_water_mark_  = -1;
  ACE_Thread_Pool_Strategy::high_water_mark_ = MAX_HIGH_WATER_MARK;
}


template <class SVC_HANDLER>
ACE_Thread_Pool_Strategy<SVC_HANDLER>::~ACE_Thread_Pool_Strategy (void)
{
  ACE_TRACE ("ACE_Thread_Strategy<SVC_HANDLER>::~ACE_Thread_Strategy");
}


template <class SVC_HANDLER> int
ACE_Thread_Pool_Strategy<SVC_HANDLER>::activate_svc_handler (SVC_HANDLER *svc_handler,
                                                        void *arg)
{
  ACE_TRACE ("ACE_Thread_Strategy<SVC_HANDLER>::activate_svc_handler");
  // Call up to our parent to do the SVC_HANDLER initialization.

  SvcHandler_Record *tempNode=new SvcHandler_Record(svc_handler);

	if (this->inherited::activate_svc_handler (svc_handler, arg) == -1)
	{
		return -1;
	}
	else
	{
		//Assign parent and thread manager to the received svc_handler
		//svc_handler->parent(TAO_ORB_Core_instance ());
		//svc_handler->thr_mgr (ACE_Thread_Manager::instance());

		ACE_Thread_Pool_Strategy::mutex_.acquire();
		ACE_Thread_Pool_Strategy::svc_handler_list_.insert_tail(tempNode);
		ACE_Thread_Pool_Strategy::mutex_.release();

		//Let some thread execute the Service Handler. Don't need to chech whether there are available threads or not. Since
		//a general semaphore is being used, as soon as we get a request, we call release, so as soon as there is an available
		//it can enter the critical region.
		ACE_Thread_Pool_Strategy::sem_->release();

		//////////////////////////////////////////////////////////////////////////////////////////////////////////////
		ACE_Thread_Pool_Strategy::mutex_.acquire();
		if(ACE_Thread_Pool_Strategy::available_threads_>0)
			ACE_Thread_Pool_Strategy::available_threads_--;
		else
			waiting_requests_++;
		
		//TO-DO:In case of a reconfiguration, it must be checked if it is necessary to spawn threads!!!
		if(this->available_threads_==ACE_Thread_Pool_Strategy::low_water_mark_)
		{  
			this->thr_mgr_->spawn_n (n_threads_-ACE_Thread_Pool_Strategy::low_water_mark_,ACE_THR_FUNC (&ACE_Thread_Pool_Strategy::entry_point));
			available_threads_+=(n_threads_-ACE_Thread_Pool_Strategy::low_water_mark_);
		}
		ACE_Thread_Pool_Strategy::mutex_.release();
		//////////////////////////////////////////////////////////////////////////////////////////////////////////////////
	}
	return 0;
}
#endif
