/******************************************************************************
*                                                                             *
*  File: ACE_Persistent_Repository.cpp                                        *
*                                                                             *
*  Description: A hierarchical repository of data implemented on top of the   *
*               ACE_OS file system.                                           *
*                                                                             *
*  Project: 2K - dynamically configurable ORB.                                *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************

   Revision Control Information

   $Revision: 1.1.1.1 $

   $Log: ACE_Persistent_Repository.cpp,v $
   Revision 1.1.1.1  1999/10/12 00:43:30  f-kon
   Version compatible with TAO 1.0

   Revision 1.1.1.1  1999/04/01 18:16:36  f-kon
   This new 2K source tree contains the most up-to-date version of ACE and TAO
   and including the dynamicTAO code as well as some 2K tests and preliminary
   versions of some of its high-level systems.


   Revision 1.9  1999/02/22 22:21:02  f-kon
   Introduced NiceConfigurator and LockingConfigurator and performed several
   fixes on dynamic reconfiguration.

   Revision 1.8  1999/02/16 21:10:19  ckhess
   Don't compare return of readdir to NULL (a void *) in get_contents.
   readdir returns BOOL, so now treated as a boolean, not void *.

   Revision 1.7  1999/02/08 21:13:37  mroman
   no message

   Revision 1.6  1999/01/28 18:01:35  mroman
   Fixed the version problem. Versions are only incremented when 

   loading and implementation succeds.

   Revision 1.5  1999/01/11 23:13:46  f-kon
   Upgrade to DCP v1.0 and improvements to the PersistentRepository.

   Revision 1.4  1998/10/23 17:14:27  f-kon
   Solaris working version with Manuel's Thread-Pool

   Revision 1.3  1998/10/06 18:16:15  mroman
   Version that works fine on NT, Win98 and compiles fine

   Revision 1.2  1998/08/30 14:56:19  f-kon
   First modifications to try to make it work in Windows

******************************************************************************/

#include "ace/OS.h"
#include "ace/SOCK_Stream.h"
#include "ACE_2K_OS.h"
#include <sys/types.h>
#include "ACE_Persistent_Repository.h"

ACE_Persistent_Repository::ACE_Persistent_Repository ()
{
	root_dir_=0;
}

ACE_Persistent_Repository::~ACE_Persistent_Repository ()
{
	ACE_OS::free (root_dir_);
}

int ACE_Persistent_Repository::init (const char *rootDir)
{
  struct stat tmp;

  if (ACE_OS::stat (rootDir, &tmp) == -1)
    ACE_ERROR_RETURN ((LM_NOTICE, "Persistent_Repository: could not open directory\n"),
                      -1);

  root_dir_ = ACE_OS::strdup (rootDir);
  return 0;

}// init()


//////////////////////////////////////////////////////////////////////////////
// Fills buf with a list of strings containing the contents of a
// directory. Returns -1 if the directory cannot be accessed and 0 o/w.
// It ignores entries starting with '.' and entries called "CVS" and removes
// unwanted prefixes and suffixes from the entry names.
//////////////////////////////////////////////////////////////////////////////
int ACE_Persistent_Repository::get_contents (char *buf, int maxsize, const char *dir)
{
  char path[MAXPATHLEN]; 
  
  DIR_HANDLE dirp;
  
  char name[MAXNAMELEN];
  
  // For security reasons, we don't allow ".." in the pathname.
  if (buf == 0 || (dir != 0 && ACE_OS::strstr(dir, "..") != 0))
    return -1;
  
  build_pathname (path, 0, dir);
  
  dirp = ACE_2k_os::opendir (path);
  
  if (dirp == NULL)
    ACE_ERROR_RETURN ((LM_WARNING, "Persistent_Repository: could not open directory\n"),
                      -1);
  buf[0] = '\0';
  int size = 0;
  
  while ((ACE_2k_os::readdir (dirp, name)))
    {
      if (name[0] != '.' && ACE_OS::strcmp(name, "CVS")) // ignore names starting with '.' 
        {                                                // and "CVS".
          char *cleaned = clean_name(name); // get rid of unwanted prefixes and suffixes.
          ACE_OS::strcat (buf, cleaned);
          ACE_OS::strcat (buf, "\r\n");
          size += (ACE_OS::strlen(cleaned) +2);
        }
      
      if (size > maxsize - MAXNAMLEN)
        {
          ACE_OS::strcat (buf, "There are too many entries, truncating list...\r\n");
          break;
        }
    }
  
  ACE_2k_os::closedir (dirp);
  return 0;
}// get_contents()







//////////////////////////////////////////////////////////////////////////////
// Fills buf with a list of strings containing the contents of a
// directory. Returns -1 if the directory cannot be accessed and 0 o/w.
// It ignores entries starting with '.' and entries called "CVS" and removes
// unwanted prefixes and suffixes from the entry names.
//////////////////////////////////////////////////////////////////////////////
int ACE_Persistent_Repository::get_contents (char **buf, int &strings, int maxsize, const char *dir)
{
  char path[MAXPATHLEN]; 
  
  DIR_HANDLE dirp;
  
  char name[MAXNAMELEN];
  
  // For security reasons, we don't allow ".." in the pathname.
  if (buf == 0 || (dir != 0 && ACE_OS::strstr(dir, "..") != 0))
    return -1;
  
  build_pathname (path, 0, dir);
  
  dirp = ACE_2k_os::opendir (path);
  
  if (dirp == NULL)
    ACE_ERROR_RETURN ((LM_WARNING, "Persistent_Repository: could not open directory"),
                      -1);

  strings=0;
  
  while ((ACE_2k_os::readdir (dirp, name)))
    {
      if (name[0] != '.' && ACE_OS::strcmp(name, "CVS")) // ignore names starting with '.' 
        {                                                // and "CVS".
          char *cleaned = clean_name(name); // get rid of unwanted prefixes and suffixes.
          buf[strings]=new char[ACE_OS::strlen(cleaned)+1];
		  ACE_OS::strcpy(buf[strings], cleaned);
          ACE_OS::strcat (buf[strings], "\0");
          strings ++;
        }
      
      if (strings > maxsize - 1)
        {
          ACE_OS::strcpy (buf[strings], "There are too many entries, truncating list...\r\n");
          break;
        }
    }
  
  ACE_2k_os::closedir (dirp);
  return 0;
}// get_contents()









////////////////////////////////////////////////////////////////////////////
// Reads len bytes from the input stream and saves them in entryname at the
// specified path. Returns -1 on failure and 0 otherwise.
// It blocks while reading from the input.
////////////////////////////////////////////////////////////////////////////
// The user of this method must be aware that if it's updating an existing entry
// and the connection fails in the middle of the process, it will damage the 
// existing implementation.
int 
ACE_Persistent_Repository::create_entry_from_stream (ACE_SOCK_Stream *input, 
                                const char *path, 
                                const char *entryname, 
                                long len)
{
  char pathname[MAXPATHLEN];

  // For security reasons, we don't allow ".." in the pathname.
  if (input == 0 || path == 0 || entryname == 0 || len == 0 ||
      ACE_OS::strstr(path, "..") != 0)
    return -1;

  // build complete pathname
  build_pathname (pathname, path, entryname);

  ACE_HANDLE entry = ACE_OS::open(pathname, O_CREAT | O_TRUNC | O_WRONLY, 
                                  S_IRWXU | S_IRWXG); // mode = 770

  if (entry <= 0)
    return -1;

  // Now, read from the input streams in chuncks of size BUFSIZ and write the
  // data to the file we have created.

  char buf[BUFSIZ];
  long remainingToRead = len;

  while (remainingToRead > 0)
    {
      int bytesToRead = ACE_MIN (remainingToRead, BUFSIZ);

      // read from stream
      int ret = input->recv_n (buf, bytesToRead);

      // write to file
      if (ret < 0 || ACE_OS::write(entry, buf, ret) < ret)
        {
          ACE_OS::close(entry);
          return -1;
        }

      remainingToRead -= ret;
    }

  return ACE_OS::close(entry);

}// create_entry_from_stream ()

////////////////////////////////////////////////////////////////////////////
// Reads len bytes from a buffer and saves them in entryname at the
// specified path. Returns -1 on failure and 0 otherwise.
////////////////////////////////////////////////////////////////////////////
int 
ACE_Persistent_Repository::create_entry_from_buffer (char *input, 
                                const char *path, 
                                const char *entryname, 
                                long len)
{
  ACE_DEBUG((LM_DEBUG,"Creating entry from stream of length %d bytes\n",
            len));
  char pathname[MAXNAMELEN];

  // For security reasons, we don't allow ".." in the pathname.
  if (input == 0 || path == 0 || entryname == 0 || len == 0 ||
      ACE_OS::strstr(path, "..") != 0)
    return -1;

  // build complete pathname
  build_pathname (pathname, path, entryname);

  ACE_HANDLE entry = ACE_OS::open(pathname, O_CREAT | O_TRUNC | O_WRONLY, 
                                  S_IRWXU | S_IRWXG); // mode = 770

  if (entry <= 0)
    return -1;

  // write to file
  if (ACE_OS::write(entry, input, len) < len)
    {
      ACE_OS::close(entry);
      return -1;
    }

  return ACE_OS::close(entry);
}// create_entry_from_buffer ()


///////////////////////////////////////////////////////////////////////////
// Delete an entry
///////////////////////////////////////////////////////////////////////////
int 
ACE_Persistent_Repository::delete_entry (const char *path, const char *entryname)
{
  char pathname[MAXPATHLEN];
  
  // For security reasons, we don't allow ".." in the pathname.
  if (path == 0 || entryname == 0 || ACE_OS::strstr(path, "..") != 0)
    return -1;

  // build complete pathname
  build_pathname (pathname, path, entryname);

  if (ACE_OS::unlink (pathname) < 0)
    {
      // The deletion failed. The problem may be that the user didn't type the
      // DLL prefix and sufix, let's add it ourselves and try again.

      char libname[MAXNAMELEN];
      ACE_OS::strcpy(libname, ACE_DLL_PREFIX);
      ACE_OS::strcat(libname, entryname);
      ACE_OS::strcat(libname, ACE_DLL_SUFFIX);

      build_pathname (pathname, path, libname);
      return ACE_OS::unlink (pathname);
    }
  else
    return 0;
}// delete_entry ()



///////////////////////////////////////////////////////////////////////////
// Assembles a pathname. It assumes that all input parameters are valid.
// It doesn't check for overflows. Use with caution.
///////////////////////////////////////////////////////////////////////////
void
ACE_Persistent_Repository::build_pathname (char *pathname, const char *path, 
                                       const char *entryname)
{
  if (!root_dir_)
    {
      *pathname = NULL;
      return;
    }
  
  ACE_OS::strcpy(pathname, root_dir_);
  
  if (path != 0 && *path != '\0')
    {
      ACE_OS::strcat(pathname,ACE_DIRECTORY_SEPARATOR_STR_A);
      ACE_OS::strcat(pathname,path);
    }

  if (entryname != 0 && *entryname != '\0')
    {
      ACE_OS::strcat(pathname, ACE_DIRECTORY_SEPARATOR_STR_A);
      ACE_OS::strcat(pathname, entryname);
    }

}// build_pathname ()

///////////////////////////////////////////////////////////////////////////
// Returns a pointer to a string not containing unwanted prefixes and suffixes
// from the given string.
// In Unix, the "lib" prefix and the ".so" suffixes are unwanted.
// In Windows, the ".dll" suffix is unwanted.
///////////////////////////////////////////////////////////////////////////
char * ACE_Persistent_Repository::clean_name (char *name)
{
  char *retVal;
  if (!name)
    return NULL;
  
  // remove prefix
  if (ACE_OS::strncmp(name, ACE_DLL_PREFIX, sizeof (ACE_DLL_PREFIX)-1) == 0)
    retVal = &name[sizeof (ACE_DLL_PREFIX)-1];
  else
    retVal = name;
  
  // remove suffix
  int size = ACE_OS::strlen(retVal);
  if (size >= sizeof (ACE_DLL_SUFFIX)-1)
    if (ACE_OS::strncmp(&retVal[size - (sizeof (ACE_DLL_SUFFIX)-1)], ACE_DLL_SUFFIX,
                        sizeof (ACE_DLL_SUFFIX)-1) == 0)
      retVal[size - (sizeof (ACE_DLL_SUFFIX)-1)] = '\0';

  return retVal;

}// clean_name ()

