/******************************************************************************
*                                                                             *
*  File: DCP_Broker.cpp                                                       *
*                                                                             *
*  Description: A Dynamic Config Broker that implements DCP, the              *
*               Distributed Configuration Protocol.                           *
*                                                                             *
*  Project: 2K - dynamically configurable ORB.                                *
*  Author: Fabio Kon (f-kon@cs.uiuc.edu)                                      *
*                                                                             *
*******************************************************************************

   Revision Control Information

   $Revision: 1.2 $

   $Log: DCP_Broker.cpp,v $
   Revision 1.2  1999/11/22 16:20:47  f-kon
   bug fixes and performance optimizations

   Revision 1.1.1.1  1999/10/12 00:43:30  f-kon
   Version compatible with TAO 1.0

   Revision 1.4.1.1  1999/09/22 00:01:22  yamane
   Change
   	Dynamic_Service_Config::XXXX()
                  --> Dynamic_Service_Config::instance()->XXXX()

   Revision 1.4  1999/04/24 17:29:16  f-kon
   *** empty log message ***

   Revision 1.3  1999/04/23 22:34:22  f-kon

   Revision 1.2  1999/04/07 20:09:02  f-kon
   added new DCP commands, upgrading to version 1.1

   Revision 1.1.1.1  1999/04/01 18:16:36  f-kon
   This new 2K source tree contains the most up-to-date version of ACE and TAO
   and including the dynamicTAO code as well as some 2K tests and preliminary
   versions of some of its high-level systems.


   Revision 1.13  1999/03/03 16:12:11  f-kon
   Implemented DCP operation configure_implementation <loaded_implementation> <configuration_string>

   Revision 1.12  1999/01/29 23:05:35  mroman
   no message

   Revision 1.11  1999/01/28 18:01:18  mroman
   Fixed the version problem. Versions are only incremented when 

   loading and implementation succeds.

   Revision 1.10  1999/01/26 20:03:43  f-kon
   added call to netBroker->fini() to fix segmentation fault bug in client

   Revision 1.9  1999/01/26 18:22:31  mroman
   DomainConfigurator running on Windows NT

   Revision 1.8  1999/01/11 23:13:50  f-kon
   Upgrade to DCP v1.0 and improvements to the PersistentRepository.

   Revision 1.7  1999/01/07 22:43:29  f-kon
   Added TAOConfigurator and new DCP commands

   Revision 1.6  1998/10/27 23:52:43  mroman
   Version Control and a strategy hooked cannot be removed

   Revision 1.5  1998/10/06 18:16:18  mroman
   Version that works fine on NT, Win98 and compiles fine

   Revision 1.4  1998/08/30 14:56:26  f-kon
   First modifications to try to make it work in Windows

******************************************************************************/

#include "DCP_Broker.h"

ACE_ALLOC_HOOK_DEFINE(DCP_Broker)

#include "ace/Reactor.h"
#include "ace/Service_Repository.h"
#include "ace/Dynamic_Service.h"
#include "DCP.h"
#include "Dynamic_Service_Config.h"
#include "Simple_Strategy_Factory.h"

int 
DCP_Broker::info (char **strp, size_t length) const
{
  ACE_TRACE ("DCP_Broker::info");
  ACE_INET_Addr sa;
  char buf[BUFSIZ];

  if (this->acceptor_.get_local_addr (sa) == -1)
    return -1;
  
  ACE_OS::sprintf (buf, "%d/%s %s %d.%d", sa.get_port_number (), "tcp", 
                   "# DCP Broker version", 
                   DCP::major_version, DCP::minor_version);

  if (*strp == 0 && (*strp = ACE_OS::strdup (buf)) == 0)
    return -1;
  else
    ACE_OS::strncpy (*strp, buf, length);
  return ACE_OS::strlen (buf);
}// info ()


///////////////////////////////////////////////////////////////////////////
// Calls the superclass init method to parse arguments, open the server 
// socket, and register us with the Reactor.
///////////////////////////////////////////////////////////////////////////
int 
DCP_Broker::initialize() 
{
  int argc = 0;
  char *argv[1] = {0};

  return this->init(argc, argv, "telnet"); // hard-coded parameters by now.
}


///////////////////////////////////////////////////////////////////////////
// Handles one client request and returns.
///////////////////////////////////////////////////////////////////////////
int 
DCP_Broker::handle_request ()
{
  // Read request from client.
  int ret = client_stream_.recv (request_ + request_index_, 
                                 sizeof request_ - request_index_);

  switch (ret)
    {
    case -1:
      if (this->debug_)
        {
          ACE_DEBUG ((LM_ERROR, "%p\n", "recv"));
        }
      this->quit (0); // quit connection but do not send ACK.
      break;

    case 0:
      // recv returned 0, this is suspicious. 
      // Let's send an ACK to the client to check if the connection
      // is still alive.
      if (this->client_stream_.send_n (DCP_Reply::ACK, 
                                       ACE_OS::strlen(DCP_Reply::ACK))
          < 0)
        {
          this->quit (0); // quit connection but do not send ACK.
        }
      return 0;
      /* NOTREACHED */

    default:
      {
        // Check whether we received a whole line or not.
        // We assume that the line is terminated by "\r\n" and that
        // "\n" cannot occur without being preceded by a "\r".
	char *p = request_ + request_index_;
        char *last = p + ret;
	while ((*p != '\0') && (*p != '\n') && (p < last))
          {
            p++;
          }

        if (p == last || *p != '\n')
          {
            // We didn't receive a whole line yet.
            request_index_ += ret;
            return 0;     
          }

        // We did receive a whole line. Kill trailing CR/LF.

	*(p-1) = '\0'; // override the '\r' with '\0'.
        request_index_ = 0;

        // Why do we need to do this signal_handler stuff????? Fabio.
        // Note that ACE_Reactor::instance is not necessarily the reactor 
        // we use. 
	ACE_Event_Handler *old_signal_handler = 0;
	ACE_Reactor::instance ()->register_handler (SIGPIPE, this, 0,
						    &old_signal_handler);


	

        /******************** Parse DCP Requests ********************/

        // The following chain of if's must be sorted so that common requests
        // come first.	
	if (ACE_OS::strcmp (request_, "help") == 0)
          {
            this->help ();
          }
        else if (!ACE_OS::strncmp (request_, DCP::LIST_CATEGORIES, 
                                   ACE_OS::strlen (DCP::LIST_CATEGORIES)))
          {
            this->list_categories();
          }
        else if (!ACE_OS::strncmp (request_, DCP::LIST_IMPLEMENTATIONS, 
                                   ACE_OS::strlen (DCP::LIST_IMPLEMENTATIONS)))
          {
            this->list_implementations (request_ + 
                                        ACE_OS::strlen (DCP::LIST_IMPLEMENTATIONS));
          }
        else if (!ACE_OS::strncmp (request_, DCP::GET_LATEST_VERSION, 
                                   ACE_OS::strlen (DCP::GET_LATEST_VERSION)))
          {
            this->get_latest_version (request_ + 
                                        ACE_OS::strlen (DCP::GET_LATEST_VERSION));
          }
        else if (!ACE_OS::strncmp (request_, DCP::LIST_LOADED_IMPLEMENTATIONS, 
                                   ACE_OS::strlen (DCP::LIST_LOADED_IMPLEMENTATIONS)))
          {
            this->list_loaded_implementations ();
          }
        else if (!ACE_OS::strncmp (request_, DCP::GET_IMPL_INFO, 
                                   ACE_OS::strlen (DCP::GET_IMPL_INFO)))
          {
            this->get_impl_info (request_+ACE_OS::strlen (DCP::GET_IMPL_INFO));
          }
        else if (!ACE_OS::strncmp (request_, DCP::LIST_DOMAIN_COMPONENTS, 
                                   ACE_OS::strlen (DCP::LIST_DOMAIN_COMPONENTS)))
          {
            this->list_domain_components ();
          }
        else if (!ACE_OS::strncmp (request_, DCP::LIST_HOOKS, 
                                   ACE_OS::strlen (DCP::LIST_HOOKS)))
          {
            this->list_hooks (request_+ACE_OS::strlen (DCP::LIST_HOOKS));
          }
        else if (!ACE_OS::strncmp (request_, DCP::GET_COMP_INFO, 
                                   ACE_OS::strlen (DCP::GET_COMP_INFO)))
          {
            this->get_comp_info (request_+ACE_OS::strlen (DCP::GET_COMP_INFO));
          }
        else if (!ACE_OS::strncmp (request_, DCP::GET_HOOKED_COMP, 
                                   ACE_OS::strlen (DCP::GET_HOOKED_COMP)))
          {
            this->get_hooked_comp (request_+ACE_OS::strlen (DCP::GET_HOOKED_COMP));
          }
        else if (!ACE_OS::strncmp (request_, DCP::LOAD_IMPLEMENTATION, 
                                   ACE_OS::strlen (DCP::LOAD_IMPLEMENTATION)))
          {
            this->load_implementation (request_+ACE_OS::strlen (DCP::LOAD_IMPLEMENTATION), p);
          }
        else if (!ACE_OS::strncmp (request_, DCP::ADD_HOOK, 
                                   ACE_OS::strlen (DCP::ADD_HOOK)))
          {
            this->add_hook (request_+ACE_OS::strlen (DCP::ADD_HOOK));
          }
        else if (!ACE_OS::strncmp (request_, DCP::REMOVE_HOOK, 
                                   ACE_OS::strlen (DCP::REMOVE_HOOK)))
          {
            this->remove_hook (request_+ACE_OS::strlen (DCP::REMOVE_HOOK));
          }
        else if (!ACE_OS::strncmp (request_, DCP::HOOK_IMPLEMENTATION, 
                                   ACE_OS::strlen (DCP::HOOK_IMPLEMENTATION)))
          {
            this->hook_implementation (request_+ACE_OS::strlen (DCP::HOOK_IMPLEMENTATION));
          }
        else if (!ACE_OS::strncmp (request_, DCP::SUSPEND_IMPLEMENTATION, 
                                   ACE_OS::strlen (DCP::SUSPEND_IMPLEMENTATION)))
          {
            this->suspend_implementation (request_+ACE_OS::strlen (DCP::SUSPEND_IMPLEMENTATION));
          }
        else if (!ACE_OS::strncmp (request_, DCP::RESUME_IMPLEMENTATION, 
                                   ACE_OS::strlen (DCP::RESUME_IMPLEMENTATION)))
          {
            this->resume_implementation (request_+ACE_OS::strlen (DCP::RESUME_IMPLEMENTATION));
          }
        else if (!ACE_OS::strncmp (request_, DCP::REMOVE_IMPLEMENTATION, 
                                   ACE_OS::strlen (DCP::REMOVE_IMPLEMENTATION)))
          {
            this->remove_implementation (request_+ACE_OS::strlen (DCP::REMOVE_IMPLEMENTATION));
          }
        else if (!ACE_OS::strncmp (request_, DCP::CONFIGURE_IMPLEMENTATION, 
                                   ACE_OS::strlen (DCP::CONFIGURE_IMPLEMENTATION)))
          {
            this->configure_implementation (request_+ACE_OS::strlen (DCP::CONFIGURE_IMPLEMENTATION), p);
          }
        else if (!ACE_OS::strncmp (request_, DCP::UPLOAD_IMPLEMENTATION, 
                                   ACE_OS::strlen (DCP::UPLOAD_IMPLEMENTATION)))
          {
            this->upload_implementation (request_ +
                                         ACE_OS::strlen (DCP::UPLOAD_IMPLEMENTATION));
          }
        else if (!ACE_OS::strncmp (request_, DCP::DELETE_IMPLEMENTATION, 
                                   ACE_OS::strlen (DCP::DELETE_IMPLEMENTATION)))
          {
            this->delete_implementation (request_ +
                                         ACE_OS::strlen (DCP::DELETE_IMPLEMENTATION));
          }
	else if (ACE_OS::strcmp (request_, DCP::RESTART) == 0)
          {
            this->restart ();
          }

        /**************************       QUIT         **************************/

	else if (ACE_OS::strcmp (request_, "quit") == 0)
          {
            this->quit (1);
          }
        else if (ACE_OS::strlen (request_) == 0)
          {
            // just ignore it
          }
        else // unknown request
          {
            if (this->debug_)
              ACE_DEBUG ((LM_ERROR, "DCP Broker received unknown request.\n"));

            this->client_stream_.send_n (DCP_Reply::UNKNOWN_REQUEST, 
                                         ACE_OS::strlen (DCP_Reply::UNKNOWN_REQUEST));
          }

	// Restore existing SIGPIPE handler
	ACE_Reactor::instance ()->register_handler (SIGPIPE, old_signal_handler);
      }
    }

  return 0;
}// handle_request()



///////////////////////////////////////////////////////////////////////////
// Determine all the loaded implementations offered by this process and
// return the information back to the client.
///////////////////////////////////////////////////////////////////////////
void
DCP_Broker::list_loaded_implementations (void) const
{
  int listed_something = 0;
  ACE_Service_Repository_Iterator sri (*ACE_Service_Repository::instance ());

  for (const ACE_Service_Type *sr; 
       sri.next (sr) != 0; 
       sri.advance ())
    {
      this->client_stream_.send_n (sr->name (), ACE_OS::strlen (sr->name ()));
      this->client_stream_.send_n ("\r\n", 2);      
      listed_something++;
    }

  if (listed_something == 0)
    {
      this->client_stream_.send_n (DCP_Reply::EMPTY, ACE_OS::strlen(DCP_Reply::EMPTY));
    }
  else
    {
      this->client_stream_.send_n ("\r\n", 2);      
    }
}// list_loaded_implementations()


///////////////////////////////////////////////////////////////////////////
// Sends a help message containing the protocol version number and the 
// URL of the proocol specification.
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::help () const
{
  char buf[128];

  int size = 
    ACE_OS::sprintf (buf, "DCP version %d.%d.\r\n"
                     "Visit http://choices.cs.uiuc.edu/2K/DCP "
                     "to get the DCP specification.\r\n", 
                     DCP::major_version, DCP::minor_version);

  this->client_stream_.send_n (buf, size);


#define help_msg "The following operations are currently supported:\r\n" \
  "\tlist_categories\r\n" \
  "\tlist_implementations   <category_name>\r\n" \
  "\tget_latest_version     <category_name>\r\n" \
  "\tlist_loaded_implementations\r\n" \
  "\tget_impl_info          <loaded_implementation>\r\n" \
  "\tlist_domain_components\r\n" \
  "\tlist_hooks             <component_name>\r\n" \
  "\tget_comp_info          <component_name>\r\n" \
  "\tget_hooked_comp        <component_name> <hook_name>\r\n" \
  "\tadd_hook               <component_name> <hook_name>\r\n" \
  "\tremove_hook            <component_name> <hook_name>\r\n" \
  "\tload_implementation    <category_name> <implementation_name> [<init_parameters>]\r\n" \
  "\thook_implementation    <loaded_implementation> <component_name> <hook_name>\r\n" \
  "\tsuspend_implementation <loaded_implementation>\r\n" \
  "\tresume_implementation  <loaded_implementation>\r\n" \
  "\tremove_implementation  <loaded_implementation>\r\n" \
  "\tconfigure_implementation  <loaded_implementation> <configuration_string>\r\n" \
  "\tupload_implementation  <category_name> <implementation_name>\r\n" \
  "\t                       <length> CR/LF <binary-code>\r\n" \
  "\tdelete_implementation  <category_name> <implementation_name>\r\n" \
  "\trestart\r\n" \
  "\tquit\r\n" \
  "\thelp \r\n\r\n"

  this->client_stream_.send_n (help_msg, sizeof(help_msg));

}// help()


///////////////////////////////////////////////////////////////////////////
// Send an ACK, remove this handler from the reactor and closes the stream
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::quit (int sendACK)
{
  if (sendACK)
    {
      this->client_stream_.send_n (DCP_Reply::ACK, ACE_OS::strlen(DCP_Reply::ACK));
    }
  // remove this handler from the reactor 
  // Note: this DONT_CALL is important to sign that we don't want
  // remove_handler to close the event_handler, i.e., ourselves.
  this->reactor_->remove_handler
    (this->client_stream_.get_handle (),
     ACE_Event_Handler::READ_MASK |
     ACE_Event_Handler::DONT_CALL );
  
  // close the client stream and set handle to ACE_INVALID_HANDLE
  if (this->client_stream_.close () == -1 && this->debug_)
    {
      ACE_DEBUG ((LM_ERROR, "%p\n", "close"));
    }
}// quit()


/////////////////////////////////////////////////////////////////////////
// Sets arg1, arg2, arg3, and arg4 to point to the space-separated words
// in the request strings. It might ignore additional words if arg2,
// arg3, or arg4 is 0.  
// Should return -1 if an error occurs (by now, it returns 0 always).
/////////////////////////////////////////////////////////////////////////
int 
DCP_Broker::get_params (char *request, char **arg1, char **arg2, 
                        char **arg3, char **arg4) const
{
  char *lasts = request;

  *arg1 = ACE_OS::strtok_r (request, " \t", &lasts);
  if(arg2)
  {
      *arg2 = ACE_OS::strtok_r (NULL, " \t", &lasts);
      if(arg3)
        {
          *arg3 = ACE_OS::strtok_r (NULL, " \t", &lasts);
          if(arg4)
            {
              *arg4 = ACE_OS::strtok_r (NULL, " \t", &lasts);
            }
        }
  }

  return 0;
}// get_params()

///////////////////////////////////////////////////////////////////////////
// List categories in the persistent repository. By now, it is implemented by
// listing all the entities in the repository root directory.
// It sends a string containing the category names separated by "\r\n". 
// The list is terminated by "\r\n\r\n".
///////////////////////////////////////////////////////////////////////////
int
DCP_Broker::list_categories () const
{
  if (pers_rep_->get_contents ((char *)buf_, BUFSIZE_) == -1)
    {
      this->client_stream_.send_n (DCP_Reply::NACK, ACE_OS::strlen(DCP_Reply::NACK));
    }

  else if (buf_[0] == '\0')
    {
      this->client_stream_.send_n (DCP_Reply::EMPTY, ACE_OS::strlen(DCP_Reply::EMPTY));
    }
  else
    {
      ACE_OS::strcat ((char *)buf_, "\r\n");
      this->client_stream_.send_n (buf_, ACE_OS::strlen(buf_));
    }
  
  return 0;
}// list_categories ()

///////////////////////////////////////////////////////////////////////////
// List the implementations of a particular category in the persistent
// repository. By now, it is implemented by listing all the entities in
// the subdirectory named after the category in the repository root
// directory.  It sends a string containing the implementation names
// separated by "\r\n".  The list is terminated by "\r\n\r\n".
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::list_implementations (char *request) const
{
  char *category;

  this->get_params (request, &category);

  if(category == 0)
    {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
    }
  
  if (this->pers_rep_->get_contents ((char*)buf_, BUFSIZE_, category) != -1)
    {
      if (buf_[0] == '\0')
        {
          this->client_stream_.send_n (DCP_Reply::EMPTY, 
                                       ACE_OS::strlen(DCP_Reply::EMPTY));
        }
      else
        {
          ACE_OS::strcat ((char*)buf_, "\r\n");
          this->client_stream_.send_n (buf_, ACE_OS::strlen(buf_));
        }
    }
  else // could not find entity
    {
      this->client_stream_.send_n (DCP_Reply::UNKNOWN_ENTITY, 
                                   ACE_OS::strlen(DCP_Reply::UNKNOWN_ENTITY));
    }
  
}// list_implementations ()

///////////////////////////////////////////////////////////////////////////
// Sends a string containing the latest version number for loaded 
// implementations of the given category.
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::get_latest_version (char *request) const
{
  char *category;
  char temp[MAXNAMELEN];
  char *implementationName;
  
  this->get_params (request, &category);
  
  if(category == 0)
    {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
    }
  
  int version = Dynamic_Service_Config::instance ()->get_latest_version (category, &implementationName);
  
  if(implementationName != 0)
    {
      ACE_OS::sprintf(temp, "%s:%d\t%s\r\n\r\n", category, version, implementationName);
      this->client_stream_.send_n (temp, ACE_OS::strlen(temp));
    }
  else 
    this->client_stream_.send_n (DCP_Reply::NACK, 
                                 ACE_OS::strlen(DCP_Reply::NACK));
  
  
}// get_latest_version ()


///////////////////////////////////////////////////////////////////////////
// List the components registered with the DomainConfigurator.
// It sends a string containing the component names
// separated by "\r\n".  The list is terminated by "\r\n\r\n".
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::list_domain_components () const
{
    Dynamic_Service_Config::instance()->get_components((char*)buf_, BUFSIZE_);
    
    if (buf_[0] == '\0')
        {
            this->client_stream_.send_n (DCP_Reply::EMPTY, 
                                         ACE_OS::strlen(DCP_Reply::EMPTY));
        }
    else
        {
            ACE_OS::strcat ((char*)buf_, "\r\n");
            this->client_stream_.send_n (buf_, ACE_OS::strlen(buf_));
        }
  
}// list_domain_components ()


///////////////////////////////////////////////////////////////////////////
// List the hooks of a given component registered with the 
// domain configurator. 
// It sends a string containing the implementation names
// separated by "\r\n".  The list is terminated by "\r\n\r\n".
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::list_hooks (char *request) const
{
  char *component;

  this->get_params (request, &component);

  if(component == 0)
    {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
    }
  
  if (Dynamic_Service_Config::instance()->get_hooks (component, (char*)buf_, BUFSIZE_) == -1)
    {
      this->client_stream_.send_n (DCP_Reply::UNKNOWN_ENTITY, 
                                   ACE_OS::strlen(DCP_Reply::UNKNOWN_ENTITY));
    }
  else
    {
      if (buf_[0] == '\0')
        {
          this->client_stream_.send_n (DCP_Reply::EMPTY, 
                                       ACE_OS::strlen(DCP_Reply::EMPTY));
        }
      else
        {
          ACE_OS::strcat ((char*)buf_, "\r\n");
          this->client_stream_.send_n (buf_, ACE_OS::strlen(buf_));
        }
    }

}// list_hooks ()



//////////////////////////////////////////////////////////////////////////////
// Sends to the client an information string about the given domain component.
//////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::get_comp_info (char *request) const
{
  char *component;

  this->get_params (request, &component);

  if(component == 0)
    {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
    }
  
  if (Dynamic_Service_Config::instance()->get_comp_info (component, (char*)buf_, BUFSIZE_) == -1)
    {
      this->client_stream_.send_n (DCP_Reply::UNKNOWN_ENTITY, 
                                   ACE_OS::strlen(DCP_Reply::UNKNOWN_ENTITY));
    }
  else
    {
      if (buf_[0] == '\0')
        {
          this->client_stream_.send_n (DCP_Reply::EMPTY, 
                                       ACE_OS::strlen(DCP_Reply::EMPTY));
        }
      else
        {
          ACE_OS::strcat ((char*)buf_, "\r\n\r\n");
          this->client_stream_.send_n (buf_, ACE_OS::strlen(buf_));
        }
    }

}// get_comp_info ()



////////////////////////////////////////////////////////////////////////////
// Returns the name of the component attached to a given hook.
////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::get_hooked_comp (char *request) const
{
 char *componentName;
 char *hookName;
 this->get_params (request, &componentName, &hookName);

 if (componentName == 0 || hookName == 0)
   {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }
 
 const char *msg = Dynamic_Service_Config::instance()->get_hooked_comp (hookName, componentName);
 
 if (msg != 0)
   {
     this->client_stream_.send_n (msg, ACE_OS::strlen(msg));
     this->client_stream_.send_n ("\r\n\r\n", sizeof("\r\n\r\n"));
   }
  else
   this->client_stream_.send_n (DCP_Reply::NACK, 
                                ACE_OS::strlen(DCP_Reply::NACK));
}// get_hooked_component ()

////////////////////////////////////////////////////////////////////////////
// Adds a new hook to a component
////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::add_hook (char *request) 
{
 char *componentName;
 char *hookName;
 this->get_params (request, &componentName, &hookName);

 if (componentName == 0 || hookName == 0)
   {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }
 
  this->AckOrNack (Dynamic_Service_Config::instance()->add_hook (componentName, hookName));
 
}// add_hook ()


////////////////////////////////////////////////////////////////////////////
// Removes a hook from a component
////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::remove_hook (char *request) 
{
 char *componentName;
 char *hookName;
 this->get_params (request, &componentName, &hookName);

 if (componentName == 0 || hookName == 0)
   {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }
 
  this->AckOrNack (Dynamic_Service_Config::instance()->delete_hook (componentName, hookName));
 
}// remove_hook ()



/////////////////////////////////////////////////////////////////////////////
// Sends information about a given loaded component stored in the ACE service
// repository.
/////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::get_impl_info (char *request) const
{
 char *implementationName;
 char buf[BUFSIZ];
 int len;
 this->get_params (request, &implementationName);
 
 if(implementationName == 0)
   {
     this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                  ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
     return;
   }

 // Now, an ugly hack to recognize ACE Services that are not 2K components: 2K
 // components must have the ":" to separate the category name from the version
 // number. If it doesn't have the ":", we assume it's not a 2K component.
 if (!strchr(implementationName, ':')) 
   {
     // Not a 2K component.
     const ACE_Service_Type *srvc = 0;
     int ret = ACE_Service_Repository::instance ()->find(implementationName, 
                                                         &srvc, 1);
     if (ret == -1 || srvc == 0) // Didn't find service in the repository.
       {
         this->client_stream_.send_n (DCP_Reply::UNKNOWN_ENTITY, 
                                      ACE_OS::strlen(DCP_Reply::UNKNOWN_ENTITY));
         return;
       }

     // The service was found, get its info.
     char *bufP = buf;
     len = srvc->type ()->info (&bufP, BUFSIZ - 4);
   }
 else 
   {
     // It is a 2K component.
     
     // Try to find implementationName in the repository
     Simple_Strategy_Factory *ssf =
       ACE_Dynamic_Service<Simple_Strategy_Factory>::instance (implementationName);
     
     // IMPORTANT: The previous call should return 0 if implementationName is not
     // of type Simple_Strategy_Factory but the implementation of
     // ACE_Dynamic_Service<>::instance does NOT use RTTI yet. While this is not
     // fixed, this won't be completely safe.
     if (ssf == 0) // Didn't find a valid implementation in the repository.
       {
         this->client_stream_.send_n (DCP_Reply::UNKNOWN_ENTITY, 
                                      ACE_OS::strlen(DCP_Reply::UNKNOWN_ENTITY));
         return;
       }
 
     // The implementationName was found, get its info.
     ACE_OS::strcpy (buf, ssf->localConfigurator ()->info ());
     len = ACE_OS::strlen(buf);
   }
 
 if (len > 0)
   {
     // Add the proper number of "\r\n" 's.
     if(buf[len - 2] != '\r' || buf[len - 1] != '\n')
       {
         buf[len++] = '\r';
         buf[len++] = '\n';
       }
     buf[len++] = '\r';
     buf[len++] = '\n';
          
     ssize_t n = this->client_stream_.send_n (buf, len);

     if (n != len || (n == -1 && errno != EPIPE))
       ACE_ERROR ((LM_ERROR, "%p\n", "send_n"));
   }
 else 
   {
     this->client_stream_.send_n (DCP_Reply::EMPTY, 
                                  ACE_OS::strlen(DCP_Reply::EMPTY));
   }
}//get_impl_info()



////////////////////////////////////////////////////////////////////////////
// Asks the dynamic service config to load the file <implementation_name>
// and register as service <category_name>. 
////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::load_implementation (char *request, char *end_of_request)
{
 char *categoryName;
 char *implementationName;
 char *parameters = 0;
 
 this->get_params (request, &categoryName, &implementationName);

 if(implementationName)
   parameters = implementationName + ACE_OS::strlen(implementationName)+1;

 if (categoryName == 0 || implementationName == 0)
   {
  	  this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }

 char pathname[MAXPATHLEN]; 

 pers_rep_->build_pathname (pathname, categoryName);

 if (parameters >= end_of_request)
   parameters = 0;

 int version = Dynamic_Service_Config::instance()->install (pathname, implementationName, 
                                                categoryName, parameters); 
 if (version > 0)
   {
     char temp[10];
     
     ACE_OS::sprintf(temp, "%d\r\n\r\n", version);
     this->client_stream_.send_n (temp, ACE_OS::strlen(temp));
   }
 else
   {
     this->client_stream_.send_n (DCP_Reply::NACK, 
                                  ACE_OS::strlen(DCP_Reply::NACK));
   }
}// load_implementation()


////////////////////////////////////////////////////////////////////////////
// Attaches a loaded implementation to a given component hook.
////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::hook_implementation (char *request)
{
 char *implementationName;
 char *componentName;
 char *hookName;
 this->get_params (request, &implementationName, &componentName, &hookName);

 if (implementationName == 0 || componentName == 0 || hookName == 0)
   {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }

 this->AckOrNack (Dynamic_Service_Config::instance()->hook (implementationName, hookName, componentName));

}// hook_implementation ()


///////////////////////////////////////////////////////////////////////////
// Suspends the execution of a loaded implementation.
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::suspend_implementation  (char *request)
{  
 char *implementationName;

 this->get_params (request, &implementationName);

 if (implementationName == 0)
   {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }

 this->AckOrNack (Dynamic_Service_Config::instance()->suspend (implementationName)); 
}// suspend_implementation()


///////////////////////////////////////////////////////////////////////////
// Resumes a previously-suspended implementation.
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::resume_implementation  (char *request)
{  
 char *implementationName;

 this->get_params (request, &implementationName);

 if (implementationName == 0)
   {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }

 this->AckOrNack (Dynamic_Service_Config::instance()->resume (implementationName)); 
}// resume_implementation()

///////////////////////////////////////////////////////////////////////////
// Removes a implementation from the in-memory ACE service repository.
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::remove_implementation (char *request)
{  
 char *implementationName;

 this->get_params (request, &implementationName);

 if (implementationName == 0)
   {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }

 int ret = Dynamic_Service_Config::instance()->remove (implementationName); 
 if (ret == 1)
   {
     this->client_stream_.send_n (DCP_Reply::DEFERRED, 
                                  ACE_OS::strlen(DCP_Reply::DEFERRED));
     return;
   }
 else
   this->AckOrNack (ret);
}// remove_implementation()

///////////////////////////////////////////////////////////////////////////
// Sends a configuration message to a loaded implementation.
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::configure_implementation (char *request, char *end_of_request)
{  
 char *implementationName;
 char *messageString = 0;
 
 this->get_params (request, &implementationName);

 if(implementationName)
   messageString = implementationName + ACE_OS::strlen(implementationName)+1;

 if (implementationName == 0 || messageString == 0)
   {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
   }
 
 // Now, an ugly hack to recognize ACE Services that are not 2K components: 2K
 // components must have the ":" to separate the category name from the version
 // number. If it doesn't have the ":", we assume it's not a 2K component.
 if (!strchr(implementationName, ':')) 
   {
     // Not a 2K component.
     const char *const msg = "configuration of non-2K component not implemented yet\r\n\r\n";
     this->client_stream_.send_n (msg, ACE_OS::strlen (msg));
     return;
   }
 else 
   {
     // It is a 2K component.
     // Try to find implementationName in the repository
     Simple_Strategy_Factory *ssf =
       ACE_Dynamic_Service<Simple_Strategy_Factory>::instance (implementationName);
     
     // IMPORTANT: The previous call should return 0 if implementationName is not
     // of type Simple_Strategy_Factory but the implementation of
     // ACE_Dynamic_Service<>::instance does NOT use RTTI yet. While this is not
     // fixed, this won't be completely safe.
     if (ssf == 0) // Didn't find a valid implementation in the repository.
       {
         this->client_stream_.send_n (DCP_Reply::UNKNOWN_ENTITY, 
                                      ACE_OS::strlen(DCP_Reply::UNKNOWN_ENTITY));
         return;
       }
 
     // The implementationName was found, sends it the messageString
     AckOrNack (ssf->configure (messageString));
   }
}// configure_implementation()

///////////////////////////////////////////////////////////////////////////
// Uploads a category implementation to the local persistent repository.
///////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::upload_implementation (char *request)
{  
  long len = -1;
  char *categoryName;
  char *implementationName;
  char *length;
  this->get_params (request, &categoryName, &implementationName, &length);
  
  if (length)
    len = ACE_OS::strtol(length, 0, 10);
  
  if (categoryName == 0 || implementationName == 0 || len <= 0)
    {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
    }

  // Creates an entry in the local persistent repository containing the first
  // len bytes received from the client_stream_.  
  // Note that this can be blocking or non-blocking according to the way that
  // the persistent repository subclass is implemented.
  this->AckOrNack 
    (pers_rep_->create_entry_from_stream(&this->client_stream_,
                                         categoryName, implementationName, 
                                         len));

}// upload_implementation ()


////////////////////////////////////////////////////////////////////////////
// Removes an implementation from the local persistent repository.
////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::delete_implementation (char *request)
{  
  char *categoryName;
  char *implementationName;
  this->get_params (request, &categoryName, &implementationName);
  
  if (categoryName == 0 || implementationName == 0)
    {
      this->client_stream_.send_n (DCP_Reply::MALFORMED_REQUEST, 
                                   ACE_OS::strlen(DCP_Reply::MALFORMED_REQUEST));
      return;
    }

  this->AckOrNack (pers_rep_-> delete_entry(categoryName, implementationName));

}// delete_implementation ()

////////////////////////////////////////////////////////////////////////////
// Restarts the ORB processing the configuration file again.
////////////////////////////////////////////////////////////////////////////
void 
DCP_Broker::restart ()
{  
  this->reconfigure_services ();
  this->AckOrNack (1);
}// delete_implementation ()


////////////////////////////////////////////////////////////////////////////
// Sends an ACK if value >= 0 and sends a NACK otherwise.
////////////////////////////////////////////////////////////////////////////
void
DCP_Broker::AckOrNack(int value) const
{
 if (value >= 0)
   {
     this->client_stream_.send_n (DCP_Reply::ACK, 
                                  ACE_OS::strlen(DCP_Reply::ACK));
   }
 else
   {
     this->client_stream_.send_n (DCP_Reply::NACK, 
                                  ACE_OS::strlen(DCP_Reply::NACK));
   }
}// AckOrNack
