/* -*- C++ -*- */
// $Id: Dynamic_Service_Config.h,v 1.2 2000/11/03 16:56:26 f-kon Exp $

// ============================================================================
//
// = LIBRARY
//    ace2K
//
// = FILENAME
//    Dynamic_Service_Config.h
//
// = AUTHOR
//    Fabio Kon
//    Based on Doug Schmidt's Service Configurator
//
// = NOTE
//    Everything in this class is static. This characteristic comes from the
//    ACE_Service_Config on which this class was based. But, we might decide
//    to change this if we find the it's a good idea to have multiple dynamic
//    service configurator objects in the future.
//
//
// ============================================================================

#if !defined (DYNAMIC_SERVICE_CONFIG_H)
#define       DYNAMIC_SERVICE_CONFIG_H

#include "ace/Service_Config.h"
#include "ace/Service_Manager.h"
#include "ace/Free_List.h"

#include "dynamicTAO/src/ComponentConfigurator/SimpleConfigurator.h"
#include "versions_list.h"
#include "dynamicTAO/src/ace/ACE_Persistent_Repository.h"
#include "ACE_2K_OS.h"


class Net_Broker;

class ACE_2K_Export Dynamic_Service_Config : public ACE_Cleanup
{
  // = DESCRIPTION
  //     Supplies common server operations for dynamic
  //     configuration of services during execution time
  // 
public:
  virtual void cleanup (void *param = 0);
  // Cleanup method, used by <ace_cleanup_destroyer> to destroy the
  // <Dynamic_Service_Config>.

  enum {MAX_SERVICES = ACE_DEFAULT_SELECT_REACTOR_SIZE};
  static const int VERSION_NUMBER_LENGTH;

  //************* Singleton Management ***************************
  static  Dynamic_Service_Config *instance (void);
  // Get pointer to a process-wide <Dynamic_Service_Config>.

  void close_singleton (void);
  // Deletes the singleton. Use with caution.


  //********** Initialization and termination methods. **********//

  // Initialize the Service Repository.
  Dynamic_Service_Config (size_t size = ACE_Service_Config::MAX_SERVICES);

  // Performs an open without parsing command-line arguments.
  Dynamic_Service_Config (const char program_name[], 
                      LPCTSTR logger_key = ACE_DEFAULT_LOGGER_KEY);


  // Performs an open without parsing command-line arguments.  Returns
  // -1 on failure and 0 otherwise.
  int open (const ASYS_TCHAR  program_name[],
                   LPCTSTR logger_key = ACE_DEFAULT_LOGGER_KEY,
                   int ignore_static_svcs = 1,
                   int ignore_default_svc_conf = 0);

  // This is the primary entry point into the Dynamic_Service_Config (the
  // constructor just handles simple initializations).  It parses
  // arguments passed in from the command-line.  Returns -1 on failure
  // and 0 otherwise.
  // Note that this initialize the Dynamic_Service_Config and not the services
  // that it contains.
  int open (int argc, ASYS_TCHAR *argv[],
                   LPCTSTR logger_key = ACE_DEFAULT_LOGGER_KEY,
                   int ignore_static_svcs = 1,
                   int ignore_default_svc_conf = 0);

  // Perform user-specified close activities and remove dynamic
  // memory.
  virtual ~Dynamic_Service_Config (void);

  // Closes Dynamic_Service_Config. Returns 0.
  // This might need to call user-specified close hooks on all of the configured
  // services in the <Service_Repository>, then delete the
  // <Service_Repository> itself. OR MAYBE NOT. DECIDE LATER.
  int close (void);

  //********** Member functions used by various other parts **********//
  //**********  of the Service Configurator class category  **********//

  static int initialize (const char *pathname,
                         const char *factoryFunc,
                         const char *serviceName,
                         int active,
                         char parameters[]);
  // dynamicall link a service getting its implementation from the file
  // system, creates an <ACE_Service_Type> and delegates the initialization to
  // initialize (const ACE_Service_Type *, char parameters[]);
    
  static int initialize (const ACE_Service_Type *, char parameters[]);
  // Inserts this service in the ACE_Service_Repository and initalizes it.
  // Dynamically link the shared object file and retrieve
  // a pointer to the designated shared object in this file.

  static int initialize (const char svc_name[], char parameters[]);
  // Initialize and activate a statically <svc_name> service.

  int install (const char *path,  const char *implementation,
                      const char *categoryName,
                      char parameters[], const char *hookName = 0, 
                      ComponentConfigurator *compConfig = 0);
  // Installs an implementation loaded from the file pathname and hooks it to
  // hookName using an application-specific configurator
  // (e.g. TAO_Configurator).


  int add_hook (char *componentName, const char *hookName);
  int delete_hook (char *componentName, const char *hookName);
  // adds and removes hooks from a given component.

  int hook (const char *implementationName, const char *hookName,
                   ComponentConfigurator *compConfig);
  int hook (const char *inplementationName, const char *hookName, 
                   char *compName);
  // Assigns an implementation to a component hook.

  int get_latest_version (const char *categoryName, char **implementationName);
  // Returns the latest version among the implementations of categoryName.
  // implementationName is an out parameter.

  int resume (const char svc_name[]);
  // Resume a <svc_name> that was previously suspended or has not yet
  // been resumed (e.g., a static service).

  int suspend (const char svc_name[]);
  // Suspend <svc_name>.  Note that this will not unlink the service
  // from the daemon if it was dynamically linked, it will mark it
  // as being suspended in the Service Repository and call the
  // suspend() member function on the appropriate ACE_Service_Object.
  // A service can be resumed later on by calling the RESUME()
  // member function...

  int remove (const char svc_name[]);
  // Remove <svc_name> from the process by removing it from the
  // ACE_Service_Repository, and our internal lists.
  // Returns: 0  if successful
  //          -1 if something failed
  //          +1 if the removal was deferred because the implementation has clients
  
  void dump (void) const;
  // Dump the state of an object.

  static ACE_INLINE void netBroker (Net_Broker *netBroker);
  static ACE_INLINE Net_Broker *netBroker ();
  // Gets or sets the Network Broker.

  static ACE_INLINE void netBroker2 (Net_Broker *netBroker2);
  static ACE_INLINE Net_Broker *netBroker2 ();
  // Gets or sets the second Network Broker.

  ACE_INLINE SimpleConfigurator *domainConfigurator ();
  // Gets the Domain Configurator.

  void get_components (char *buf, int maxsize);
  // Fills buf with list of components in this domain.
  void get_components (char **components, int &size, int maxsize);
  //Fills the array of strings with the names of the components

  int get_hooks (char *comp, char *buf, int maxsize);
  // Fills buf with a list of hooks for this component. 
  int get_hooks (char *comp, char **hooks, int &size, int maxsize);
  // Fills an array of strings with the names of the hooks.
  
  
  int get_comp_info (char *component, char *buf, int maxsize);
  // Fills buf with information about this component.

  const char *get_hooked_comp (const char *hookName, char *comp);
  // Returns the name of the component attached to the given hook.

  ACE_ALLOC_HOOK_DECLARE;
  // Declare the dynamic allocation hooks.  

protected:
  static int process_directives (void);
  // forwards the call to ACE_Service_Config in order to process the
  // <service_config_file>.

  static void parse_args (int, char *argv[]);
  // Handle the command-line options intended for the
  // <Dynamic_Service_Config>. Then calls ACE_Service_Config to take care of
  // the remaining options.  
  // Note that <argv[0]> is assumed to be the program name.

  static int load_static_svcs (void);
  // forwards the call to ACE_Service_Config. to add the default
  // statically-linked services to the <ACE_Service_Repository>.

private:

  friend class DynamicConfigurator_i;
  friend class Net_Broker;
	
  static int net_broker_entry_point();
  static int net_broker2_entry_point();

  int getTentativeVersion(const char *servName, const char *impName, 
                          Version_Record **incrementLater, int *alreadyLoaded);
  // Find whether or not this is the first time we are adding the service, and
  // return a correct version number. If the implementation specified has
  // already been loaded it returns -1. It does not add any element in any list and it does not
  // increment the stored version number (because implementation could fail when being loaded
  // into memory. UpdateVersion is the responsible for storing the changes.

  int updateVersion(Version_Record *tempCat, const char *catName, const char *impName);
  //Updates all the changes.
  
  static void listHooksOnString (ComponentConfigurator *c, char *buf, int maxsize);
  // Fills buf with a list of hooks for the given ComponentConfigurator

  static void listHooksOnString (ComponentConfigurator *c, char **buf, int &size, int maxsize);
  // Fills the array of strings "buf" with a list of hooks for the given ComponentConfigurator

  ComponentConfigurator *getConfigurator (char *componentPathName);
  // Returns a pointer to the configurator of the component given by the
  // pathname in relation to the domain configurator. Pathname components must
  // be separated by the '/' character. The "/" path returns the domainConfigurator_.
  
  // = Set by command-line options.
  static char verbose_;
  static int debug_;

  static Dynamic_Service_Config *instance_;
  static int delete_instance_;
  // Later we must consider if this should  really be a singleton or not.

  static Net_Broker * netBroker_;
  static Net_Broker * netBroker2_;
  // the network brokers that receives remote requests.

  static ACE_Persistent_Repository *pers_rep_;

  static ACE_Persistent_Repository *pers_rep();
  //Accessor method for the persistent repository

  static void pers_rep(ACE_Persistent_Repository *rep);

  SimpleConfigurator domainConfigurator_;

  Versions_List versions_list_;
  Implementations_List imp_list_;

  // Why are the following iterators class static variables and not local variables in the
  // method in which they are used? Fabio.
  ACE_Double_Linked_List_Iterator<Version_Record> *ver_li_; // iterator
  ACE_Double_Linked_List_Iterator<Implementation_Record> *imp_li_; // iterator
};

#if defined (__ACE_INLINE__)
#include "Dynamic_Service_Config.i"
#endif /* __ACE_INLINE__ */


#endif /* ACE_SERVICE_CONFIG_H */

